/*
 *  pe_chain_encoding.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/10/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern int seek_previous_target_atom(int ind);
extern int seek_next_target_atom(int ind);
extern int search_sequence(int ind);
extern void clear_slt_tile(struct pe_slt_tile * pt_dst_tile);
extern int seek_init_tile(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct,
			  PE_DIRECTION direction);
extern void parallel_cancelation(int ind, int seq_ct, int t_ind,
				 struct pe_slt_tile * slt_tiles,
				 PE_DIRECTION direction);
extern int step_toward_new_ca_atom(int t_ind, struct pe_slt_tile * slt_tiles,
				   int atm_ct, PE_DIRECTION direction);
extern int step_toward_new_atom(int t_ind, struct pe_slt_tile * slt_tiles,
				int atm_ct, int seq_ct, PE_DIRECTION direction);
extern void move_tile_to_target(int t_ind, struct pe_slt_tile * slt_tiles,
				int atm_ct);
extern void rotate_tile_and_trn_table(int t_ind, struct pe_slt_tile * slt_tiles);

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;
extern struct pe_trn_table my_trn_table[];

void copy_slt_tile(struct pe_slt_tile * pt_dst_tile,
		   struct pe_slt_tile * pt_src_tile)
{
	pt_dst_tile->cx = pt_src_tile->cx;
	pt_dst_tile->cy = pt_src_tile->cy;
	pt_dst_tile->cz = pt_src_tile->cz;
	pt_dst_tile->tx = pt_src_tile->tx;
	pt_dst_tile->ty = pt_src_tile->ty;
	pt_dst_tile->tz = pt_src_tile->tz;
	pt_dst_tile->dx = pt_src_tile->dx;
	pt_dst_tile->dy = pt_src_tile->dy;
	pt_dst_tile->dz = pt_src_tile->dz;
	pt_dst_tile->bx = pt_src_tile->bx;
	pt_dst_tile->by = pt_src_tile->by;
	pt_dst_tile->bz = pt_src_tile->bz;
	pt_dst_tile->error = pt_src_tile->error;
	pt_dst_tile->error_atom = pt_src_tile->error_atom;
	pt_dst_tile->atom_flag = pt_src_tile->atom_flag;
	pt_dst_tile->top[0] = pt_src_tile->top[0];
	pt_dst_tile->top[1] = pt_src_tile->top[1];
	pt_dst_tile->top[2] = pt_src_tile->top[2];
	pt_dst_tile->top[3] = pt_src_tile->top[3];
	pt_dst_tile->type = pt_src_tile->type;
	pt_dst_tile->code = pt_src_tile->code;
	pt_dst_tile->ter = pt_src_tile->ter;
	return;
}

int number_of_target_atoms(int seq_ct)
{
	int ct;
	int ind;
	
	ct = 0;
	for (ind = my_protein.sequence[seq_ct].atom_head;
	     ind <= my_protein.sequence[seq_ct].atom_tail; ind++) {
		if (THIS_IS_NOT_CA_ATOM(ind)) continue;
		ct++;
	}
	return ct;
}

int set_end_tile_to_down(int t_ind, struct pe_slt_tile * slt_tiles)
{
	int prv_ind;
	int prv_type, cur_type;
	int i;
	
	if (t_ind <= 1) {
		fprintf(STDERR,
			"set_end_tile_to_down() : tile index=%d <= 1\n",
			t_ind);
		return FAIL;
	}
	slt_tiles[t_ind].code = C_DOWN;
	prv_ind = t_ind -1;
	prv_type = slt_tiles[prv_ind].type;
	
	if (slt_tiles[prv_ind].code == C_DOWN) {/* D (prv) -> D (end) */
		slt_tiles[t_ind].type = my_trn_table[prv_type].downward_D_type;
		for (i = 0; i < 4; i++)
			slt_tiles[t_ind].top[i]
				= my_trn_table[prv_type].downward_D_top[i]
				+ slt_tiles[prv_ind].top[i];
	} else if (slt_tiles[prv_ind].code == C_UP) {	/* U (prv) -> D (end) */
		slt_tiles[t_ind].type = my_trn_table[prv_type].upward_D_type;
		for (i = 0; i < 4; i++)
			slt_tiles[t_ind].top[i]
				= my_trn_table[prv_type].upward_D_top[i]
				+ slt_tiles[prv_ind].top[i];
	} else {
		fprintf(STDERR,
			"set_end_tile_to_down() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			prv_ind);
		exit(1);
	}
	cur_type = slt_tiles[t_ind].type;
	slt_tiles[t_ind].tx = slt_tiles[t_ind].cx - my_trn_table[cur_type].cx;
	slt_tiles[t_ind].ty = slt_tiles[t_ind].cy - my_trn_table[cur_type].cy;
	slt_tiles[t_ind].tz = slt_tiles[t_ind].cz - my_trn_table[cur_type].cz;
	return SUCCESS;
}

int set_top_tile_to_down(int t_ind, struct pe_slt_tile * slt_tiles)
{
	int nxt_ind;
	int nxt_type, cur_type;
	int i;
	
	if (t_ind >= my_protein.tile_max_num) {
		fprintf(STDERR,
			"set_end_tile_to_down() : ERROR t_ind=%d >= max_tile\n",
			t_ind);
		return FAIL;
	}
	slt_tiles[t_ind].code = C_DOWN;
	nxt_ind = t_ind + 1;
	nxt_type = slt_tiles[nxt_ind].type;
	
	if (slt_tiles[nxt_ind].code == C_DOWN) { /* D (top) <- D (next) */
		slt_tiles[t_ind].type = my_trn_table[nxt_type].upward_U_type;
		for (i = 0; i < 4; i++)
			slt_tiles[t_ind].top[i]
				= my_trn_table[nxt_type].upward_U_top[i]
				+ slt_tiles[nxt_ind].top[i];
	} else if (slt_tiles[nxt_ind].code == C_UP) {	/* D (top) <- U (next) */
		slt_tiles[t_ind].type = my_trn_table[nxt_type].downward_U_type;
		for (i = 0; i < 4; i++)
			slt_tiles[t_ind].top[i]
				= my_trn_table[nxt_type].downward_U_top[i]
				+ slt_tiles[nxt_ind].top[i];
	} else {
		fprintf(STDERR,
			"set_top_tile_to_down() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			nxt_ind);
		exit(1);
	}
	cur_type = slt_tiles[t_ind].type;
	slt_tiles[t_ind].tx = slt_tiles[t_ind].cx - my_trn_table[cur_type].cx;
	slt_tiles[t_ind].ty = slt_tiles[t_ind].cy - my_trn_table[cur_type].cy;
	slt_tiles[t_ind].tz = slt_tiles[t_ind].cz - my_trn_table[cur_type].cz;
	return SUCCESS;
}

/* In the case of CA backbones, add terminal N and C atoms at both ends. */
void set_terminal_tile(int t_ind, struct pe_slt_tile * slt_tiles, 
		       PE_DIRECTION direction)
{
	if (direction == DIR_FORWARD) {
		if (set_end_tile_to_down(t_ind, slt_tiles) == FAIL)
			exit(1);
	} else {
		if (set_top_tile_to_down(t_ind, slt_tiles) == FAIL)
			exit(1);
	}
	return;
}

void set_terminal_atom(int t_ind, int seq_ct, struct pe_slt_tile * slt_tiles,
		       PE_DIRECTION direction)
{
	/* Set the first tile of the current chain */
	if (direction == DIR_BACKWARD)
		my_protein.sequence[seq_ct].tile_head = t_ind;
	else {
		my_protein.sequence[seq_ct].tile_tail = t_ind;
		slt_tiles[t_ind].ter = 1;
	}
	
	if (my_settings.ca_only_flag == TRUE
	    && my_settings.rotate_tile != TRUE
	    && number_of_target_atoms(seq_ct) > 1) {
		set_terminal_tile(t_ind, slt_tiles, direction);
	}
	return;
}

int seek_terminal_of_new_chain(int ind, PE_DIRECTION direction, int * pseq_ct)
{
	int seq_ct;
	
	seq_ct = *pseq_ct;
	if (direction == DIR_BACKWARD) {
		seq_ct--;
		while ((seq_ct > 1)
		       && (my_protein.sequence[seq_ct].atom_head <= 0))
			seq_ct--;	/* Skip DNA and RNA */
		ind--;	
		if  (THIS_IS_NOT_CA_ATOM(ind))
			ind = seek_previous_target_atom(ind);
	} else {
		seq_ct++;
		while ((seq_ct < my_protein.seq_num)
		       && (my_protein.sequence[seq_ct].atom_head <= 0))
			seq_ct++;	/* Skip DNA and RNA */
		ind++;
		if  (THIS_IS_NOT_CA_ATOM(ind))
			ind = seek_next_target_atom(ind);
	}
	
	if (ind < my_protein.sequence[seq_ct].atom_head
	    || ind > my_protein.sequence[seq_ct].atom_tail || ind < 0) {
		fprintf(STDERR,
			"search_new_chain() : ERROR No target atom (from %d to %d of seq %d).\n",
			my_protein.sequence[seq_ct].atom_head,
			my_protein.sequence[seq_ct].atom_tail, seq_ct);
		return -1;;
	}
	*pseq_ct = seq_ct;
	return ind;
}

int set_tile_of_terminal_atom(int ind, int seq_ct, int tile_ind,
			      struct pe_slt_tile * slt_tiles,
			      PE_DIRECTION direction)
{
	int t_ind;
	int prv_ind, nxt_ind;
	
	t_ind = seek_init_tile(tile_ind, slt_tiles, ind, direction);
	slt_tiles[t_ind].atom_flag = 1;	/* It is associated with an atom. */
	if (direction == DIR_BACKWARD) {
		my_protein.sequence[seq_ct].tile_tail = t_ind;
		slt_tiles[t_ind].ter = 1;
	} else {
		my_protein.sequence[seq_ct].tile_head = t_ind;
	}
	
	if (my_settings.move_to_target == TRUE
	    || my_settings.rotate_tile == TRUE) {
		move_tile_to_target(t_ind, slt_tiles, ind);
	}
	prv_ind = seek_previous_target_atom(ind);
	if (prv_ind < 0)
		prv_ind = ind;
	nxt_ind = seek_next_target_atom(ind);
	if (nxt_ind < 0)
		nxt_ind = ind;
	slt_tiles[t_ind].dx = my_protein.atom[nxt_ind].x
		- my_protein.atom[prv_ind].x;
	slt_tiles[t_ind].dy = my_protein.atom[nxt_ind].y
		- my_protein.atom[prv_ind].y;
	slt_tiles[t_ind].dz = my_protein.atom[nxt_ind].z
		- my_protein.atom[prv_ind].z;
	if (my_settings.rotate_tile == TRUE) {
		rotate_tile_and_trn_table(t_ind, slt_tiles);
	}
	return t_ind;
}

int set_tile_of_new_target_atom(int ind, int seq_ct, int tile_ind,
				struct pe_slt_tile * slt_tiles,
				PE_DIRECTION direction)
{
	int t_ind;
	int prv_ind, nxt_ind;
	
	if (my_settings.ca_only_flag == TRUE && my_settings.rotate_tile != TRUE)
		t_ind = step_toward_new_ca_atom(tile_ind, slt_tiles, ind,
						direction);
	else
		t_ind = step_toward_new_atom(tile_ind, slt_tiles, ind, seq_ct,
					     direction);
	slt_tiles[t_ind].atom_flag = 1;	/* It is associated with an atom. */
	
	/* Set dx, dy, dy. */
	prv_ind = seek_previous_target_atom(ind);
	if (prv_ind < 0)
		prv_ind = ind;
	nxt_ind = seek_next_target_atom(ind);
	if (nxt_ind < 0)
		nxt_ind = ind;
	slt_tiles[t_ind].dx = my_protein.atom[nxt_ind].x
		- my_protein.atom[prv_ind].x;
	slt_tiles[t_ind].dy = my_protein.atom[nxt_ind].y
		- my_protein.atom[prv_ind].y;
	slt_tiles[t_ind].dz = my_protein.atom[nxt_ind].z
		- my_protein.atom[prv_ind].z;
	if (my_settings.rotate_tile == TRUE) {
		rotate_tile_and_trn_table(t_ind, slt_tiles);
	}
	return t_ind;
}

/* Encoding toward the N-terminus (downward): from atm_tail to atm_head */
/* It returns the brand-new current tile if successful and -1 otherwise. */
/* (t_ind = the current tile) */
int backward_encoding(int atm_head, int atm_tail,
		      struct pe_slt_tile * slt_tiles,
		      int cur_tile /* current tile */,
		      int tile_head /* top of empty entries */)
{
	int ind, seq_ct;
	int t_ind;
	int i;
	
	if ((cur_tile <= 0) || (cur_tile > my_protein.tile_max_num)) {
		fprintf(STDOUT,
			"backward_encoding() : ERROR cur_tile = %d <=> my_protein.tile_max_num = %d\n",
			cur_tile, my_protein.tile_max_num);
		return -1;
	}
	
	t_ind = cur_tile;
	ind = seek_previous_target_atom(atm_tail + 1 /* include current one */);
	if (ind < atm_head || ind < 0) {
		fprintf(STDOUT,
			"backward_encoding() : CAUTION There exists no target atom (from %d to %d)\n",
			atm_head, atm_tail);
		return cur_tile;
	}
	seq_ct = search_sequence(ind);
	while (ind > atm_head) {
		if (ind == my_protein.sequence[seq_ct].atom_head) {
			set_terminal_atom(t_ind, seq_ct, slt_tiles,
					  DIR_BACKWARD);
			ind = seek_terminal_of_new_chain(ind, DIR_BACKWARD,
							 &seq_ct);
			t_ind = set_tile_of_terminal_atom(ind, seq_ct, t_ind,
						slt_tiles, DIR_BACKWARD);
		} else {
			ind--;
			if (THIS_IS_NOT_CA_ATOM(ind)) continue;
			t_ind = set_tile_of_new_target_atom(ind, seq_ct, t_ind,
						slt_tiles, DIR_BACKWARD);
		}
	}
	if (ind == my_protein.sequence[seq_ct].atom_head) {
		set_terminal_atom(t_ind, seq_ct, slt_tiles, DIR_BACKWARD);
	}
	
	/* Move tiles: slt_tiles[tile_head] <== slt_tiles[t_ind] */
	for (i = tile_head; t_ind <= my_protein.tile_max_num; i++, t_ind++) {
		copy_slt_tile(slt_tiles + i, slt_tiles + t_ind);
		clear_slt_tile(slt_tiles + t_ind);
	}
	return (i - 1);	
}

/* Encoding toward the C-terminus (upward): from atm_head to atm_tail */
int forward_encoding(int atm_head, int atm_tail,
		     struct pe_slt_tile * slt_tiles,
		     int cur_tile /* current tile */)
{
	int ind, seq_ct;
	int t_ind;
	
	if ((cur_tile > my_protein.tile_max_num) || (cur_tile <= 0)) {
		fprintf(STDOUT,
			"forward_encoding() : ERROR cur_tile= %d <=> my_protein.tile_max_num= %d\n",
			cur_tile, my_protein.tile_max_num);
		return -1;
	}
	
	t_ind = cur_tile;
	ind = seek_next_target_atom(atm_head - 1 /* include current one */);
	if (ind > atm_tail || ind < 0) {
		fprintf(STDOUT,
			"forward_encoding() : CAUTION There exists no target atom (from %d to %d)\n",
			atm_head, atm_tail);
		return cur_tile;/* brand-new current tile */
	}
	seq_ct = search_sequence(ind);
	while (ind < atm_tail) {
		if (0) {
			fprintf(STDOUT,
				"forward_encoding() : t_ind= %d <= my_protein.tile_max_num= %d\n",
				t_ind, my_protein.tile_max_num);
		}
		if (ind == my_protein.sequence[seq_ct].atom_tail) {
			set_terminal_atom(t_ind, seq_ct, slt_tiles,
					  DIR_FORWARD);
			ind = seek_terminal_of_new_chain(ind, DIR_FORWARD,
							 &seq_ct);
			t_ind = set_tile_of_terminal_atom(ind, seq_ct, t_ind,
							slt_tiles, DIR_FORWARD);
		} else {
			ind++;
			if (THIS_IS_NOT_CA_ATOM(ind)) continue;
			t_ind = set_tile_of_new_target_atom(ind, seq_ct, t_ind,
							slt_tiles, DIR_FORWARD);
		}
	}
	if (ind == my_protein.sequence[seq_ct].atom_tail) {
		set_terminal_atom(t_ind, seq_ct, slt_tiles, DIR_FORWARD);
	}
	return t_ind;
}

/* Toward the N-terminus (backward) */
/* Toward the C-terminus (upward) */
int chain_encoding(int ind, int * ptile_ind, struct pe_slt_tile * slt_tiles)
{
	int t_ind;
	
	t_ind = *ptile_ind;
	if (t_ind >= 1) {
		t_ind = backward_encoding(1	/* N-terminus */,
					  ind	/* current atom */,
					  slt_tiles, t_ind /* current tile */,
					  1 /* top empty entry of the table */);
	}
	if (DEBUG && ALL_ENC) {
		fprintf(STDOUT,
			"chain_encoding() :  Backward encoding done (t_ind= %d, ind= %d)\n",
			t_ind, ind);
		if (0) print_my_slt_tiles(slt_tiles);
		fflush(STDOUT);
	}

	if (t_ind < 0) {
		*ptile_ind = t_ind;
		return FAIL;
	}
	if (my_settings.move_to_target == TRUE) {
		move_tile_to_target(t_ind, slt_tiles, ind);
	}
	if (my_settings.rotate_tile == TRUE) {
		rotate_tile_and_trn_table(t_ind, slt_tiles);
	}
	if (t_ind <= my_protein.tile_max_num) {
		t_ind = forward_encoding(ind /* current atom */,
					 my_protein.atm_num /* C-terminus */,
					 slt_tiles, t_ind /* current tile */);
	}
	if (DEBUG && ALL_ENC) {
		fprintf(STDOUT,
			"chain_encoding() :  Forward encoding done(t_ind= %d, ind= %d)\n",
			t_ind, ind);
		if (0) print_my_slt_tiles(slt_tiles);
		fflush(STDOUT);
	}
	*ptile_ind = t_ind;
	return SUCCESS;
}
