/*
 *  pe_encoding.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/01/05.
 *  Copyright (C) 2005, 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-03 fix MidPoint setting
 *  2006-11-04 add "if (0)" in check_tiles()
 *  2006-11-20 version 0.2 release
 *  2006-11-20 version 0.2 release
 *  2007-01-12 add my_settings.try_all_type (remove TRY_ALL_TYPE)
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 *  2007-08-22 option "R" is supported in distance_check(), encode_protein_structure()
 *  2007-08-22 Change the initial atom if neccesary in encode_protein_structure()
 *  2008-08-22 version 0.5 release
 *  2007-12-23 Relax a condition in check_tile_direction()
 *  2008-01-02 version 0.6 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern void print_trn_table(void);
extern int seek_previous_target_atom(int ind);
extern int seek_next_target_atom(int ind);
extern int setup_tables_and_init_tile(float ave_dis, int ind,
				      struct pe_slt_tile * slt_tiles,
				      int init_tile_type);
extern int chain_encoding(int ind, int * ptile_ind,
			  struct pe_slt_tile * slt_tiles);
extern int search_sequence(int ind );
extern void encoding_suite(int best_ind, float ave_dis,
			   struct pe_slt_tile * slt_tiles);

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;

struct pe_coor_system my_coor;
struct local_table {
	float ave_error;
	int tile_num;
};

void print_loop_label(struct pe_init_tile global_bst_tile, int for_ct,
		      float ave_dis)
{
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, 
		"--------------------------------------------------------------------------\n");
	fprintf(STDOUT,
		"|  [%d-th for loop]    ave_dis = %f\n", for_ct, ave_dis);
	fprintf(STDOUT,
		"|  global_best_ind = %d, global_best_t_ind = %d, global_best_t_type = %d\n",
		global_bst_tile.ind, global_bst_tile.t_ind, global_bst_tile.t_type);
	fprintf(STDOUT,
		"|  best_ave_dis = %f, global_best_ave_error= %f\n",
		global_bst_tile.ave_dis, global_bst_tile.ave_error);
	fprintf(STDOUT,
		"--------------------------------------------------------------------------\n");
	return;
}

void print_global_best(struct pe_init_tile global_bst_tile)
{
	int seq_ct;
	
	seq_ct = search_sequence(global_bst_tile.ind);
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"print_global_best() : ************************************\n");
	fprintf(STDOUT,
		"print_global_best() : Best atom = the %d-th atom",
		global_bst_tile.ind);
	if (my_protein.atom[global_bst_tile.ind].atom_name == PE_CC)
		fprintf(STDOUT, " (C) ");
	else if (my_protein.atom[global_bst_tile.ind].atom_name == PE_CA)
		fprintf(STDOUT, " (CA) ");
	else if (my_protein.atom[global_bst_tile.ind].atom_name == PE_NN)
		fprintf(STDOUT, " (N) ");
	else
		fprintf(STDOUT, " (?) ");
	fprintf(STDOUT, " in the %d-th chain (Num of tiles= %d)\n",
		seq_ct, global_bst_tile.t_ind);
	fprintf(STDOUT,
		"print_global_best() :     best_ave_error= %f, best_ave_dis= %f\n",
		global_bst_tile.ave_error, global_bst_tile.ave_dis);
	fprintf(STDOUT,
		"print_global_best() :     best_t_ind= %d, best_t_type= %d\n",
		global_bst_tile.t_ind, global_bst_tile.t_type);
	fprintf(STDOUT,
		"print_global_best() : ************************************\n");
	return;
}

void print_all_ave_diss(float for_head, float for_tail, float for_step,
			struct local_table * local_tbl)
{
	int for_ct;
	float ave_dis;
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"print_all_ave_diss() :     | ave_dis  | ave_error | tile num. |\n");
	fprintf(STDOUT,
		"print_all_ave_diss() :     |----------|-----------|-----------|\n");
	for_ct = 0;
	for (ave_dis = for_head; ave_dis <= for_tail;
	     ave_dis = ave_dis + for_step, for_ct++) {
		fprintf(STDOUT,
			"print_all_ave_diss() :     | %8.3f |  %8.3f |  %8d |\n",
			ave_dis, local_tbl[for_ct].ave_error,
			local_tbl[for_ct].tile_num);
	}
	return;
}

void print_untried_atoms(void)
{
	int ind, count;
	
	fprintf(STDOUT, "print_untried_atoms() : Untried atoms\n");
	count = 0;
	for (ind = 1; ind <= my_protein.atm_num; ind++) {
		if (((my_protein.atom[ind].atom_name == PE_CA)
		     || (my_protein.atom[ind].atom_name == PE_CC)
		     || (my_protein.atom[ind].atom_name == PE_NN))
		    && (my_protein.atom[ind].can_be_init_atom == YES)) {
			fprintf(STDOUT,
				"print_untried_atoms() : Untried atom[%d] (err= %f, flag= %d,",
				ind, my_protein.atom[ind].err,
				my_protein.atom[ind].can_be_init_atom);
			if (my_protein.atom[ind].atom_name == PE_CC)
				fprintf(STDOUT, " C )\n");
			else if (my_protein.atom[ind].atom_name== PE_CA)
				fprintf(STDOUT, " CA)\n");
			else if (my_protein.atom[ind].atom_name== PE_NN)
				fprintf(STDOUT, " N )\n");
			else
				fprintf(STDOUT, " UN)\n");
			count++;
		}
	}
	fprintf(STDOUT,
		"print_untried_atoms() : Nom of Untried atom= %d\n",
		count);
	return;
}

void print_init_atom(int ind)
{
	int seq_ct;
	
	seq_ct = search_sequence(ind);
	if (my_settings.ca_only_flag == TRUE) {
		fprintf(STDOUT,
			"print_init_atom() : Initial CA = the %d-th atom in the %d-th chain\n",
			ind, seq_ct);
	} else {
		fprintf(STDOUT,
			"print_init_atom() : Initial atom = the %d-th atom in the %d-th chain\n",
			ind, seq_ct);
	}
	return;
}

void clear_slt_tile(struct pe_slt_tile * pt_dst_tile)
{
	pt_dst_tile->cx = 0.0;
	pt_dst_tile->cy = 0.0;
	pt_dst_tile->cz = 0.0;
	pt_dst_tile->tx = 0.0;
	pt_dst_tile->ty = 0.0;
	pt_dst_tile->tz = 0.0;
	pt_dst_tile->bx = 0.0;
	pt_dst_tile->by = 0.0;
	pt_dst_tile->bz = 0.0;
	pt_dst_tile->dx = 0.0;
	pt_dst_tile->dy = 0.0;
	pt_dst_tile->dz = 0.0;
	pt_dst_tile->error = -1.0;
	pt_dst_tile->error_atom = 0;
	pt_dst_tile->atom_flag = 0;
	pt_dst_tile->top[0] = 0;
	pt_dst_tile->top[1] = 0;
	pt_dst_tile->top[2] = 0;
	pt_dst_tile->top[3] = 0;
	pt_dst_tile->type = -1;
	pt_dst_tile->code = C_UNKNOWN;
	pt_dst_tile->ter = 0;
	return;
}

/* Bubble sort */
void b_sort(struct pe_error_table * pt_tab, int tab_tail)
{
	int i, flag;
	int tmp_ind;
	float tmp_err;
	
	if (tab_tail == 0) return;
	flag = 1;
	while (flag > 0) {
		flag = 0;
		for (i = 0; i < tab_tail; i++) {
			if (pt_tab[i].err < pt_tab[i + 1].err) {
				flag = 1;
				tmp_err = pt_tab[i].err;
				tmp_ind = pt_tab[i].ind;
				pt_tab[i].err = pt_tab[i + 1].err;
				pt_tab[i].ind = pt_tab[i + 1].ind;
				pt_tab[i + 1].err = tmp_err;
				pt_tab[i + 1].ind = tmp_ind;
			}
		}
	}
	return;	
}

int nxt_init_error_check(int i, struct pe_slt_tile * slt_tiles)
{
	if (slt_tiles[i].error > 100.0) {
		fprintf(STDERR,
			"nxt_init_error_check() : ERROR slt_tiles[%d].error= %f > 100.0 against %d-th atom ",
			i, slt_tiles[i].error, slt_tiles[i].error_atom);
		return FAIL;
	}
	return SUCCESS;
}

/* pt_tab: error_tab[ErrorTableSize] */
int renew_error_table(struct pe_error_table * pt_tab,
		      struct pe_slt_tile * slt_tiles)
{
	int i, ind, t_ind;
	int tab_tail;
	
	/* Reset table */
	tab_tail = -1;
	for (i = 0; i < my_settings.error_tab_size; i++) {
		pt_tab[i].err = -1.0;
		pt_tab[i].ind = 0;
	}
	
	/* Renew error value of atoms */
	for (ind = 1; ind <= my_protein.atm_num; ind++) {
		my_protein.atom[ind].err = -1.0;
	}
	for (t_ind = 1; t_ind <= my_protein.tile_num; t_ind++) {
		if (slt_tiles[t_ind].error > 0) {
			ind = slt_tiles[t_ind].error_atom;
			if (my_protein.atom[ind].err <= 0) {
				my_protein.atom[ind].err
					= slt_tiles[t_ind].error;
			} else if (my_protein.atom[ind].err
				   > slt_tiles[t_ind].error) {
				/* a nearer tile */
				my_protein.atom[ind].err
					= slt_tiles[t_ind].error;
			}
		}
	}
	
	/* Pick the worst untried ErrorTableSize CA/C/N atoms */
	/* for (ind=1; ind <= my_protein.atm_num; ind++) { */
	for (ind = 2; ind < my_protein.atm_num; ind++) {
		if (((my_protein.atom[ind].atom_name == PE_CA)
		     || (my_protein.atom[ind].atom_name == PE_CC)
		     || (my_protein.atom[ind].atom_name == PE_NN))
		    && (my_protein.atom[ind].can_be_init_atom == YES)) {
			if (my_protein.atom[ind].err > 0) {
				if (tab_tail < my_settings.error_tab_size - 1) {
					tab_tail++;
					pt_tab[tab_tail].err
						= my_protein.atom[ind].err;
					pt_tab[tab_tail].ind = ind;
					b_sort(pt_tab, tab_tail);
				} else if (pt_tab[tab_tail].err
					   < my_protein.atom[ind].err) {
					pt_tab[tab_tail].err
						= my_protein.atom[ind].err;
					pt_tab[tab_tail].ind = ind;
					b_sort(pt_tab, tab_tail);
				}
			}
		}
	}
	
	return tab_tail;	/* Number of valid entries */
}

#define CONTAINS_ATOM_HEAD(seq_ct) \
(my_protein.sequence[(seq_ct)].atom_head <= atm_head \
 && atm_head <= my_protein.sequence[(seq_ct)].atom_tail)
#define CONTAINS_ATOM_TAIL(seq_ct) \
(my_protein.sequence[(seq_ct)].atom_head <= atm_tail \
 && atm_tail <= my_protein.sequence[(seq_ct)].atom_tail)

int convert_atom_range_into_seq_range(int atm_head, int atm_tail,
				      int * pseq_head, int * pseq_tail)
{
	int seq_head, seq_tail;
	int mod_ct, seq_ct;
	
	seq_head = -1;
	seq_tail = -1;
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			if (CONTAINS_ATOM_HEAD(seq_ct)) {
				seq_head = seq_ct;
			}
			if (CONTAINS_ATOM_TAIL(seq_ct)) {
				seq_tail = seq_ct;
				break;
			}
		}
	}
	*pseq_head = seq_head;
	*pseq_tail = seq_tail;
	
	if ((seq_head > 0) && (seq_tail > 0))
		return SUCCESS;
	else
		return FAIL;
}

#define LENGTH_UPPER_LIMIT 1.8
#define LENGTH_LOWER_LIMIT 1.25
#define LENGTH_UPPER_CA_LIMIT 5.0
#define LENGTH_LOWER_CA_LIMIT 3.0
#define LENGTH_UPPER_P_LIMIT 10.0	// 2007-08-22
#define LENGTH_LOWER_P_LIMIT  4.0	// 2007-08-22

void distance_check(float cur_dis, int atm_ct, int nxt_atm_ct)
{
	float upper_limit, lower_limit;
	if (my_settings.encode_RNA != TRUE) {	// 2007-08-22
		upper_limit = (my_settings.ca_only_flag == TRUE) ? LENGTH_UPPER_CA_LIMIT
								 : LENGTH_UPPER_LIMIT;
		lower_limit = (my_settings.ca_only_flag == TRUE) ? LENGTH_LOWER_CA_LIMIT
								 : LENGTH_LOWER_LIMIT;
	} else {
		upper_limit = LENGTH_UPPER_P_LIMIT;
		lower_limit = LENGTH_LOWER_P_LIMIT;
	}
	
	if (cur_dis < lower_limit) {
		fprintf(STDERR,
			"distance_check() : DATA ERROR (bond length = %8.4f < %3.2f at %8d-th residue of chain %c [%.3s-%.3s])\n",
			cur_dis, lower_limit, my_protein.atom[atm_ct].res_seq,
			my_protein.atom[atm_ct].chain_id,
			my_protein.atom[atm_ct].res_name,
			my_protein.atom[nxt_atm_ct].res_name);
	} else if (cur_dis > upper_limit) {
		fprintf(STDERR,
			"distance_check() : DATA ERROR (bond length = %8.4f > %3.2f at %8d-th residue of chain %c [%.3s-%.3s])\n",
			cur_dis, upper_limit, my_protein.atom[atm_ct].res_seq,
			my_protein.atom[atm_ct].chain_id,
			my_protein.atom[atm_ct].res_name,
			my_protein.atom[nxt_atm_ct].res_name);
	}
	return;
}

void invalidate_terminal_and_isolated_atoms(void)
{
	int mod_ct, seq_ct, atm_ct;
	int num_of_atoms;
	
	/* Invalidate terminal atoms */
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
				my_protein.atom[atm_ct].can_be_init_atom = NONO;
				break;
			}
			for (atm_ct = my_protein.sequence[seq_ct].atom_tail;
			     atm_ct >= my_protein.sequence[seq_ct].atom_head;
			     atm_ct--) {
				if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
				my_protein.atom[atm_ct].can_be_init_atom = NONO;
				break;
			}
		}
	}
	
	/* Invalidate isolated atoms */
	num_of_atoms = 0;
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
				num_of_atoms++;
			}
			if (num_of_atoms >= 3) continue; /* head + 1 + tail */
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
				my_protein.atom[atm_ct].can_be_init_atom = NONO;
			}
		}
	}
	return;
}

int choose_initial_atom(int atm_head, int atm_tail)
{
	int ind;
	int atm_ct, seq_ct;
	
	ind = 0;
	for (atm_ct = atm_head; atm_ct <= atm_tail; atm_ct++) {
		if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
		if (my_protein.atom[atm_ct].can_be_init_atom == YES) {
			ind = atm_ct;
			break;	/* atm_ct */
		}
	}
	
	if (ind == 0) {
		fprintf(STDERR,
			"choose_initial_atom() : ERROR  - no initial candidate from the %d-th atom to the %d-th atom -\n",
			atm_head, atm_tail);
		exit(1);
	}
	if (0) {
		seq_ct = search_sequence(ind);
		fprintf(STDOUT,
			"choose_initial_atome() : The first atom = the %d-th atom in the %d-th chain\n",
			ind, seq_ct);
		fflush(STDOUT);
	}
	return ind;
}

int choose_next_initial_atom(struct pe_init_tile cur_tile,
			     struct pe_slt_tile * slt_tiles,
			     struct pe_error_table * error_tab,
			     int * perr_tab_ct, int * perr_tab_max,
			     struct pe_init_tile * pbst_tile)
{
	int i;
	int count;
	float new_ave_error;
	
	/* Error calculation */
	new_ave_error = 0.0;
	count = 0;
	for (i = 1; i <= cur_tile.t_ind; i++) {
		/*
		 if (my_protein.atom[slt_tiles[i].error_atom].atom_name
		     == PE_CA) {
			 new_ave_error = new_ave_error
			 + slt_tiles[i].error*slt_tiles[i].error;
		 }
		 */
		if (slt_tiles[i].atom_flag > 0) {
			if (nxt_init_error_check(i, slt_tiles) == FAIL) exit(1);
			new_ave_error = new_ave_error + slt_tiles[i].error;
			count++;
		}
	}
	cur_tile.ave_error = new_ave_error / (float) count;
	if (0)
		fprintf(STDOUT,
			"choose_next_initial_atom() :     new_ave_error= %f <=> best_ave_error= %f (#atom = %d)\n",
			cur_tile.ave_error, pbst_tile->ave_error, count);
	
	if ((cur_tile.ave_error < pbst_tile->ave_error)
	    || (pbst_tile->ave_error < 0.0)) {
		/* Create a new error table. */
		pbst_tile->ave_error = cur_tile.ave_error;
		pbst_tile->ind = cur_tile.ind;
		pbst_tile->t_ind = cur_tile.t_ind;
		pbst_tile->t_type = cur_tile.t_type;
		
		/* Choose the worstly-approximated
			ErrorTableSize atoms. */
		*perr_tab_max = renew_error_table(error_tab, slt_tiles);
		*perr_tab_ct = 0;
		if (*perr_tab_max < 0) {
			if (DEBUG)
				fprintf(STDOUT, "choose_next_initial_atom() :  No more atoms (err_tab_max = %d < 0)\n",
					*perr_tab_max);
			/* break; */
			return -1;
		}
		if (DEBUG)
			fprintf(STDOUT, "choose_next_initial_atom() :    *Renew ErrorTable (err_tab_max = %d)\n", 
				*perr_tab_max);
	} else {
		*perr_tab_ct = *perr_tab_ct + 1;
	}
	
	if (0) {
		fprintf(STDOUT, "choose_next_initial_atom() : ErrorTable[ ] (err_tab_ct= %d)\n",
			*perr_tab_ct);
		for (i = 0; i < my_settings.error_tab_size; i++) {
			fprintf(STDOUT, "choose_next_initial_atom() : ErrorTable[%d] (ind= %d, err= %f, flag= %d)\n",
				i, error_tab[i].ind, error_tab[i].err,
				my_protein.atom[error_tab[i].ind].
				can_be_init_atom);
		}
	}
	
	return error_tab[*perr_tab_ct].ind;
}

/* Calculate the average distance of N-CA, CA-C along a backbone */
float average_distance(int atm_head, int atm_tail)
{
	int seq_head, seq_tail;
	int seq_ct, atm_ct, base_atm_ct;
	int bond_ct;
	int atm_to_seq_flag;
	float ave_dis;
	float cur_dis;
	float wrk_vx, wrk_vy, wrk_vz;
	
	atm_to_seq_flag = convert_atom_range_into_seq_range(atm_head, atm_tail,
							&seq_head, &seq_tail);
	if(atm_to_seq_flag == FAIL) {
		fprintf(STDERR,
			"average_distance() : ERROR invalid atom range [%d,%d]\n",
			atm_head, atm_tail);
		exit(1);		
	}
	
	ave_dis = 0.0;
	cur_dis = 0.0;
	bond_ct = 0;
	for (seq_ct = seq_head; seq_ct <= seq_tail; seq_ct++) {
		base_atm_ct = -1;
		for (atm_ct = my_protein.sequence[seq_ct].atom_head;
		     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
		     atm_ct++) {
			if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
			if (base_atm_ct < 0) {
				base_atm_ct = atm_ct; /* Set the first atom */
				continue;
			}
			wrk_vx = my_protein.atom[atm_ct].x
				- my_protein.atom[base_atm_ct].x;
			wrk_vy = my_protein.atom[atm_ct].y
				- my_protein.atom[base_atm_ct].y;
			wrk_vz = my_protein.atom[atm_ct].z
				- my_protein.atom[base_atm_ct].z;
			cur_dis = sqrt(wrk_vx * wrk_vx + wrk_vy * wrk_vy
				       + wrk_vz * wrk_vz);
			distance_check(cur_dis, base_atm_ct, atm_ct);
			ave_dis = ave_dis + cur_dis;
			bond_ct++;
			base_atm_ct = atm_ct;
		}
	}
	if (bond_ct == 0) {
		fprintf(STDERR,
			"average_distance() : ERROR  no atoms in sequence [%d, %d]\n",
			seq_head, seq_tail);
		exit(1);
	}
	ave_dis = ave_dis / (float) bond_ct;
	return ave_dis;
}

/* Choose the initial atom */
int setup_initial_atom(void)
{
	int ind;
	
	for (ind = 1; ind <= my_protein.atm_num; ind++) {
		my_protein.atom[ind].can_be_init_atom = YES;
	}
	invalidate_terminal_and_isolated_atoms();
	ind = choose_initial_atom(1, my_protein.atm_num);
	my_protein.atom[ind].can_be_init_atom = NONO;
	return ind;
}

int setup_next_init_atom(struct pe_init_tile cur_tile,
			 struct pe_slt_tile * slt_tiles, 
			 struct pe_error_table * error_tab,
			 int * perr_tab_ct, int * perr_tab_max,
			 struct pe_init_tile * pbst_tile)
{
	int new_ind;
	int i;
	
	new_ind = choose_next_initial_atom(cur_tile, slt_tiles, error_tab,
				       perr_tab_ct, perr_tab_max, pbst_tile);
	if (new_ind < 0) return -1;
	my_protein.atom[new_ind].can_be_init_atom = NONO;
	if (0) print_init_atom(new_ind);
	
	for (i = 0; i <= my_protein.tile_max_num; i++)
		clear_slt_tile(slt_tiles + i);
	my_protein.tile_num = 0;
	return new_ind;
}

void encode_protein_of_ave_dis(float ave_dis, struct pe_slt_tile * slt_tiles, 
			       struct pe_error_table * error_tab,
			       struct pe_init_tile * pbst_tile)
{
	struct pe_init_tile cur_tile;
	int save_ind;
	int while_ct;
	int err_tab_ct, err_tab_max;
	
	cur_tile.ind = setup_initial_atom();		
	cur_tile.t_ind = setup_tables_and_init_tile(ave_dis, cur_tile.ind,
						    slt_tiles, XYZW);
	cur_tile.t_type = XYZW;
	cur_tile.ave_error = 0.0;
	pbst_tile->ind = cur_tile.ind;
	pbst_tile->t_ind = cur_tile.t_ind;
	pbst_tile->t_type = cur_tile.t_type;
	pbst_tile->ave_error = -1.0;
	while_ct = 0;
	
	if (DEBUG) {
		fprintf(STDOUT, 
			"encode_protein_of_ave_dis() : Go to %d-th while loop (ind= %d,t_ind= %d, best_ind= %d, best_t_ind= %d, best_t_type= %d)\n",
			while_ct, cur_tile.ind, cur_tile.t_ind, pbst_tile->ind,
			pbst_tile->t_ind, pbst_tile->t_type);
	}
	while (1) {
		if (chain_encoding(cur_tile.ind, &(cur_tile.t_ind), slt_tiles)
		    == FAIL){
			fprintf(STDOUT, 
				"encode_protein_of_ave_dis() : ERROR chain_encoding() failed [XYZW]\n");
			exit(1);
		}
		save_ind = setup_next_init_atom(cur_tile, slt_tiles,
						error_tab, &err_tab_ct,
						&err_tab_max, pbst_tile);
		if (my_settings.try_all_type == TRUE) {	/* 2007-1-12 */
			int i;
			int count;
			float wrk_error;
			
			cur_tile.t_ind = setup_tables_and_init_tile(ave_dis,
						cur_tile.ind, slt_tiles, YZWX);
			cur_tile.t_type = YZWX;
			cur_tile.ave_error = 0.0;
			if (chain_encoding(cur_tile.ind, &(cur_tile.t_ind),
			    slt_tiles) == FAIL){
				fprintf(STDOUT,
					"encode_protein_of_ave_dis() : ERROR chain_encoding() failed [YZWX]\n");
				exit(1);
			}
			wrk_error = 0.0;
			count = 0;
			for (i = 1; i <= cur_tile.t_ind; i++) {
				if (slt_tiles[i].atom_flag > 0) {
					wrk_error = wrk_error
						+ slt_tiles[i].error;
					count++;
				}
			}
			wrk_error = wrk_error / (float) count;
			if ( wrk_error < pbst_tile->ave_error ) {
				cur_tile.ind = setup_next_init_atom(cur_tile,
					slt_tiles, error_tab, &err_tab_ct,
					&err_tab_max, pbst_tile);
				if (DEBUG) {
					fprintf(STDOUT,
						"encode_protein_of_ave_dis() : [YZWX] error = %f  < [XYZW] error = %f (#atom = %d)\n",
						wrk_error, pbst_tile->ave_error,
						count);
					fprintf(STDOUT,
						"encode_protein_of_ave_dis() : Go to %d-th while loop (ind= %d,t_ind= %d, best_ind= %d, best_t_ind= %d, best_t_type= %d)\n",
						while_ct, cur_tile.ind,
						cur_tile.t_ind, pbst_tile->ind,
						pbst_tile->t_ind,
						pbst_tile->t_type);
				}
			} else {
				for (i = 0; i <= my_protein.tile_max_num; i++)
					clear_slt_tile(slt_tiles + i);
				my_protein.tile_num = 0;
				cur_tile.ind = save_ind;
			}
		} else {
			cur_tile.ind = save_ind;
		}
		
		if (cur_tile.ind < 0) break;
		while_ct++;
		if (0) {
			fprintf(STDOUT,
				"encode_protein_of_ave_dis() : Go to %d-th while loop (ind= %d,t_ind= %d, best_ind= %d, best_t_ind= %d, best_t_type= %d)\n",
				while_ct, cur_tile.ind, cur_tile.t_ind,
				pbst_tile->ind, pbst_tile->t_ind,
				pbst_tile->t_type);
		}
		if (err_tab_ct >= err_tab_max || while_ct > 20) break;
		if (my_settings.move_to_target == TRUE) break;
		cur_tile.t_ind = setup_tables_and_init_tile(ave_dis,
						cur_tile.ind, slt_tiles, XYZW);
		cur_tile.t_type = XYZW;
		cur_tile.ave_error = 0.0;
	}		/* while () { */

	if (0) {
		int seq_ct;
		
		seq_ct = search_sequence(pbst_tile->ind);
		fprintf(STDOUT,
"encode_protein_of_ave_dis() : Local best atom = the %d-th atom in the %d-th chain (Num of tiles= %d)\n",
			pbst_tile->ind, seq_ct, pbst_tile->t_ind);
	}
	if (0) print_untried_atoms();
	return;
}

void compute_best_ind_and_dis(struct pe_slt_tile * slt_tiles)
{
	struct pe_init_tile global_bst_tile;
	float ave_dis;
	struct pe_init_tile bst_tile;
	int for_ct;
	float for_head, for_tail, for_step;
	struct pe_error_table * error_tab;
	struct local_table my_local_table[MAX_LOCAL_AVE_TBL];
		
	global_bst_tile.ind = 0;
	global_bst_tile.t_ind = 0;
	global_bst_tile.t_type = XYZW;
	global_bst_tile.ave_error = -1.0;
	global_bst_tile.ave_dis = -1.0;
	error_tab = (struct pe_error_table *) malloc(my_settings.error_tab_size
					* sizeof(struct pe_error_table));
	for_head = my_settings.encoding_for_head;
	for_tail = my_settings.encoding_for_tail;
	for_step = my_settings.encoding_for_step;
	if (DEBUG) {
		fprintf(STDOUT,
"compute_best_ind_and_dis() : for (ave_dis = %8.3f; ave_dis <= %8.3f; ave_dis = ave_dis + %8.3f)\n",
			for_head, for_tail, for_step);
	}
	/* ave_dis: average distance between successive atoms */
	for (ave_dis = for_head, for_ct = 0; ave_dis <= for_tail;
	     ave_dis = ave_dis + for_step, for_ct++) {
		if (DEBUG) {
			print_loop_label(global_bst_tile, for_ct, ave_dis);
		}
		encode_protein_of_ave_dis(ave_dis, slt_tiles, error_tab,
					  &bst_tile);
		if (DEBUG) {
			fprintf(STDOUT,
"compute_best_ind_and_dis() :     best_ave_error= %f <=> global_best_ave_error= %f (ave_dis = %f)\n",
				bst_tile.ave_error, global_bst_tile.ave_error,
				ave_dis);
			if (for_ct > MAX_LOCAL_AVE_TBL) {
				fprintf(STDOUT,
"compute_best_ind_and_dis() :     for_ct = %d > MAX_LOCAL_AVE_TBL = %d\n",
					for_ct, MAX_LOCAL_AVE_TBL);
				exit(1);
			}
			my_local_table[for_ct].ave_error = bst_tile.ave_error;
			my_local_table[for_ct].tile_num = bst_tile.t_ind;
		}
		if ((global_bst_tile.ave_error > bst_tile.ave_error)
		    || (global_bst_tile.ave_error < 0.0)) {
			global_bst_tile.ind = bst_tile.ind;
			global_bst_tile.t_ind = bst_tile.t_ind;
			global_bst_tile.t_type = bst_tile.t_type;
			global_bst_tile.ave_error = bst_tile.ave_error;
			global_bst_tile.ave_dis = ave_dis;
		}
	}	/* for (ave_dis = for_head, for_ct = 0; ... ) { */

	if (DEBUG) {
		print_global_best(global_bst_tile);
		print_all_ave_diss(for_head, for_tail, for_step,
					   my_local_table);
	}
	my_protein.best_ind = global_bst_tile.ind;
	my_protein.best_t_type = global_bst_tile.t_type;
	my_protein.ave_dis = global_bst_tile.ave_dis;
	my_protein.ave_error = my_local_table[(int)
		((global_bst_tile.ave_dis - for_head) / for_step)].ave_error;
	return;
}

int seek_tile(struct pe_slt_tile * s_tiles, int atm_ct)
{
	int t_ct, seq_ct;
	
	seq_ct = search_sequence(atm_ct);
	for (t_ct = my_protein.sequence[seq_ct].tile_head;
	     t_ct <= my_protein.sequence[seq_ct].tile_tail;
	     t_ct++) {
		if (s_tiles[t_ct].atom_flag == 0)
			continue;
		if (atm_ct == s_tiles[t_ct].error_atom)
			return t_ct;
	}
	fprintf(STDERR, "seek_tile() : ERROR overrun atm_ct = %d, seq_ct = %d (from %d-th to %d-th tile)\n",
		atm_ct, seq_ct, my_protein.sequence[seq_ct].tile_head,
		my_protein.sequence[seq_ct].tile_tail);
	exit(1);
}

int check_atom_position(struct pe_slt_tile * s_tiles, int atm_ct)
{
	int t_ct;
	float wrk_cx, wrk_cy, wrk_cz, wrk_error;
	
	t_ct = seek_tile(s_tiles, atm_ct);
	wrk_cx = s_tiles[t_ct].cx - my_protein.atom[atm_ct].x;
	wrk_cy = s_tiles[t_ct].cy - my_protein.atom[atm_ct].y;
	wrk_cz = s_tiles[t_ct].cz - my_protein.atom[atm_ct].z;
	if (my_settings.move_to_target == TRUE) {
		if (FLOAT_NEQ(wrk_cx, my_coor.ave_dis)
		    || FLOAT_NEQ(wrk_cy, my_coor.ave_dis)
		    || FLOAT_NEQ(wrk_cz, my_coor.ave_dis)) {
			fprintf(STDERR,
				"check_atom_position() : Error slt_tiles[%d].c* = (%8.3f, %8.3f, %8.3f)\n",
				t_ct, s_tiles[t_ct].cx,
				s_tiles[t_ct].cy,
				s_tiles[t_ct].cz);
			fprintf(STDERR,
				"check_atom_position() :       != my_protein.atom[%d].* = (%8.3f, %8.3f, %8.3f)\n",
				atm_ct, my_protein.atom[atm_ct].x,
				my_protein.atom[atm_ct].y,
				my_protein.atom[atm_ct].z);
			return FAIL;
		}
	} else {
		wrk_error = sqrt(wrk_cx * wrk_cx + wrk_cy * wrk_cy
				 + wrk_cz * wrk_cz);
		if (wrk_error > 2.0 * my_coor.ave_dis) {
			fprintf(STDERR,
				"check_atom_position() : Error |slt_tiles[%d] - my_protein.atom[%d]| = %8.3f\n",
				t_ct, atm_ct, wrk_error);
			fprintf(STDERR,
				"check_atom_position() :       >  my_coor.ave_dis = %8.3f\n",
				my_coor.ave_dis);
			if (0) {
				fprintf(STDERR,
					"check_atom_position() :       where\n");
				fprintf(STDERR,
					"check_atom_position() :             slt_tiles[%d].c* = (%8.3f, %8.3f, %8.3f)\n",
					t_ct, s_tiles[t_ct].cx,
					s_tiles[t_ct].cy,
					s_tiles[t_ct].cz);
				fprintf(STDERR,
					"check_atom_position() :             my_protein.atom[%d].* = (%8.3f, %8.3f, %8.3f)\n",
					atm_ct, my_protein.atom[atm_ct].x,
					my_protein.atom[atm_ct].y,
					my_protein.atom[atm_ct].z);
			}
			/* [Sep202006]
			if (wrk_error > 3.0 * my_coor.ave_dis)
				return FAIL;
			 */
		}
	}
	return SUCCESS;
}

int check_tile_direction(struct pe_slt_tile * s_tiles, int t_ct)
{
	int ind, prv_ind, nxt_ind;
	float wrk_dx, wrk_dy, wrk_dz, wrk_length;
	float wrk_dx2, wrk_dy2, wrk_dz2, wrk_length2;
	
	ind = s_tiles[t_ct].error_atom;
	prv_ind = seek_previous_target_atom(ind);
	if (prv_ind < 0)
		prv_ind = ind;
	nxt_ind = seek_next_target_atom(ind);
	if (nxt_ind < 0)
		nxt_ind = ind;
	wrk_dx = my_protein.atom[nxt_ind].x
		- my_protein.atom[prv_ind].x;
	wrk_dy = my_protein.atom[nxt_ind].y
		- my_protein.atom[prv_ind].y;
	wrk_dz = my_protein.atom[nxt_ind].z
		- my_protein.atom[prv_ind].z;
	if (prv_ind == nxt_ind) {
		return SUCCESS;
	}
	if (FLOAT_NEQ(s_tiles[t_ct].dx - wrk_dx, my_coor.ave_dis)
		|| FLOAT_NEQ(s_tiles[t_ct].dy - wrk_dy, my_coor.ave_dis)
		|| FLOAT_NEQ(s_tiles[t_ct].dz - wrk_dz, my_coor.ave_dis)) {
		float wrk_length;
		
		wrk_length = wrk_dx * wrk_dx + wrk_dx * wrk_dx + wrk_dx * wrk_dx;
		if (FLOAT_NEQ(wrk_length, 1.0)) {	// 20071223  if (wrk_length != 0.0)
			fprintf(STDERR,
				"check_tile_direction() : Error slt_tiles[%d].d* = (%8.3f, %8.3f, %8.3f) != (%8.3f, %8.3f, %8.3f)\n",
				t_ct, s_tiles[t_ct].dx, s_tiles[t_ct].dy,
				s_tiles[t_ct].dz, wrk_dx, wrk_dy, wrk_dz);
			print_tile(t_ct, s_tiles);
			return FAIL;
		}
	}
	
	if (s_tiles[t_ct].code == C_UP) {
		wrk_dx = s_tiles[t_ct].tx - s_tiles[t_ct].bx;
		wrk_dy = s_tiles[t_ct].ty - s_tiles[t_ct].by;
		wrk_dz = s_tiles[t_ct].tz - s_tiles[t_ct].bz;
		wrk_length = sqrt(wrk_dx * wrk_dx + wrk_dy * wrk_dy
				  + wrk_dz * wrk_dz);
		wrk_dx = wrk_dx / wrk_length;
		wrk_dy = wrk_dy / wrk_length;
		wrk_dz = wrk_dz / wrk_length;
	} else {
		wrk_dx = s_tiles[t_ct].bx - s_tiles[t_ct].tx;
		wrk_dy = s_tiles[t_ct].by - s_tiles[t_ct].ty;
		wrk_dz = s_tiles[t_ct].bz - s_tiles[t_ct].tz;
		wrk_length = sqrt(wrk_dx * wrk_dx + wrk_dy * wrk_dy
				  + wrk_dz * wrk_dz);
		wrk_dx = wrk_dx / wrk_length;
		wrk_dy = wrk_dy / wrk_length;
		wrk_dz = wrk_dz / wrk_length;
	}
	wrk_dx2 = s_tiles[t_ct].dx;
	wrk_dy2 = s_tiles[t_ct].dy;
	wrk_dz2 = s_tiles[t_ct].dz;
	wrk_length2 = sqrt(wrk_dx2 * wrk_dx2 + wrk_dy2 * wrk_dy2
			  + wrk_dz2 * wrk_dz2);
	wrk_dx2 = wrk_dx2 / wrk_length2;
	wrk_dy2 = wrk_dy2 / wrk_length2;
	wrk_dz2 = wrk_dz2 / wrk_length2;
	if (FLOAT_NEQ(wrk_dx - wrk_dx2, my_coor.ave_dis)
	    || FLOAT_NEQ(wrk_dy - wrk_dy2, my_coor.ave_dis)
	    || FLOAT_NEQ(wrk_dz - wrk_dz2, my_coor.ave_dis)) {
		fprintf(STDERR,
			"check_tile_direction() : Error slt_tiles[%d].d* = (%8.3f, %8.3f, %8.3f)\n",
			t_ct, s_tiles[t_ct].dx, s_tiles[t_ct].dy,
			s_tiles[t_ct].dz);
		fprintf(STDERR,
			"check_tile_direction() :       <=> |slt_tiles[%d].t* - b*| = (%8.3f, %8.3f, %8.3f)\n",
			t_ct, wrk_dx, wrk_dy, wrk_dz);
		print_tile(t_ct, s_tiles);
		return FAIL;
	}
	return SUCCESS;
}

void check_tiles(struct pe_slt_tile * slt_tiles)
{
	int mod_ct, seq_ct, atm_ct, t_ct;
	
	if (0) print_my_protein(&my_protein);
	/* atoms */
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (atm_ct == 0) break;	/* no residue */
				if (THIS_IS_NOT_CA_ATOM(atm_ct)) continue;
				if (check_atom_position(slt_tiles, atm_ct)
				    == FAIL)
					exit(1);
			}
		}
		break;	/* Consider only one model per protein */
	}
	/* 2006-11-04 add if (0) */
	if (0) fprintf(STDOUT, "check_tiles() : DEBUG check_atom_position() done\n");
	
	/* tiles */
	if (my_settings.rotate_tile != TRUE)
		return; 
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			for (t_ct = my_protein.sequence[seq_ct].tile_head;
			     t_ct <= my_protein.sequence[seq_ct].tile_tail;
			     t_ct++) {
				if (slt_tiles[t_ct].atom_flag == 0) continue;
				if (check_tile_direction(slt_tiles, t_ct)
				    == FAIL)
					exit(1);
			}
		}
		break;	/* Consider only one model per protein */
	}
	/* 2006-11-04 add if (0) */
	if (0) fprintf(STDOUT, "check_tiles() : DEBUG check_tile_direction() done\n");
	return;
}

/* Set the initi tile to XYZW and YZWX and choose the better one 
*/
int compute_best_init_type(int atm_ind, float ave_dis,
			   struct pe_slt_tile * slt_tiles)
{	
	float save_error;
	
	if (my_settings.try_all_type != TRUE) {	/* 2007-1-12 */
		if (0) fprintf(STDOUT, "compute_best_init_type() : DEBUG init_type is always XYZW\n");
		return XYZW;
	}
	my_protein.best_t_type = XYZW;
	encoding_suite(my_protein.best_ind, my_protein.ave_dis, slt_tiles);
	save_error = my_protein.ave_error;
	
	my_protein.best_t_type = YZWX;
	encoding_suite(my_protein.best_ind, my_protein.ave_dis, slt_tiles);
	
	if (save_error == my_protein.ave_error) {
		if (0) fprintf(STDOUT, "compute_best_init_type() : DEBUG error_XYZW = %f <=> error_YZWX =%f)\n",
		       save_error, my_protein.ave_error);
		if (0) fprintf(STDOUT, "compute_best_init_type() : DEBUG init_type = XYZW (YZWX)\n");
		return XYZW;		
	} else if (save_error < my_protein.ave_error) {
		if (0) fprintf(STDOUT, "compute_best_init_type() : DEBUG init_type = XYZW\n");
		return XYZW;		
	} else {
		if (1) fprintf(STDOUT, "compute_best_init_type() : DEBUG init_type = YZWX\n");
		return YZWX;
	}		
}

void encode_protein_structure(struct pe_slt_tile * slt_tiles)
{
	my_protein.ave_bond_length = average_distance(1, my_protein.atm_num);
	if (DEBUG) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT,
"encode_protein_structure() : Average distance = %f   (atom_num=%d)\n",
			my_protein.ave_bond_length, my_protein.atm_num);
	}
	if (0) print_my_protein(&my_protein);
	if (0) print_my_sequences();
	
	if (my_settings.initial_atom == BESTPOINT)
		compute_best_ind_and_dis(slt_tiles);
	else if (my_settings.initial_atom == MIDPOINT) {
		int res_num;
		int new_atm;
		
		if (my_settings.encode_RNA != TRUE)	// 2007-08-22
			res_num = (my_protein.atm_num/2) / 3;	/* 2006-11-03 */
		else
			res_num = (my_protein.atm_num/2) / 2;
		/*
		if (3 * (2 * res_num + 1) != my_protein.atm_num) {
			fprintf(STDOUT,
				"encode_protein_structure() : initial atom not midpoint (atm_num = %d)\n",
				my_protein.atm_num);
		}
		 if (my_protein.atom[my_protein.best_ind].atom_name != PE_CA) {
			 fprintf(STDOUT,
				 "encode_protein_structure() : ERROR initial atom = %d-th",
				 my_protein.best_ind);
			 if (my_protein.atom[my_protein.best_ind].atom_name
			     == PE_CC)
				 fprintf(STDOUT, " (C )");
			 
			 else if (my_protein.atom[my_protein.best_ind].atom_name
				  == PE_CA)
				 fprintf(STDOUT, " (CA)");
			 
			 else if (my_protein.atom[my_protein.best_ind].atom_name
				  == PE_NN)
				 fprintf(STDOUT, " (N )");
			 
			 else
				 fprintf(STDOUT, " (UN)");
			 fprintf(STDOUT, " is not CA atom\n");
			 exit(1);
		 }
		 */
		if (my_settings.encode_RNA != TRUE)	// 2007-08-22
			my_protein.best_ind = 3 * res_num + 2;
		else
			my_protein.best_ind = 2 * res_num + 1;
		if ((my_settings.ca_only_flag == TRUE) && 
		    (my_protein.atom[my_protein.best_ind].atom_name != PE_CA) ) {
			new_atm = seek_previous_target_atom(my_protein.best_ind);
			if (new_atm < 0)
				new_atm = seek_next_target_atom(my_protein.best_ind);
			if (new_atm < 0) {
				fprintf(STDOUT,
					"encode_protein_structure() : ERROR No CA atom\n");
				exit(1);
			}
			my_protein.best_ind = new_atm;
		}

		{	// 2007-08-22
			int save_ind;
			
			new_atm = seek_previous_target_atom(my_protein.best_ind);
			if (new_atm < 0) {
				new_atm = seek_next_target_atom(my_protein.best_ind);
				if (new_atm < 0) {
					fprintf(STDOUT,
						"encode_protein_structure() : ERROR initial atom assignment failed (%d-th atom)\n",
						my_protein.best_ind);
					exit(1);
				}
				my_protein.best_ind = new_atm;
			}
			save_ind = my_protein.best_ind;
			new_atm = seek_next_target_atom(my_protein.best_ind);
			if (new_atm < 0) {
				new_atm = seek_previous_target_atom(my_protein.best_ind);
				if (new_atm < 0 || new_atm == save_ind) {
					fprintf(STDOUT,
						"encode_protein_structure() : ERROR initial atom assignment failed (%d-th atom)\n",
						my_protein.best_ind);
					exit(1);
				}
				my_protein.best_ind = new_atm;
			}
		}
		
		if (DEBUG)
			fprintf(STDOUT,
				"encode_protein_structure() : DEBUG initial atom = %d-th [MIDPOINT]\n",
				my_protein.best_ind);
		my_protein.ave_dis = my_settings.encoding_for_tail; /* [TODO] */
		my_protein.best_t_type = compute_best_init_type(
					my_protein.best_ind, my_protein.ave_dis,
					slt_tiles);
	} else {
		my_protein.best_ind = my_settings.initial_atom;
		if ((my_settings.ca_only_flag == TRUE) && 
		    (my_protein.atom[my_protein.best_ind].atom_name != PE_CA) ) {
			int new_atm;
			
			if (DEBUG)
				fprintf(STDOUT,
				"encode_protein_structure() : DEBUG the specified initial atom is not CA atom \n");
			new_atm = seek_previous_target_atom(my_protein.best_ind);
			if (new_atm < 0)
				new_atm = seek_next_target_atom(my_protein.best_ind);
			if (new_atm < 0) {
				fprintf(STDOUT,
					"encode_protein_structure() : ERROR No CA atom\n");
				exit(1);
			}
			my_protein.best_ind = new_atm;
		}
		if (DEBUG)
			fprintf(STDOUT,
				"encode_protein_structure() : DEBUG initial atom = %d-th [SPECIFIED]\n",
				my_protein.best_ind);
		my_protein.ave_dis = my_settings.encoding_for_tail; /* [TODO] */
		my_protein.best_t_type = compute_best_init_type(
					my_protein.best_ind, my_protein.ave_dis,
					slt_tiles);
	}
	encoding_suite(my_protein.best_ind, my_protein.ave_dis, slt_tiles);
	if (0) print_my_coordinates();
	if (0) print_trn_table();
	if (0) print_my_protein(&my_protein);
	if (0) print_my_slt_tiles(slt_tiles);
	
	if (DEBUG) check_tiles(slt_tiles);
	return;
}
