/*
 *  pe_encoding_suite.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/10/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-01-12 add if (DEBUG) print_my_sequences(); in 
               compute_direction_of_N_atom() and compute_direction_of_C_atom()
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 *  2007-08-22 modifiy error message in compute_direction_of_N_atom()
 *  2008-08-22 version 0.5 release
 *  2007-12-23 Check the denominator and chenge the return value in compute_V*_vector()
 *  2007-12-23 Recompute if V* is zero vector in compute_3d_coordinates()
 *  2007-12-23 Modify compute_VB_vector() to make sure that VA _|_ VB
 *  2008-01-02 version 0.6 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern void print_trn_table(void);
extern void setup_transition_table(float ave_dis);
extern int set_initial_tile(struct pe_slt_tile * slt_tiles, int ind, int tile_type);
extern int chain_encoding(int ind, int * ptile_ind,
			  struct pe_slt_tile * slt_tiles);
extern void clear_slt_tile(struct pe_slt_tile * pt_dst_tile);

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;
extern struct pe_coor_system my_coor;
extern struct pe_trn_table my_trn_table[];

struct pe_path_of_length_n ** npath_table;
struct pe_path_of_length_n * np_table[NP_TBL_SIZE][24];

/* U/D sequence */
/* p_ct =  code[0]*2**(n-1) + code[1]*2**(n-2) + ... + code[n-2]*2 + code[n-1]
*/
void print_npath_table(struct pe_path_of_length_n ** npath_tbl, int path_length)
{
	int i, p_ct, t_ct;
	int max_path;
	
	max_path = (int) pow(2.0, (double) path_length);
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"%d-tile transition table: Dump list starts (max_path= %d)\n",
		path_length, max_path);
	if (npath_tbl[0] == NULL) {
		fprintf(STDOUT,
			"N-tile transition table: Memory not assigned.\n");
		fprintf(STDOUT, "\n");
		return;
	}
	for (i = 0; i < 24; i++) {
		fprintf(STDOUT,
"npath_table[%2d]:---------------------------------------------------------------\n",
			i);
		for (p_ct = 0; p_ct < max_path; p_ct++) {
			/* U/D sequence */
			fprintf(STDOUT,
				"npath_table[%2d]: code[0]->code[%d] = ",
				i, path_length);
			for (t_ct = 0; t_ct < path_length - 1; t_ct++) {
				if (npath_tbl[i][p_ct].code[t_ct] == C_DOWN)
					fprintf(STDOUT, "D-");
				else
					fprintf(STDOUT, "U-");
			}
			if (npath_tbl[i][p_ct].code[path_length - 1] == C_DOWN)
				fprintf(STDOUT, "D\n");
			else
				fprintf(STDOUT, "U\n");
			fprintf(STDOUT,
				"npath_table[%2d]: direction (%8.3f, %8.3f, %8.3f), length= %8.3f\n",
				i,
				npath_tbl[i][p_ct].c.x - my_trn_table[i].cx,
				npath_tbl[i][p_ct].c.y - my_trn_table[i].cy,
				npath_tbl[i][p_ct].c.z - my_trn_table[i].cz,
				npath_tbl[i][p_ct].c.length);
			
			fprintf(STDOUT,
				"npath_table[%2d]:           from (%8.3f, %8.3f, %8.3f)",
				i, my_trn_table[i].cx, my_trn_table[i].cy,
				my_trn_table[i].cz);
			fprintf(STDOUT, " to (%8.3f, %8.3f, %8.3f)\n",
				npath_tbl[i][p_ct].c.x,
				npath_tbl[i][p_ct].c.y,
				npath_tbl[i][p_ct].c.z);
		}
		fprintf(STDOUT, "\n");
		fflush(STDOUT);
	}
	fprintf(STDOUT, "N-tile transition table: Dump list ends.\n");
	fprintf(STDOUT, "\n");
	return;
}

void print_npath_table_backbone_vector(void)
{
	int i, p_ct, t_ct;
	int path_length;
	int max_path;
	
	path_length = my_settings.path_length;
	max_path = my_settings.max_path;
	
	fprintf(STDOUT, "[ntr_table (backbone vectors)]\n");
	for (i = 0; i < 24; i++) {
		for (p_ct = 0; p_ct < max_path; p_ct++) {
			for (t_ct = 0; t_ct < path_length; t_ct++) {
				fprintf(STDOUT, "npath_table[%2d][%d].bb_c[%d]= (%8.3f, %8.3f, %8.3f), length= %8.3f\n",
					i, p_ct, t_ct,
					npath_table[i][p_ct].bb_c[t_ct].x,
					npath_table[i][p_ct].bb_c[t_ct].y,
					npath_table[i][p_ct].bb_c[t_ct].z,
					npath_table[i][p_ct].bb_c[t_ct].length);
			}
		}
	}
	fprintf(STDOUT, "\n");
}

/* Set tile_head and tile_tail for each chain */
void set_tile_heads_and_tails(struct pe_slt_tile * slt_tiles)
{
	int t_ind;
	int mod_ct, seq_ct;
	
	t_ind = 1;
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			if (my_protein.sequence[seq_ct].atom_head <= 0) {
				/* RNA or DNA */
				my_protein.sequence[seq_ct].tile_head = 0;
				my_protein.sequence[seq_ct].tile_tail = -1;
			} else {
				my_protein.sequence[seq_ct].tile_head = t_ind;
				while ((t_ind < my_protein.tile_num)
				       && (slt_tiles[t_ind].ter == 0))
					t_ind++;
				if (t_ind > my_protein.tile_num) {
					fprintf(STDERR,
						"set_heads_and_tails() : my_protein.sequence[%d].tile_tail overflow \n",
						seq_ct);
					exit(1);
				}
				my_protein.sequence[seq_ct].tile_tail = t_ind;
				t_ind++;
			}
		}
	}
	return;
}

/* Error calculation */
/* slt_tiles[nxt_ind].error */
/* slt_tiles[nxt_ind].error_atom */
/* cur_error = |target_atom - current_tile| */
void error_calculation(struct pe_slt_tile * slt_tiles)
{
	int i, ca_atm_ct;
	int ind;
	float error_x, error_y, error_z;
	float wrk_ave_error;
	
	ca_atm_ct = 0;
	wrk_ave_error = 0.0;
	for (i = 1; i <= my_protein.tile_num; i++) {
		ind = slt_tiles[i].error_atom;
		if (slt_tiles[i].error > 0.0) {
			error_x = my_protein.atom[ind].x - slt_tiles[i].cx;
			error_y = my_protein.atom[ind].y - slt_tiles[i].cy;
			error_z = my_protein.atom[ind].z - slt_tiles[i].cz;
			slt_tiles[i].error = sqrt(error_x * error_x
				+ error_y * error_y + error_z * error_z);
		}
		if (my_protein.atom[ind].atom_name == PE_CA) {
			ca_atm_ct++;
			wrk_ave_error = wrk_ave_error + slt_tiles[i].error;
		}
	}
	if (ca_atm_ct > 0)
		my_protein.ave_error = wrk_ave_error / (float) ca_atm_ct;
	else
		my_protein.ave_error = 0.0;
	return;
}

void compute_average_of_errors(struct pe_slt_tile * slt_tiles)
{
	int i, ca_atm_ct;
	int ind;
	float wrk_ave_error;
	
	ca_atm_ct = 0;
	wrk_ave_error = 0.0;
	for (i = 1; i <= my_protein.tile_num; i++) {
		ind = slt_tiles[i].error_atom;
		if (my_protein.atom[ind].atom_name == PE_CA) {
			ca_atm_ct++;
			wrk_ave_error = wrk_ave_error + slt_tiles[i].error;
		}
	}
	if (ca_atm_ct > 0)
		my_protein.ave_error = wrk_ave_error / (float) ca_atm_ct;
	else
		my_protein.ave_error = 0.0;
	return;
}

/* Error check */
/* slt_tiles[nxt_ind].error */
/* slt_tiles[nxt_ind].error_atom */
/* cur_error = |target_atom - current_tile| */
int check_error(struct pe_slt_tile * slt_tiles)
{
	int i;
	int ind;
	float error_x, error_y, error_z, wrk_error;
	
	for (i = 1; i <= my_protein.tile_num; i++) {
		if (slt_tiles[i].error > 0.0) {
			ind = slt_tiles[i].error_atom;
			error_x = my_protein.atom[ind].x - slt_tiles[i].cx;
			error_y = my_protein.atom[ind].y - slt_tiles[i].cy;
			error_z = my_protein.atom[ind].z - slt_tiles[i].cz;
			wrk_error = sqrt(error_x * error_x + error_y * error_y
					 + error_z * error_z);
			if (FLOAT_NEQ(slt_tiles[i].error - wrk_error,
					my_coor.ave_dis)) {
				fprintf(STDERR,
"check_error() : slt_tiles[%d].error= %f != actual error= %f (against = %d-th atom",
					i, slt_tiles[i].error, wrk_error, ind);
				if (my_protein.atom[ind].atom_name == PE_CC)
					fprintf(STDERR, " (C ) )\n");
				else if (my_protein.atom[ind].atom_name ==PE_CA)
					fprintf(STDERR, " (CA) )\n");
				else if (my_protein.atom[ind].atom_name ==PE_NN)
					fprintf(STDERR, " (N ) )\n");
				else
					fprintf(STDERR, " (UN) )\n");
				return FAIL;
			}
		}
	}
	if (0) fprintf(STDOUT,
"check_error() : DEBUG error check done (tile_num= %d).\n",
		my_protein.tile_num);
	return SUCCESS;
}

/* The position of (t_ind+1)-th tile */
int check_next_tile(int t_ind, struct pe_slt_tile * slt_tiles,
		    int * ptop_0, int * ptop_1, int * ptop_2, int * ptop_3,
		    float * pwrk_x, float * pwrk_y, float * pwrk_z)
{
	int cur_type, wrk_type;
	
	cur_type = slt_tiles[t_ind].type;
	if (slt_tiles[t_ind].code == C_UP) {
		*pwrk_x = my_trn_table[cur_type].upward_cx
			- my_trn_table[cur_type].cx + *pwrk_x;
		*pwrk_y = my_trn_table[cur_type].upward_cy
			- my_trn_table[cur_type].cy + *pwrk_y;
		*pwrk_z = my_trn_table[cur_type].upward_cz
			- my_trn_table[cur_type].cz + *pwrk_z;
		if (slt_tiles[t_ind + 1].code == C_UP) {
			wrk_type = my_trn_table[cur_type].upward_U_type;
			*ptop_0 = my_trn_table[cur_type].upward_U_top[0]
				+ *ptop_0;
			*ptop_1 = my_trn_table[cur_type].upward_U_top[1]
				+ *ptop_1;
			*ptop_2 = my_trn_table[cur_type].upward_U_top[2]
				+ *ptop_2;
			*ptop_3 = my_trn_table[cur_type].upward_U_top[3]
				+ *ptop_3;
		} else if (slt_tiles[t_ind + 1].code == C_DOWN) {
			wrk_type = my_trn_table[cur_type].upward_D_type;
			*ptop_0 = my_trn_table[cur_type].upward_D_top[0]
				+ *ptop_0;
			*ptop_1 = my_trn_table[cur_type].upward_D_top[1]
				+ *ptop_1;
			*ptop_2 = my_trn_table[cur_type].upward_D_top[2]
				+ *ptop_2;
			*ptop_3 = my_trn_table[cur_type].upward_D_top[3]
				+ *ptop_3;
		} else {
			fprintf(STDERR,
				"check_next_tile() : Error slt_tiles[%d].code = UNKWON\n",
				t_ind + 1);
			return FAIL;
		}
	} else if (slt_tiles[t_ind].code == C_DOWN) {
		*pwrk_x = my_trn_table[cur_type].downward_cx
			- my_trn_table[cur_type].cx + *pwrk_x;
		*pwrk_y = my_trn_table[cur_type].downward_cy
			- my_trn_table[cur_type].cy + *pwrk_y;
		*pwrk_z = my_trn_table[cur_type].downward_cz
			- my_trn_table[cur_type].cz + *pwrk_z;
		if (slt_tiles[t_ind + 1].code == C_UP) {
			wrk_type = my_trn_table[cur_type].downward_U_type;
			*ptop_0 = my_trn_table[cur_type].downward_U_top[0]
				+ *ptop_0;
			*ptop_1 = my_trn_table[cur_type].downward_U_top[1]
				+ *ptop_1;
			*ptop_2 = my_trn_table[cur_type].downward_U_top[2]
				+ *ptop_2;
			*ptop_3 = my_trn_table[cur_type].downward_U_top[3]
				+ *ptop_3;
		} else if (slt_tiles[t_ind + 1].code == C_DOWN) {
			wrk_type = my_trn_table[cur_type].downward_D_type;
			*ptop_0 = my_trn_table[cur_type].downward_D_top[0]
				+ *ptop_0;
			*ptop_1 = my_trn_table[cur_type].downward_D_top[1]
				+ *ptop_1;
			*ptop_2 = my_trn_table[cur_type].downward_D_top[2]
				+ *ptop_2;
			*ptop_3 = my_trn_table[cur_type].downward_D_top[3]
				+ *ptop_3;
		} else {
			fprintf(STDERR,
				"check_next_tile() : Error slt_tiles[%d].code = UNKWON\n",
				t_ind + 1);
			return FAIL;
		}
	} else {
		fprintf(STDERR,
			"check_next_tile() : Error slt_tiles[%d].code = UNKOWN\n",
			t_ind);
		return FAIL;
	}

	/* Slant tile Check */
	if ((slt_tiles[t_ind + 1].top[0] != *ptop_0)
	    || (slt_tiles[t_ind + 1].top[1] != *ptop_1)
	    || (slt_tiles[t_ind + 1].top[2] != *ptop_2)
	    || (slt_tiles[t_ind + 1].top[3] != *ptop_3)
	    || (slt_tiles[t_ind + 1].type != wrk_type)) {
		fprintf(STDERR,
"check_next_tile() : Error slt_tiles[%d].top = (%d, %d, %d, %d) <=> (%d, %d, %d, %d)\n",
			t_ind + 1, slt_tiles[t_ind + 1].top[0],
			slt_tiles[t_ind + 1].top[1],slt_tiles[t_ind + 1].top[2],
			slt_tiles[t_ind + 1].top[3], *ptop_0, *ptop_1, *ptop_2,
			*ptop_3);
		fprintf(STDERR,
"check_next_tile() : Error slt_tiles[%d].type = %d <=> %d\n",
			t_ind + 1, slt_tiles[t_ind + 1].type, wrk_type);
		return FAIL;
	}
	return SUCCESS;
}

int check_center(int t_ind, struct pe_slt_tile * slt_tiles, float * pwrk_x,
		 float * pwrk_y, float * pwrk_z)
{
	/* Flat tile Check */
	if (FLOAT_NEQ((slt_tiles[t_ind].cx - *pwrk_x), my_coor.ave_dis)
	    || FLOAT_NEQ((slt_tiles[t_ind].cy - *pwrk_y), my_coor.ave_dis)
	    || FLOAT_NEQ((slt_tiles[t_ind].cz - *pwrk_z), my_coor.ave_dis)) {
		fprintf(STDERR,
"check_center() : Error slt_tiles[%d].c* = (%8.3f, %8.3f, %8.3f) != (%8.3f, %8.3f, %8.3f)\n",
			t_ind, slt_tiles[t_ind].cx,
			slt_tiles[t_ind].cy, slt_tiles[t_ind].cz,
			*pwrk_x, *pwrk_y, *pwrk_z);
		return FAIL;
	}
	return SUCCESS;
}

/* Tile position check */
int check_tile_position(struct pe_slt_tile * slt_tiles)
{
	int t_ind;
	int mod_ct, seq_ct;
	int top_0, top_1, top_2, top_3;
	float wrk_x, wrk_y, wrk_z;
	
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			t_ind = my_protein.sequence[seq_ct].tile_head;
			wrk_x = slt_tiles[t_ind].cx;
			wrk_y = slt_tiles[t_ind].cy;
			wrk_z = slt_tiles[t_ind].cz;
			top_0 = slt_tiles[t_ind].top[0];
			top_1 = slt_tiles[t_ind].top[1];
			top_2 = slt_tiles[t_ind].top[2];
			top_3 = slt_tiles[t_ind].top[3];
			for (; t_ind < my_protein.sequence[seq_ct].tile_tail;
			     t_ind++) {
				if (check_next_tile(t_ind, slt_tiles, &top_0,
				    &top_1, &top_2, &top_3, &wrk_x, &wrk_y, 
				    &wrk_z) == FAIL)
					return FAIL;
			}
			if (check_center(t_ind, slt_tiles, &wrk_x, &wrk_y,
					 &wrk_z) == FAIL) {
				return FAIL;
			}
		}	/* seq_ct */
	}		/* mod_ct */
	return SUCCESS;
}

/* Top vertex Check */
int check_top_vertex(struct pe_slt_tile * slt_tiles)
{
	int mod_ct, seq_ct, t_ind;
	float top_x, top_y, top_z, top_length;
	
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			for (t_ind = my_protein.sequence[seq_ct].tile_head;
			     t_ind < my_protein.sequence[seq_ct].tile_tail;
			     t_ind++) {
				top_x = my_coor.origin.x
				  + slt_tiles[t_ind].top[0] * my_coor.axes[0].x
				  + slt_tiles[t_ind].top[1] * my_coor.axes[1].x
				  + slt_tiles[t_ind].top[2] * my_coor.axes[2].x
				  + slt_tiles[t_ind].top[3] * my_coor.axes[3].x;
				top_y = my_coor.origin.y
				  + slt_tiles[t_ind].top[0] * my_coor.axes[0].y
				  + slt_tiles[t_ind].top[1] * my_coor.axes[1].y
				  + slt_tiles[t_ind].top[2] * my_coor.axes[2].y
				  + slt_tiles[t_ind].top[3] * my_coor.axes[3].y;
				top_z = my_coor.origin.z
				  + slt_tiles[t_ind].top[0] * my_coor.axes[0].z
				  + slt_tiles[t_ind].top[1] * my_coor.axes[1].z
				  + slt_tiles[t_ind].top[2] * my_coor.axes[2].z
				  + slt_tiles[t_ind].top[3] * my_coor.axes[3].z;
				top_length = sqrt(top_x * top_x + top_y * top_y
						  + top_z * top_z);
				
				if (FLOAT_NEQ(slt_tiles[t_ind].tx - top_x,
					      top_length * my_coor.ave_dis)
				    || FLOAT_NEQ(slt_tiles[t_ind].ty - top_y,
						 top_length * my_coor.ave_dis)
				    || FLOAT_NEQ(slt_tiles[t_ind].tz - top_z,
						 top_length * my_coor.ave_dis)){
					fprintf(STDERR,
"check_top_vertex() : Error slt_tiles[%d].t* = (%8.3f, %8.3f, %8.3f) != (%8.3f, %8.3f, %8.3f)\n",
						t_ind, slt_tiles[t_ind].tx,
						slt_tiles[t_ind].ty,
						slt_tiles[t_ind].tz, top_x, 
						top_y, top_z);
					return FAIL;
				}
			}	/* t_ind */
		}	/* seq_ct */
	}	/* mod_ct */
	return SUCCESS;
}

int npath_code_conversion(struct pe_path_of_length_n ** npath_tbl,
			  int path_length, int tile_type, int path_ct)
{
	int code, tmp;
	int t_ct;
	
	code = path_ct;
	for (t_ct = path_length - 1; t_ct >= 0; t_ct--) {
		tmp = code % 2;
		if (tmp == 0)
			npath_tbl[tile_type][path_ct].code[t_ct]= C_DOWN;
		else
			npath_tbl[tile_type][path_ct].code[t_ct] = C_UP;
		code = code / 2;
	}
	if (DEBUG) {
		if (code != 0) {
			fprintf(STDERR,
				"npath_code_conversioq(): Error conversion failed (code= %d)\n",
				code);
			return FAIL;
		}
	}
	return SUCCESS;
}

int alloc_npath_code_seq(struct pe_path_of_length_n ** npath_tbl,
			 int path_length, int tile_type)
{
	int p_ct;
	int max_path;
	
	max_path = (int) pow(2.0, (double) path_length);
	for (p_ct = 0; p_ct < max_path; p_ct++) {
		npath_tbl[tile_type][p_ct].code
		= (PE_CODE *) malloc(path_length * sizeof(PE_CODE));
		if (npath_tbl[tile_type][p_ct].code == NULL) {
			fprintf(STDERR, 
				"alloc_npath_code_seq() : ERROR npath_table[%d][%d].code not allocated\n",
				tile_type, p_ct);
			return FAIL;
		}
		if (my_settings.along_backbone == TRUE) {
			npath_tbl[tile_type][p_ct].bb_c
			= (struct pe_vector *)
			malloc(path_length *sizeof(struct pe_vector));
			if (npath_tbl[tile_type][p_ct].bb_c == NULL) {
				fprintf(STDERR, 
					"alloc_npath_code_seq() : ERROR npath_table[%d][%d].bb_c not allocated\n",
					tile_type, p_ct);
				return FAIL;
			}
		}
		if (npath_code_conversion(npath_tbl, path_length, tile_type,
					  p_ct) == FAIL)
			return FAIL;
	}
	return SUCCESS;
}

int search_sequence(int ind /* atom No. */ )
{
	int mod_ct, seq_ct;
	
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			if ((my_protein.sequence[seq_ct].atom_head <= ind)
			    && (ind <= my_protein.sequence[seq_ct].atom_tail)) {
				return seq_ct;
			}
		}
	}
	fprintf(STDERR, "search_sequence() : The %d-th atom does not exists.\n",
		ind);
	return -1;
}

int seek_previous_target_atom(int ind)
{
	int seq_ct;
	int prv_ind;
	
	seq_ct = search_sequence(ind);
	for (prv_ind = ind - 1;
	     prv_ind >= my_protein.sequence[seq_ct].atom_head; prv_ind--) {
		if (THIS_IS_NOT_CA_ATOM(prv_ind)) continue;
		return prv_ind;
	}
	return -1;
}

int seek_next_target_atom(int ind)
{
	int seq_ct;
	int nxt_ind;
	
	if (ind < 0) return -1;
	if (ind > 0)
		seq_ct = search_sequence(ind);
	else
		seq_ct = 1;
	for (nxt_ind = ind + 1;
	     nxt_ind <= my_protein.sequence[seq_ct].atom_tail; nxt_ind++) {
		if (THIS_IS_NOT_CA_ATOM(nxt_ind)) continue;
		return nxt_ind;
	}
	return -1;
}

/* 
   VA // N->C
   |VA| = sqrt(6)*ave_dis/t_fac
    (t_fac = tiles_per_res/3 = tiles per atom)
 */
int compute_VA_vector(int ind_atm, float ave_dis, 
		       double N_vx, double N_vy, double N_vz,
		       double C_vx, double C_vy, double C_vz,
		       double * pVA_vx, double * pVA_vy, double * pVA_vz)
{
	double VA_vx, VA_vy, VA_vz;
	double wrk_length;
	
	VA_vx = C_vx - N_vx;
	VA_vy = C_vy - N_vy;
	VA_vz = C_vz - N_vz;
	wrk_length = sqrt(VA_vx * VA_vx + VA_vy * VA_vy + VA_vz * VA_vz)
		* my_settings.tiles_per_res / 3.0;
	{	// 20071223
		if (!FLOAT_NEQ(wrk_length, 1.0)) {	/* if wrk_length == 0.0 */
			*pVA_vx = (sqrt(6.0) * ((double) ave_dis) * VA_vx);
			*pVA_vy = (sqrt(6.0) * ((double) ave_dis) * VA_vy);
			*pVA_vz = (sqrt(6.0) * ((double) ave_dis) * VA_vz);
			return FAIL;
		}
	}
	*pVA_vx = (sqrt(6.0) * ((double) ave_dis) * VA_vx) / wrk_length;
	*pVA_vy = (sqrt(6.0) * ((double) ave_dis) * VA_vy) / wrk_length;
	*pVA_vz = (sqrt(6.0) * ((double) ave_dis) * VA_vz) / wrk_length;
	return SUCCESS;		// 20071223  change the return value
}

/* 
  VB // (CA->N)+(CA->C)            
 |VB| = ave_dis/t_fac
 */
int compute_VB_vector(int ind_atm, float ave_dis,
		       double N_vx, double N_vy, double N_vz,
		       double C_vx, double C_vy, double C_vz,
		       double VA_vx, double VA_vy, double VA_vz,	// 20071223
		       double * pVB_vx, double * pVB_vy, double * pVB_vz)
{
	double VB_vx, VB_vy, VB_vz;
	double wrk_length;
	
	VB_vx = C_vx + N_vx;
	VB_vy = C_vy + N_vy;
	VB_vz = C_vz + N_vz;
	wrk_length = sqrt(VB_vx * VB_vx + VB_vy * VB_vy + VB_vz * VB_vz)
		* my_settings.tiles_per_res / 3.0;
	{	// 20071223
		if (!FLOAT_NEQ(wrk_length, 1.0)) {	/* if wrk_length == 0.0 */
			*pVB_vx = (((double) ave_dis) * VB_vx);
			*pVB_vy = (((double) ave_dis) * VB_vy);
			*pVB_vz = (((double) ave_dis) * VB_vz);
			return FAIL;
		}
	}

	if (0) {	// 2007-12-23 Adjust VB to make VB _|_ VA
		*pVB_vx = (((double) ave_dis) * VB_vx) / wrk_length;	
		*pVB_vy = (((double) ave_dis) * VB_vy) / wrk_length;
		*pVB_vz = (((double) ave_dis) * VB_vz) / wrk_length;
	} else {
		double tmp_vx, tmp_vy, tmp_vz;
		double VBB_vx, VBB_vy, VBB_vz;
	
		tmp_vx = (((double) ave_dis) * VB_vx) / wrk_length;
		tmp_vy = (((double) ave_dis) * VB_vy) / wrk_length;
		tmp_vz = (((double) ave_dis) * VB_vz) / wrk_length;
	
		{	// Compute tmp2 = (<VA,tmp>/<VA,VA>)VA and
			//         VBB = tmp - tmp2   (Note that VA _|_ VB)
			double tmp2_vx, tmp2_vy, tmp2_vz;
			double inner_product;
			double VA_length;
		
			inner_product = tmp_vx * VA_vx + tmp_vy * VA_vy + tmp_vz * VA_vz;
			VA_length = VA_vx * VA_vx + VA_vy * VA_vy + VA_vz * VA_vz;
			tmp2_vx = inner_product * VA_vx / VA_length;
			tmp2_vy = inner_product * VA_vy / VA_length;
			tmp2_vz = inner_product * VA_vz / VA_length;
		
			VBB_vx = tmp_vx - tmp2_vx;
			VBB_vy = tmp_vy - tmp2_vy;
			VBB_vz = tmp_vz - tmp2_vz;		
		}
	
		{
			double wrk_length;
		
			wrk_length = sqrt(VBB_vx * VBB_vx + VBB_vy * VBB_vy + VBB_vz * VBB_vz)
					* my_settings.tiles_per_res / 3.0;
			*pVB_vx = (((double) ave_dis) * VBB_vx) / wrk_length;
			*pVB_vy = (((double) ave_dis) * VBB_vy) / wrk_length;
			*pVB_vz = (((double) ave_dis) * VBB_vz) / wrk_length;
		}
	}
	return SUCCESS;		// 20071223  change the return value
}

/* 
   VC // VA x VB
   |VC| = sqrt(3)*ave_dis/t_fac
 */
int compute_VC_vector(int ind_atm, float ave_dis,
		       double VA_vx, double VA_vy, double VA_vz,
		       double VB_vx, double VB_vy, double VB_vz,
		       double * pVC_vx, double * pVC_vy, double * pVC_vz)
{
	double VC_vx, VC_vy, VC_vz;
	double wrk_length;
	
	VC_vx = VA_vy * VB_vz - VA_vz * VB_vy;
	VC_vy = VA_vz * VB_vx - VA_vx * VB_vz;
	VC_vz = VA_vx * VB_vy - VA_vy * VB_vx;
	wrk_length = sqrt(VC_vx * VC_vx + VC_vy * VC_vy + VC_vz * VC_vz)
		* my_settings.tiles_per_res / 3.0;
	{	// 20071223
		if (!FLOAT_NEQ(wrk_length, 1.0)) {	/* if wrk_length == 0.0 */
			*pVC_vx = (sqrt(3.0) * ((double) ave_dis) * VC_vx);
			*pVC_vy = (sqrt(3.0) * ((double) ave_dis) * VC_vy);
			*pVC_vz = (sqrt(3.0) * ((double) ave_dis) * VC_vz);
			return FAIL;
		}
	}
	*pVC_vx = (sqrt(3.0) * ((double) ave_dis) * VC_vx) / wrk_length;
	*pVC_vy = (sqrt(3.0) * ((double) ave_dis) * VC_vy) / wrk_length;
	*pVC_vz = (sqrt(3.0) * ((double) ave_dis) * VC_vz) / wrk_length;
	return SUCCESS;		// 20071223  change the return value
}

int coordinates_validation(int ind_atm, float ave_dis)
{
	double wrk_sum_x, wrk_sum_y, wrk_sum_z;
	double wrk_length, wrk_double;
	double N_vx, N_vy, N_vz;
	double C_vx, C_vy, C_vz;
	int i;
	
	/* VX + VY + VZ + VW = 0 */
	wrk_sum_x = 0.0;
	wrk_sum_y = 0.0;
	wrk_sum_z = 0.0;
	for (i = 0; i < 4; i++) {
		wrk_sum_x = wrk_sum_x + my_coor.axes[i].x;
		wrk_sum_y = wrk_sum_y + my_coor.axes[i].y;
		wrk_sum_z = wrk_sum_z + my_coor.axes[i].z;
	}
	if (FLOAT_NEQ(wrk_sum_x, ave_dis) || FLOAT_NEQ(wrk_sum_y, ave_dis)
	    || FLOAT_NEQ(wrk_sum_z, ave_dis)) {
		fprintf(STDERR,
			"coordinates_validation() : ERROR  VX+VY+VZ+VW=(%f,%f, %f)\n",
			wrk_sum_x, wrk_sum_y, wrk_sum_z);
		return FAIL;
	}
	/* |VX| = |VY| = |VZ| = |VW|
		=sqrt(6)*ave_dis/(my_settings.tiles_per_res/3) */
	wrk_double = sqrt(6.0) * ((double) ave_dis) * 3.0
		/ my_settings.tiles_per_res;
	for (i = 0; i < 4; i++) {
		if (FLOAT_NEQ((my_coor.axes[i].length - wrk_double), ave_dis)) {
			fprintf(STDERR,
				"coordinates_validation() : ERROR  my_coor.axes[%d].length=%f <=> sqrt(6)*ave_dis=%f\n",
				i, my_coor.axes[i].length, wrk_double);
			return FAIL;			
		}
	}
	
	/* |XY|=|ZW| */
	N_vx = my_coor.axes[0].x - my_coor.axes[1].x;
	N_vy = my_coor.axes[0].y - my_coor.axes[1].y;
	N_vz = my_coor.axes[0].z - my_coor.axes[1].z;
	wrk_length = sqrt(N_vx * N_vx + N_vy * N_vy + N_vz * N_vz);
	C_vx = my_coor.axes[2].x - my_coor.axes[3].x;
	C_vy = my_coor.axes[2].y - my_coor.axes[3].y;
	C_vz = my_coor.axes[2].z - my_coor.axes[3].z;
	wrk_double = sqrt(C_vx * C_vx + C_vy * C_vy + C_vz * C_vz);
	if (FLOAT_NEQ(wrk_length - wrk_double, ave_dis)) {
		fprintf(STDERR,
			"coordinates_validation() : ERROR  |XY|= %f <=> |ZW|= %f\n",
			wrk_length, wrk_double);
		return FAIL;
	}
	/* |XY|=|XZ| */
	C_vx = my_coor.axes[0].x - my_coor.axes[2].x;
	C_vy = my_coor.axes[0].y - my_coor.axes[2].y;
	C_vz = my_coor.axes[0].z - my_coor.axes[2].z;
	wrk_double = sqrt(C_vx * C_vx + C_vy * C_vy + C_vz * C_vz);
	if (FLOAT_NEQ(wrk_length - wrk_double, ave_dis)) {
		fprintf(STDERR,
			"coordinates_validation() : ERROR  |XY|= %f <=> |XZ|= %f\n",
			wrk_length, wrk_double);
		return FAIL;
	}
	/* |XY|=|XW| */
	C_vx = my_coor.axes[0].x - my_coor.axes[3].x;
	C_vy = my_coor.axes[0].y - my_coor.axes[3].y;
	C_vz = my_coor.axes[0].z - my_coor.axes[3].z;
	wrk_double = sqrt(C_vx * C_vx + C_vy * C_vy + C_vz * C_vz);
	if (FLOAT_NEQ(wrk_length - wrk_double, ave_dis)) {
		fprintf(STDERR,
			"coordinates_validation() : ERROR  |XY|= %f <=> |XW|= %f\n",
			wrk_length, wrk_double);
		return FAIL;
	}
	/* |XY|=|YZ| */
	C_vx = my_coor.axes[1].x - my_coor.axes[2].x;
	C_vy = my_coor.axes[1].y - my_coor.axes[2].y;
	C_vz = my_coor.axes[1].z - my_coor.axes[2].z;
	wrk_double = sqrt(C_vx * C_vx + C_vy * C_vy + C_vz * C_vz);
	if (FLOAT_NEQ(wrk_length - wrk_double, ave_dis)) {
		fprintf(STDERR,
			"coordinates_validation() : ERROR  |XY|= %f <=> |YZ|= %f\n",
			wrk_length, wrk_double);
		return FAIL;
	}
	/* |XY|=|YW| */
	C_vx = my_coor.axes[1].x - my_coor.axes[3].x;
	C_vy = my_coor.axes[1].y - my_coor.axes[3].y;
	C_vz = my_coor.axes[1].z - my_coor.axes[3].z;
	wrk_double = sqrt(C_vx * C_vx + C_vy * C_vy + C_vz * C_vz);
	if (FLOAT_NEQ(wrk_length - wrk_double, ave_dis)) {
		fprintf(STDERR,
			"coordinates_validation() : ERROR  |XY|= %f <=> |YW|= %f\n",
			wrk_length, wrk_double);
		return FAIL;
	}
	if (0) {
		/* Origin check: Origin = CA - (3x+2y+z)/4 for XYZW */
		/* Origin check: Origin = CA - (3y+2z+w)/4 for YZWX */
		wrk_sum_x = 3.0 * my_coor.axes[0].x + 2.0 * my_coor.axes[1].x
		+ my_coor.axes[2].x;
		wrk_sum_x = my_coor.origin.x + wrk_sum_x / 4.0;
		wrk_sum_x = wrk_sum_x - my_protein.atom[ind_atm].x;
		wrk_sum_y = 3.0 * my_coor.axes[0].y + 2.0 * my_coor.axes[1].y
			+ my_coor.axes[2].y;
		wrk_sum_y = my_coor.origin.y + wrk_sum_y / 4.0;
		wrk_sum_y = wrk_sum_y - my_protein.atom[ind_atm].y;
		wrk_sum_z = 3.0 * my_coor.axes[0].z + 2.0 * my_coor.axes[1].z
			+ my_coor.axes[2].z;
		wrk_sum_z = my_coor.origin.z + wrk_sum_z / 4.0;
		wrk_sum_z = wrk_sum_z - my_protein.atom[ind_atm].z;
		if (FLOAT_NEQ(wrk_sum_x, ave_dis) || FLOAT_NEQ(wrk_sum_y, ave_dis)
		    || FLOAT_NEQ(wrk_sum_z, ave_dis)) {
			fprintf(STDERR,
				"coordinates_validation() : ERROR  - diff of Origin = (%f,%f, %f) -\n",
				wrk_sum_x, wrk_sum_y, wrk_sum_z);
			return FAIL;
		}
	}
	return SUCCESS;
}

/* Count of paths = 2**n
* p_ct =  code[0]*2**(n-1) + code[1]*2**(n-2) + ... + code[n-1]
* where n = tiles_per_res
*/
int alloc_npath_table(struct pe_path_of_length_n ** npath_tbl, int path_length)
{
	int max_path;
	int tile_type;
	
	if (npath_tbl[0] != NULL) {
		fprintf(STDERR, 
			"alloc_npath_table() : CAUTION npath_table[0] already allocated\n");
		return SUCCESS;
	}
	max_path = (int) pow(2.0, (double) path_length);
	for (tile_type = 0; tile_type < 24; tile_type++) {
		npath_tbl[tile_type] = (struct pe_path_of_length_n *)
			malloc(max_path * sizeof(struct pe_path_of_length_n));
		if (npath_tbl[tile_type] == NULL) {
			fprintf(STDERR, 
				"alloc_npath_code_seq() : ERROR npath_table[%d] not allocated\n",
				tile_type);
			return FAIL;
		}
		if (alloc_npath_code_seq(npath_tbl, path_length, tile_type) == FAIL)
			return FAIL;
	}
	return SUCCESS;
}

void compute_position_of_last_tile(struct pe_path_of_length_n ** npath_tbl,
				   int path_length, int tile_type, int p_ct)
{
	int cur_type;
	int t_ct;
	float wrk_x, wrk_y, wrk_z;
	
	cur_type = tile_type;
	wrk_x = my_trn_table[cur_type].cx;
	wrk_y = my_trn_table[cur_type].cy;
	wrk_z = my_trn_table[cur_type].cz;
	if (my_settings.along_backbone == TRUE) {
		npath_tbl[tile_type][p_ct].bb_c[0].x = wrk_x;
		npath_tbl[tile_type][p_ct].bb_c[0].y = wrk_y;
		npath_tbl[tile_type][p_ct].bb_c[0].z = wrk_z;
		npath_tbl[tile_type][p_ct].bb_c[0].length = 0.0;
	}
	/* The position of the (t_ct+1)-th tile (from 1 to n-1 th) */
	for (t_ct = 0; t_ct < path_length - 1; t_ct++) {
		if (npath_tbl[tile_type][p_ct].code[t_ct] == C_DOWN) {
			wrk_x = wrk_x + my_trn_table[cur_type].downward_cx
				- my_trn_table[cur_type].cx;
			wrk_y = wrk_y + my_trn_table[cur_type].downward_cy
				- my_trn_table[cur_type].cy;
			wrk_z = wrk_z + my_trn_table[cur_type].downward_cz
				- my_trn_table[cur_type].cz;
			if (npath_tbl[tile_type][p_ct].code[t_ct+1] == C_DOWN)
				cur_type = 
					my_trn_table[cur_type].downward_D_type;
			else
				cur_type = 
					my_trn_table[cur_type].downward_U_type;
		} else {
			wrk_x = wrk_x + my_trn_table[cur_type].upward_cx
				- my_trn_table[cur_type].cx;
			wrk_y = wrk_y + my_trn_table[cur_type].upward_cy
				- my_trn_table[cur_type].cy;
			wrk_z = wrk_z + my_trn_table[cur_type].upward_cz
				- my_trn_table[cur_type].cz;
			if (npath_tbl[tile_type][p_ct].code[t_ct+1] == C_DOWN)
				cur_type = my_trn_table[cur_type].upward_D_type;
			else
				cur_type = my_trn_table[cur_type].upward_U_type;
		}
		if (my_settings.along_backbone == TRUE) {
			npath_tbl[tile_type][p_ct].bb_c[t_ct+1].x = wrk_x;
			npath_tbl[tile_type][p_ct].bb_c[t_ct+1].y = wrk_y;
			npath_tbl[tile_type][p_ct].bb_c[t_ct+1].z = wrk_z;
			npath_tbl[tile_type][p_ct].bb_c[t_ct+1].length = 0.0;
		}
	}
	/* The position of the n-th tile = the next target atom */
	/* code of (n-1)-th tile = npath_table[tile_type][p_ct].code[t_ct] */
	if (npath_tbl[tile_type][p_ct].code[t_ct] == C_DOWN) {
		npath_tbl[tile_type][p_ct].c.x = wrk_x
			+ my_trn_table[cur_type].downward_cx
			- my_trn_table[cur_type].cx;
		npath_tbl[tile_type][p_ct].c.y = wrk_y
			+ my_trn_table[cur_type].downward_cy
			- my_trn_table[cur_type].cy;
		npath_tbl[tile_type][p_ct].c.z = wrk_z
			+ my_trn_table[cur_type].downward_cz
			- my_trn_table[cur_type].cz;
	} else {
		npath_tbl[tile_type][p_ct].c.x = wrk_x
			+ my_trn_table[cur_type].upward_cx
			- my_trn_table[cur_type].cx;
		npath_tbl[tile_type][p_ct].c.y = wrk_y
			+ my_trn_table[cur_type].upward_cy
			- my_trn_table[cur_type].cy;
		npath_tbl[tile_type][p_ct].c.z = wrk_z
			+ my_trn_table[cur_type].upward_cz
			- my_trn_table[cur_type].cz;
	}
	/* The type of the last tile */
	npath_tbl[tile_type][p_ct].type = cur_type;
	wrk_x = npath_tbl[tile_type][p_ct].c.x - my_trn_table[tile_type].cx;
	wrk_y = npath_tbl[tile_type][p_ct].c.y - my_trn_table[tile_type].cy;
	wrk_z = npath_tbl[tile_type][p_ct].c.z - my_trn_table[tile_type].cz;
	npath_tbl[tile_type][p_ct].c.length
		= sqrt(wrk_x*wrk_x + wrk_y*wrk_y + wrk_z*wrk_z);
	return;
}

void compute_direction_of_N_atom(int ind_atm, double * pN_vx, double * pN_vy,
				 double * pN_vz)
{
	double N_vx, N_vy, N_vz;
	double wrk_length;
	int seq_ct;
	int prv_atm;
	
	seq_ct = -1;
	prv_atm = seek_previous_target_atom(ind_atm);
	if (prv_atm < 0) {
		if (DEBUG) print_my_sequences();	/* 2007-01-15 */
		fprintf(STDERR,
			"compute_direction_of_N_atom() : ERROR There exists no predecessor of %d-th atom [backward].\n",
			ind_atm);	// 2007-08-22
		exit(1);
	}
	/* Note that |N| may not equal to |C|. */
	N_vx = my_protein.atom[prv_atm].x - my_protein.atom[ind_atm].x;
	N_vy = my_protein.atom[prv_atm].y - my_protein.atom[ind_atm].y;
	N_vz = my_protein.atom[prv_atm].z - my_protein.atom[ind_atm].z;
	wrk_length = sqrt(N_vx * N_vx + N_vy * N_vy + N_vz * N_vz);	
	*pN_vx = N_vx / wrk_length;
	*pN_vy = N_vy / wrk_length;
	*pN_vz = N_vz / wrk_length;
	return;
}

void compute_direction_of_C_atom(int ind_atm, double * pC_vx, double * pC_vy,
				 double * pC_vz)
{
	double C_vx, C_vy, C_vz;
	double wrk_length;
	int nxt_atm;
	
	nxt_atm = seek_next_target_atom(ind_atm);
	if (nxt_atm > my_protein.atm_num ) {
		if (DEBUG) print_my_sequences();	/* 2007-01-15 */
		fprintf(STDERR,
			"compute_direction_of_C_atom() : ERROR There exists no atom [forward] (FROM %d TO %d).\n",
			ind_atm + 1, nxt_atm - 1);
		exit(1);
	}
	C_vx = my_protein.atom[nxt_atm].x - my_protein.atom[ind_atm].x;
	C_vy = my_protein.atom[nxt_atm].y - my_protein.atom[ind_atm].y;
	C_vz = my_protein.atom[nxt_atm].z - my_protein.atom[ind_atm].z;
	wrk_length = sqrt(C_vx * C_vx + C_vy * C_vy + C_vz * C_vz);
	*pC_vx = C_vx / wrk_length;
	*pC_vy = C_vy / wrk_length;
	*pC_vz = C_vz / wrk_length;
	return;
}


int recompute_VA_vector(int ind_atm, float ave_dis,
			double * pVA_vx, double * pVA_vy, double * pVA_vz)
{
	double N_vx, N_vy, N_vz;
	double C_vx, C_vy, C_vz;
	int prv_atm;
	int nxt_atm;
	
	prv_atm = seek_previous_target_atom(ind_atm);
	if (prv_atm < 0) {
		compute_direction_of_N_atom(ind_atm, &N_vx, &N_vy, &N_vz);
	} else {
		compute_direction_of_N_atom(prv_atm, &N_vx, &N_vy, &N_vz);
	}
	nxt_atm = seek_next_target_atom(ind_atm);
	if (nxt_atm > my_protein.atm_num ) {
		compute_direction_of_C_atom(ind_atm, &C_vx, &C_vy, &C_vz);
	} else {
		compute_direction_of_C_atom(nxt_atm, &C_vx, &C_vy, &C_vz);
	}
	if (compute_VA_vector(ind_atm, ave_dis, N_vx, N_vy, N_vz, C_vx, C_vy, C_vz, 
			      pVA_vx, pVA_vy, pVA_vz) == FAIL) {
		if (0) fprintf(STDOUT,
			"recompute_VA_vector() : CAUTION |VA| = 0.0 (prv_atm=%d, ind_atm=%d, nxt_atm=%d, ave_dis=%f)\n",
			prv_atm, ind_atm, nxt_atm, ave_dis);
		return FAIL;
	} else {
		return SUCCESS;
	}
}


int recompute_VB_vector(int ind_atm, float ave_dis,
			double VA_vx, double VA_vy, double VA_vz,
			double * pVB_vx, double * pVB_vy, double * pVB_vz)
{
	double N_vx, N_vy, N_vz;
	double C_vx, C_vy, C_vz;
	int prv_atm;
	int nxt_atm;
	
	prv_atm = seek_previous_target_atom(ind_atm);
	if (prv_atm < 0) {
		compute_direction_of_N_atom(ind_atm, &N_vx, &N_vy, &N_vz);
	} else {
		compute_direction_of_N_atom(prv_atm, &N_vx, &N_vy, &N_vz);
	}
	nxt_atm = seek_next_target_atom(ind_atm);
	if (nxt_atm > my_protein.atm_num ) {
		compute_direction_of_C_atom(ind_atm, &C_vx, &C_vy, &C_vz);
	} else {
		compute_direction_of_C_atom(nxt_atm, &C_vx, &C_vy, &C_vz);
	}
	if (compute_VB_vector(ind_atm, ave_dis, N_vx, N_vy, N_vz, C_vx, C_vy, C_vz,
			      VA_vx, VA_vy, VA_vz, pVB_vx, pVB_vy, pVB_vz) == FAIL) {
		if (0) fprintf(STDOUT,
			"recompute_VB_vector() : CAUTION |VB| = 0.0 (prv_atm=%d, ind_atm=%d, nxt_atm=%d, ave_dis=%f)\n",
			prv_atm, ind_atm, nxt_atm, ave_dis);
		return FAIL;
	} else {
		return SUCCESS;
	}
}


void set_arbitrary_VB_vector(float ave_dis, double VA_vx, double VA_vy, double VA_vz,
			     double * pVB_vx, double * pVB_vy, double * pVB_vz)
{
	double tmp_vx, tmp_vy, tmp_vz;
	double VBB_vx, VBB_vy, VBB_vz;
	
	tmp_vx = VA_vx;
	tmp_vy = VA_vy;
	tmp_vz = VA_vz;
	if (FLOAT_NEQ(VA_vx, 1.0)) {	/* if VA_vx != 0.0 */
		tmp_vx = -1.0 * VA_vx;
	} else if (FLOAT_NEQ(VA_vy, 1.0)) {
		tmp_vy = -1.0 * VA_vy;
	} else {
		tmp_vz = -1.0 * VA_vz;
	}

	{	// Compute tmp2 = (<VA,tmp>/<VA,VA>)VA and
		//         VBB = tmp - tmp2   (Note that VA _|_ VB)
		double tmp2_vx, tmp2_vy, tmp2_vz;
		double inner_product;
		double VA_length;
		
		inner_product = tmp_vx * VA_vx + tmp_vy * VA_vy + tmp_vz * VA_vz;
		VA_length = VA_vx * VA_vx + VA_vy * VA_vy + VA_vz * VA_vz;
		tmp2_vx = inner_product * VA_vx / VA_length;
		tmp2_vy = inner_product * VA_vy / VA_length;
		tmp2_vz = inner_product * VA_vz / VA_length;
		
		VBB_vx = tmp_vx - tmp2_vx;
		VBB_vy = tmp_vy - tmp2_vy;
		VBB_vz = tmp_vz - tmp2_vz;		
	}

	{
		double wrk_length;
		
		wrk_length = sqrt(VBB_vx * VBB_vx + VBB_vy * VBB_vy + VBB_vz * VBB_vz)
			* my_settings.tiles_per_res / 3.0;
		*pVB_vx = (((double) ave_dis) * VBB_vx) / wrk_length;
		*pVB_vy = (((double) ave_dis) * VBB_vy) / wrk_length;
		*pVB_vz = (((double) ave_dis) * VBB_vz) / wrk_length;
	}
	return;
}


/* 
       VA // N->C
       VB // (CA->N)+(CA->C)
       VC // VA x VB
       |VA| = sqrt(6)*ave_dis/t_fac               
       |VB| = ave_dis/t_fac
       |VC| = sqrt(3)*ave_dis/t_fac
   where 
       t_fac = tiles_per_res/3 = tiles per atom
       ave_dis = average distance between successive atoms of the
		  protein backbone
*/
void compute_3d_coordinates(int ind_atm, float ave_dis,
			    double * pVA_vx, double * pVA_vy, double * pVA_vz,
			    double * pVB_vx, double * pVB_vy, double * pVB_vz,
			    double * pVC_vx, double * pVC_vy, double * pVC_vz)
{
	double N_vx, N_vy, N_vz;
	double C_vx, C_vy, C_vz;
	
	compute_direction_of_N_atom(ind_atm, &N_vx, &N_vy, &N_vz);
	compute_direction_of_C_atom(ind_atm, &C_vx, &C_vy, &C_vz);
	if (compute_VA_vector(ind_atm, ave_dis, N_vx, N_vy, N_vz, C_vx, C_vy, C_vz, 
			      pVA_vx, pVA_vy, pVA_vz) == FAIL) {	// 20071223
		// position of N == position of C  (|VA| = 0.0)
		if (DEBUG)
			fprintf(STDOUT,
				"compute_3d_coordinates() : CAUTION |VA| = 0.0 (ind_atm=%d, ave_dis=%f)\n",
				ind_atm, ave_dis);
		if (recompute_VA_vector(ind_atm, ave_dis, pVA_vx, pVA_vy, pVA_vz) == FAIL) {
			fprintf(STDOUT,
				"compute_3d_coordinates() : CAUTION the fragment is contracted into a point (ind_atm=%d, ave_dis=%f)\n",
				ind_atm, ave_dis);
		}
		if (compute_VB_vector(ind_atm, ave_dis, N_vx, N_vy, N_vz, C_vx, C_vy, C_vz,
				      *pVA_vx, *pVA_vy, *pVA_vz, pVB_vx, pVB_vy, pVB_vz) == FAIL) {
			// N, CA, and C are on a straight line (|VB| = 0.0)
			fprintf(STDOUT,
				"compute_3d_coordinates() : CAUTION the fragment is a straight line (ind_atm=%d, ave_dis=%f)\n",
				ind_atm, ave_dis);
			set_arbitrary_VB_vector(ave_dis, *pVA_vx, *pVA_vy, *pVA_vz, pVB_vx, pVB_vy, pVB_vz);
		}
	} else {
		if (compute_VB_vector(ind_atm, ave_dis, N_vx, N_vy, N_vz, C_vx, C_vy, C_vz,
				      *pVA_vx, *pVA_vy, *pVA_vz, pVB_vx, pVB_vy, pVB_vz) == FAIL) {	// 20071223
			// N, CA, and C are on a straight line (|VB| = 0.0)
			if (DEBUG)
				fprintf(STDOUT,
					"compute_3d_coordinates() : CAUTION |VB| = 0.0 (ind_atm=%d, ave_dis=%f)\n",
					ind_atm, ave_dis);
			if (recompute_VB_vector(ind_atm, ave_dis, *pVA_vx, *pVA_vy, *pVA_vz,
						pVB_vx, pVB_vy, pVB_vz) == FAIL) {
				fprintf(STDOUT,
					"compute_3d_coordinates() : CAUTION the fragment is a straight line (ind_atm=%d, ave_dis=%f)\n",
					ind_atm, ave_dis);
				set_arbitrary_VB_vector(ave_dis, *pVA_vx, *pVA_vy, *pVA_vz, pVB_vx, pVB_vy, pVB_vz);
			}
		}
	}
	if (compute_VC_vector(ind_atm, ave_dis, *pVA_vx, *pVA_vy, *pVA_vz,
			      *pVB_vx, *pVB_vy, *pVB_vz, 
			      pVC_vx, pVC_vy, pVC_vz) == FAIL) {	// 20071223
		fprintf(STDOUT,
			"compute_3d_coordinates() : CAUTION |VC| = 0.0 (ind_atm=%d, ave_dis=%f)\n",
			ind_atm, ave_dis);
	}
	return;
}

void set_my_coordinates(int ind_atm, float ave_dis, int init_tile_type)
{
	double VA_vx, VA_vy, VA_vz;
	double VB_vx, VB_vy, VB_vz;
	double VC_vx, VC_vy, VC_vz;
	int i;
	
	if (ind_atm < 2 || ind_atm > my_protein.atm_num) {
		fprintf(STDERR,
			"set_my_coordinates() : ERROR   initial atom = %d (number of atoms = %d)\n",
			ind_atm, my_protein.atm_num);
		exit(1);
	}
	
	compute_3d_coordinates(ind_atm, ave_dis, &VA_vx, &VA_vy, &VA_vz,
			       &VB_vx, &VB_vy, &VB_vz, &VC_vx, &VC_vy, &VC_vz);
	if (0) {
		float wrk_length, wrk_length2;
		
		wrk_length = sqrt(VA_vx * VA_vx + VA_vy * VA_vy + VA_vz * VA_vz);
		wrk_length2 = sqrt(6.0) * ((double) ave_dis) * 3.0 / my_settings.tiles_per_res;
		if (FLOAT_NEQ((wrk_length - wrk_length2), ave_dis)) {
			fprintf(STDERR,
				"set_my_coordinates() : ERROR  |VA|=%f <=> sqrt(6)*ave_dis=%f\n",
				wrk_length, wrk_length2);
			exit(1);
		}
		wrk_length = sqrt(VB_vx * VB_vx + VB_vy * VB_vy + VB_vz * VB_vz);
		wrk_length2 = ((double) ave_dis) * 3.0 / my_settings.tiles_per_res;
		if (FLOAT_NEQ((wrk_length - wrk_length2), ave_dis)) {
			fprintf(STDERR,
				"set_my_coordinates() : ERROR  |VB|=%f <=> ave_dis=%f\n",
				wrk_length, wrk_length2);
			exit(1);
		}
		wrk_length = sqrt(VC_vx * VC_vx + VC_vy * VC_vy + VC_vz * VC_vz);
		wrk_length2 = sqrt(3.0) * ((double) ave_dis) * 3.0 / my_settings.tiles_per_res;
		if (FLOAT_NEQ((wrk_length - wrk_length2), ave_dis)) {
			fprintf(STDERR,
					"set_my_coordinates() : ERROR  |VC|=%f <=> sqrt(3)*ave_dis=%f\n",
					wrk_length, wrk_length2);
			exit(1);
		}
	}
	
	if (init_tile_type == XYZW) {
		/* X = (VA + 6VB - 2VC)/3
		   Y = (VA + 4VC)/3
		   Z = (VA - 6VB - 2VC)/3
		   W = -VA
		*/
		my_coor.axes[0].x = (VA_vx + 6.0 * VB_vx - 2.0 * VC_vx) / 3.0;
		my_coor.axes[0].y = (VA_vy + 6.0 * VB_vy - 2.0 * VC_vy) / 3.0;
		my_coor.axes[0].z = (VA_vz + 6.0 * VB_vz - 2.0 * VC_vz) / 3.0;
		my_coor.axes[1].x = (VA_vx + 4.0 * VC_vx) / 3.0;
		my_coor.axes[1].y = (VA_vy + 4.0 * VC_vy) / 3.0;
		my_coor.axes[1].z = (VA_vz + 4.0 * VC_vz) / 3.0;
		my_coor.axes[2].x = (VA_vx - 6.0 * VB_vx - 2.0 * VC_vx) / 3.0;
		my_coor.axes[2].y = (VA_vy - 6.0 * VB_vy - 2.0 * VC_vy) / 3.0;
		my_coor.axes[2].z = (VA_vz - 6.0 * VB_vz - 2.0 * VC_vz) / 3.0;
		my_coor.axes[3].x = - VA_vx;
		my_coor.axes[3].y = - VA_vy;
		my_coor.axes[3].z = - VA_vz;
		for (i = 0; i < 4; i++) {
			my_coor.axes[i].length
			= sqrt(my_coor.axes[i].x * my_coor.axes[i].x
			       + my_coor.axes[i].y * my_coor.axes[i].y
			       + my_coor.axes[i].z * my_coor.axes[i].z);
		}
		
		/* Origin = (Center of CA) - (VA+2VB)/2     */
		my_coor.origin.x = my_protein.atom[ind_atm].x
			- (VA_vx + 2.0 * VB_vx) / 2.0;
		my_coor.origin.y = my_protein.atom[ind_atm].y
			- (VA_vy + 2.0 * VB_vy) / 2.0;
		my_coor.origin.z = my_protein.atom[ind_atm].z
			- (VA_vz + 2.0 * VB_vz) / 2.0;
	} else if (init_tile_type == YZWX) {
		/* Y = (VA + 6VB - 2VC)/3
		   Z = (VA + 4VC)/3
		   W = (VA - 6VB - 2VC)/3
		   X = -VA
		*/
		my_coor.axes[1].x = (VA_vx + 6.0 * VB_vx - 2.0 * VC_vx) / 3.0;
		my_coor.axes[1].y = (VA_vy + 6.0 * VB_vy - 2.0 * VC_vy) / 3.0;
		my_coor.axes[1].z = (VA_vz + 6.0 * VB_vz - 2.0 * VC_vz) / 3.0;
		my_coor.axes[2].x = (VA_vx + 4.0 * VC_vx) / 3.0;
		my_coor.axes[2].y = (VA_vy + 4.0 * VC_vy) / 3.0;
		my_coor.axes[2].z = (VA_vz + 4.0 * VC_vz) / 3.0;
		my_coor.axes[3].x = (VA_vx - 6.0 * VB_vx - 2.0 * VC_vx) / 3.0;
		my_coor.axes[3].y = (VA_vy - 6.0 * VB_vy - 2.0 * VC_vy) / 3.0;
		my_coor.axes[3].z = (VA_vz - 6.0 * VB_vz - 2.0 * VC_vz) / 3.0;
		my_coor.axes[0].x = - VA_vx;
		my_coor.axes[0].y = - VA_vy;
		my_coor.axes[0].z = - VA_vz;
		for (i = 0; i < 4; i++) {
			my_coor.axes[i].length
			= sqrt(my_coor.axes[i].x * my_coor.axes[i].x
			       + my_coor.axes[i].y * my_coor.axes[i].y
			       + my_coor.axes[i].z * my_coor.axes[i].z);
		}
		
		/* Origin = (Center of CA) - (VA+2VB)/2     */
		my_coor.origin.x = my_protein.atom[ind_atm].x
			- (VA_vx + 2.0 * VB_vx) / 2.0;
		my_coor.origin.y = my_protein.atom[ind_atm].y
			- (VA_vy + 2.0 * VB_vy) / 2.0;
		my_coor.origin.z = my_protein.atom[ind_atm].z
			- (VA_vz + 2.0 * VB_vz) / 2.0;
	} else {
		fprintf(STDERR,
			"set_my_coordinates() : ERROR  init_tile_type = %d not supported\n",
			init_tile_type);
		exit(1);
	}
	
	my_coor.ave_dis = ave_dis;
	
	if (0) print_my_coordinates();
	if (coordinates_validation(ind_atm, ave_dis) == FAIL) {
		fprintf(STDERR,
			"set_my_coordinates() : ERROR  coordinates_validation() failed\n");
		print_my_coordinates();		// 20071223
		exit(1);
	}
	return;
}

/* fn tiles for a residue (fn = 2k+1 for some k) */
int setup_npath_table(struct pe_path_of_length_n ** npath_tbl, int path_length)
{
	int tile_type;
	int p_ct;
	int max_path;
	
	if (path_length <= 0) {
		fprintf(STDERR, "setup_npath_table() :  ERROR path_length = %d\n",
			path_length);
		return FAIL;
	}
	if (npath_tbl[0] == NULL) {
		if (alloc_npath_table(npath_tbl, path_length) == FAIL)
			return FAIL;
	}
	max_path = (int) pow(2.0, (double) path_length);
	for (tile_type = 0; tile_type < 24; tile_type++) {
		for (p_ct = 0; p_ct < max_path; p_ct++) {
			compute_position_of_last_tile(npath_tbl, path_length,
						      tile_type, p_ct);
		}
	}
	return SUCCESS;
}

void setup_npath(float ave_dis)
{
	int path_length;
	
	if (my_settings.variable_path_length == TRUE) {
		int i;
		
		path_length = (int) my_settings.tiles_per_res
			- NP_TBL_DISP;
		if (path_length <= 0) {
			fprintf(STDERR,
				"setup_npathe() :  ERROR path_length = tiles_per_res - %d = %d\n",
				path_length, NP_TBL_DISP);
		}
		for (i = 0; i < NP_TBL_SIZE; i++, path_length++) {
			if (setup_npath_table(np_table[i], path_length)
			    == FAIL) {
				print_npath_table(np_table[i],
						  path_length);
				exit(1);
			}
		}
	} else {
		path_length = (int) my_settings.tiles_per_res;
		if (setup_npath_table(np_table[NP_TBL_DISP],
				      path_length) == FAIL) {
			if (0) print_npath_table(np_table[NP_TBL_DISP],
						 path_length);
			exit(1);
		}
		if (0) print_npath_table(np_table[NP_TBL_DISP], path_length);
	}
	my_settings.path_length = (int) my_settings.tiles_per_res;
	my_settings.max_path = (int) pow(2.0, (double)
					 my_settings.tiles_per_res);
	npath_table = np_table[NP_TBL_DISP];
	return;
}

/* Setup tables */
/* Set the initial tetrahedron */
/* atom[ind]  <=>  tiles[my_protein.tile_max_num]  =  1[init_tile_type] */
int setup_tables_and_init_tile(float ave_dis, int ind,
			       struct pe_slt_tile * slt_tiles,
			       int init_tile_type)
{
	int t_ind;
	
	if (0) {	// 20071223
		fprintf(STDOUT, "setup_tables_and_init_tile() : DEBUG ind = %d, ave_dis = %8.3f, init_tile_type = %d\n",
			ind, ave_dis, init_tile_type);
	}
	set_my_coordinates(ind, ave_dis, init_tile_type);
	if (0) print_my_coordinates();
	setup_transition_table(ave_dis /* used for error check */);
	if (0) print_trn_table();
	if (my_settings.ca_only_flag == TRUE)
		setup_npath(ave_dis);
	
	t_ind = my_protein.tile_max_num;	/* temporal position */
	if (set_initial_tile(slt_tiles + t_ind, ind, init_tile_type) == FAIL)
		exit(1);
	my_protein.tile_num = 1;
	return t_ind;
}

void encoding_suite(int best_ind, float ave_dis, struct pe_slt_tile * slt_tiles)
{
	int i;
	int t_ind;
	
	for (i = 0; i <= my_protein.tile_max_num; i++)
		clear_slt_tile(slt_tiles + i);
	my_protein.tile_num = 0;
	
	t_ind = setup_tables_and_init_tile(ave_dis, best_ind, slt_tiles,
					   my_protein.best_t_type);
	if (chain_encoding(best_ind, &t_ind, slt_tiles) == FAIL){
		fprintf(STDOUT,
			"encoding_suite() : ERROR chain_encoding() failed\n");
		exit(1);
	}
	set_tile_heads_and_tails(slt_tiles);
	
	if (my_settings.move_to_target != TRUE)
		error_calculation(slt_tiles);
	else
		compute_average_of_errors(slt_tiles);
	if (0) print_my_slt_tiles(slt_tiles);
	if (DEBUG) {
		if (my_settings.move_to_target != TRUE) {
			if (check_error(slt_tiles) == FAIL) exit(1);
		}
		if (my_settings.rotate_tile == FALSE) {
			if (check_tile_position(slt_tiles) == FAIL) exit(1);
			if (0) fprintf(STDOUT,
				"encoding_suite() : Position check done (tile_num= %d).\n",
				my_protein.tile_num);
			if (check_top_vertex(slt_tiles) == FAIL) exit(1);
			if (0) fprintf(STDOUT,
				"encoding_suite() : Top vertex check done (tile_num= %d).\n",
				my_protein.tile_num);
		}
	}
	return;
}
