/*
 *  pe_print_functions.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/01/05.
 *  Copyright (C) 2005, 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_pdb_data.h"
#include "pe_encoding.h"

extern struct pe_protein my_protein;
extern struct pe_trn_table my_trn_table[];
extern struct pe_coor_system my_coor;

void print_pdb_line(struct pe_pdb_line * pt_line)
{
	fprintf(STDOUT, "name  |record_body\n");
	fprintf(STDOUT, "%.6s|%.74s\n", pt_line->record_name,
		pt_line->record_body);
	fflush(STDOUT);
}

void print_pdb_header(struct pe_pdb_header * pt_hd)
{
	fprintf(STDOUT,
		"name  |class                                   |date     |idcode\n");
	fprintf(STDOUT, "%.6s|%.40s|%.9s|%.4s\n", pt_hd->record_name,
		pt_hd->class, pt_hd->date, pt_hd->idcode);
	fflush(STDOUT);
}

void print_pdb_compnd(struct pe_pdb_compnd * pt_line)
{
/*    fprintf(STDOUT,"name  |c |spec_list\n"); */
	fprintf(STDOUT, "%.6s|%.2s|%.60s\n", pt_line->record_name,
		pt_line->continuation, pt_line->spec_list);
	fflush(STDOUT);
}

void print_pdb_model(struct pe_pdb_model * pt_line)
{
/*    fprintf(STDOUT,"name  |serial_num\n"); */
	fprintf(STDOUT, "%.6s|%.4s\n", pt_line->record_name,
		pt_line->serial_num);
	fflush(STDOUT);
}

void print_pdb_atom(struct pe_pdb_atom * pt_atm)
{
	fprintf(STDOUT,
		"name  |ser  |atm |l|res|c|seq |i|x-coor. |y-coor. |z-coor. |occu  |temp  |seg |el|chg\n");
	fprintf(STDOUT,
		"%.6s|%.5s|%.4s|%c|%.3s|%c|%.4s|%c|%.8s|%.8s|%.8s|%.6s|%.6s|%.4s|%.2s|%.2s\n",
		pt_atm->record_name, pt_atm->serial_num, pt_atm->atom_name,
		pt_atm->alt_loc, pt_atm->res_name, pt_atm->chain_id,
		pt_atm->res_seq, pt_atm->i_code, pt_atm->x, pt_atm->y,
		pt_atm->z, pt_atm->occupancy, pt_atm->temp_fac, pt_atm->seg_id,
		pt_atm->element, pt_atm->charge);
	fflush(STDOUT);
}

void print_pdb_ter(struct pe_pdb_ter * pt_line)
{
	fprintf(STDOUT, "name  |ser  |res|c|ser |i\n");
	fprintf(STDOUT, "%.6s|%.5s|%.3s|%c|%.4s|%c\n", pt_line->record_name,
		pt_line->serial_num,
		pt_line->res_name,
		pt_line->chain_id, pt_line->res_serial_num, pt_line->i_code);
	fflush(STDOUT);
}

void print_pdb_master(struct pe_pdb_master * pt_line)
{
	fprintf(STDOUT,
		"name  |Rem  |zero |Het  |Hlx  |Sht  |Turn |Site |Xfor |Coor |Ter  |Cnct |Sqrs\n");
	fprintf(STDOUT,
		"%.6s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s|%.5s\n",
		pt_line->record_name, pt_line->num_remark, pt_line->zeros,
		pt_line->num_het, pt_line->num_helix, pt_line->num_sheet,
		pt_line->num_turn, pt_line->num_site, pt_line->num_xform,
		pt_line->num_coor, pt_line->num_ter, pt_line->num_conect,
		pt_line->num_seqres);
	fflush(STDOUT);
}

void print_my_molecule(struct pe_molecule * pt_strct)
{
	fprintf(STDOUT, "mol. : [chn_head, chn_tail] id name\n");
	fprintf(STDOUT, "mol. : [ %7d,  %7d]  %c %s\n", pt_strct->chain_head,
		pt_strct->chain_tail, pt_strct->id, pt_strct->name);
	fflush(STDOUT);
}

void print_my_chain(struct pe_chain * pt_strct)
{
	fprintf(STDOUT,
		"chain: [atm_head, atm_tail] chn_id |mol_id |mod_num [tile_head,tile_tail]\n");
	fprintf(STDOUT, "chain: [ %7d,  %7d] %7c|%7d|%7d [ %8d, %8d]\n",
		pt_strct->atom_head, pt_strct->atom_tail, pt_strct->id,
		pt_strct->mol_id, pt_strct->mod_num, pt_strct->tile_head,
		pt_strct->tile_tail);
	fflush(STDOUT);
}

void print_my_protein(struct pe_protein * pt_strct)
{
	int i;

	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "my_protein: mod |mol |chn |seq |res |atm |\n");
	fprintf(STDOUT, "my_protein: %4d|%4d|%4d|%4d|%4d|%4d|\n", pt_strct->mod_num,
		pt_strct->mol_num,
		pt_strct->chn_num,
		pt_strct->seq_num, pt_strct->res_num, pt_strct->atm_num);
/*
	fprintf(STDOUT,"my_protein: tiles/res|tile_max |dist_fac\n");
    fprintf(STDOUT,"my_protein: %9.4f|%9d|%9.4f\n", pt_strct->tiles_per_res, pt_strct->tile_max_num, pt_strct->dist_factor);
*/
	if (pt_strct->model != NULL) {
		fprintf(STDOUT, "my_protein: <models>\n");
		fprintf(STDOUT, "my_protein:            [seq_head, seq_tail]\n");
		for (i = 1; i <= pt_strct->mod_num; i++) {
			fprintf(STDOUT, "my_protein: model[%2d]= [ %7d,  %7d]\n", i,
				pt_strct->model[i].seq_head,
				pt_strct->model[i].seq_tail);
		}
	}

	fprintf(STDOUT, "my_protein: <molecules>\n");
	fprintf(STDOUT, "my_protein:                [chn_head, chn_tail] id name\n");
	for (i = 1; i <= pt_strct->mol_num; i++) {
		fprintf(STDOUT, "my_protein: molecules[%2d]= [ %7d,  %7d] %2d %s\n",
			i, pt_strct->molecule[i].chain_head,
			pt_strct->molecule[i].chain_tail,
			pt_strct->molecule[i].id, pt_strct->molecule[i].name);
	}

	fprintf(STDOUT, "my_protein: <chains>\n");
	fprintf(STDOUT,
		"my_protein:            [atm_head, atm_tail] chn_id |mol_id |\n");
	for (i = 1; i <= pt_strct->chn_num; i++) {
		fprintf(STDOUT, "my_protein: chain[%2d]= [ %7d,  %7d] %7c|%7d|\n", i,
			pt_strct->chain[i].atom_head,
			pt_strct->chain[i].atom_tail,
			pt_strct->chain[i].id, pt_strct->chain[i].mol_id);
	}

	if (pt_strct->sequence != NULL) {
		fprintf(STDOUT, "my_protein: <sequences>\n");
		fprintf(STDOUT,
			"my_protein:                [atm_head, atm_tail] chn_id |mol_id |mod_num [tile_head,tile_tail]\n");
		for (i = 1; i <= pt_strct->seq_num; i++) {
			fprintf(STDOUT,
				"my_protein: sequence[%3d]= [ %7d,  %7d] %7c|%7d|%7d [ %8d, %8d]\n",
				i, pt_strct->sequence[i].atom_head,
				pt_strct->sequence[i].atom_tail,
				pt_strct->sequence[i].id,
				pt_strct->sequence[i].mol_id,
				pt_strct->sequence[i].mod_num,
				pt_strct->sequence[i].tile_head,
				pt_strct->sequence[i].tile_tail);
		}
	}

	fprintf(STDOUT, "my_protein: end of my_protein \n");
	fprintf(STDOUT, "\n");
	fflush(STDOUT);
}

void print_my_model(struct pe_model * pt_strct)
{
	fprintf(STDOUT, "model: [seq_head, seq_tail]\n");
	fprintf(STDOUT, "model: [%7d, %7d]\n", pt_strct->seq_head,
		pt_strct->seq_tail);
	fflush(STDOUT);
}

void print_my_atom(struct pe_atom * pt_strct)
{
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"atom : ser. name res chn num (       x,        y,        z)    occu. id\n");
	fprintf(STDOUT,
		"atom : %4d %4d %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %8f %.4s\n",
		pt_strct->serial_num, pt_strct->atom_name, pt_strct->res_name,
		pt_strct->chain_id, pt_strct->res_seq, pt_strct->x, pt_strct->y,
		pt_strct->z, pt_strct->occupancy, pt_strct->seg_id);
	fprintf(STDOUT, "\n");
	fflush(STDOUT);
}

    /**************************************/
    /*   atm        seq           mod     */
    /*          atom_head      seq_head   */
    /*          atom_tail      seq_tail   */
    /*           (id)                      */
    /*           (mod_id)                 */
    /*           (mol_id)                  */
    /**************************************/
/*
1 <= cur_mod <= my_protein.mod_num
    my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_head].atom_head]
    my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_head].atom_tail]
        ....
    my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_tail].atom_head]
    my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_tail].atom_tail]

1 <= cur_seq <= my_protein.seq_num
    my_protein.atom[my_protein.sequence[cur_seq].atom_head]
    my_protein.atom[my_protein.sequence[cur_seq].atom_tail]
*/
/*
    for (mod_ct=1; mod_ct <= my_protein.mod_num; mod_ct++) {
        for ( seq_ct=my_protein.model[mod_ct].seq_head; seq_ct <=my_protein.model[mod_ct].seq_tail; seq_ct++) {
            for (atm_ct=my_protein.sequence[seq_ct].atom_head; atm_ct <=my_protein.sequence[seq_ct].atom_tail; atm_ct++) {
            }
        }
    }
*/

/* pt_strct = my_protein.atom */
void print_part_of_my_sequences(struct pe_atom * pt_strct, int atm_head,
				int atm_tail)
{
	int i;

	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "seq. : from %d to %d\n", atm_head, atm_tail);
	fprintf(STDOUT,
		"seq. : [atom] ser. name res chn num  (       x,        y,        z) occu. id\n");

	for (i = atm_head; i <= atm_tail; i++) {
		fprintf(STDOUT,
			"seq. : [%4d] %4d %4d %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s\n",
			i, pt_strct[i].serial_num, pt_strct[i].atom_name,
			pt_strct[i].res_name, pt_strct[i].chain_id,
			pt_strct[i].res_seq, pt_strct[i].x, pt_strct[i].y,
			pt_strct[i].z, pt_strct[i].occupancy,
			pt_strct[i].seg_id);
	}
	fprintf(STDOUT, "seq. : end of sequence\n");
	fprintf(STDOUT, "\n");
	fflush(STDOUT);
}

void print_my_sequences(void)
{
	int mod_ct, seq_ct, atm_ct;

	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "seq. : #sequences= %d, #molecules= %d, #models= %d \n",
		my_protein.seq_num, my_protein.mol_num, my_protein.mod_num);

	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(STDOUT,
				"seq. : #seq= %d (from %d-th atom to %d-th atom)\n",
				seq_ct, my_protein.sequence[seq_ct].atom_head,
				my_protein.sequence[seq_ct].atom_tail);
			fprintf(STDOUT,
				"seq. : [ atom ]  ser.  name res chn num  (       x,        y,        z) occu. id\n");
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				fprintf(STDOUT, "seq. : [%6d] %6d ", atm_ct,
					my_protein.atom[atm_ct].serial_num);
				if (my_protein.atom[atm_ct].atom_name == PE_CC)
					fprintf(STDOUT, " C  ");
				else if (my_protein.atom[atm_ct].atom_name == PE_CA)
					fprintf(STDOUT, " CA ");
				else if (my_protein.atom[atm_ct].atom_name == PE_NN)
					fprintf(STDOUT, " N  ");
				else
					fprintf(STDOUT, " UN ");

				fprintf(STDOUT,
					" %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s\n",
					my_protein.atom[atm_ct].res_name,
					my_protein.atom[atm_ct].chain_id,
					my_protein.atom[atm_ct].res_seq,
					my_protein.atom[atm_ct].x,
					my_protein.atom[atm_ct].y,
					my_protein.atom[atm_ct].z,
					my_protein.atom[atm_ct].occupancy,
					my_protein.atom[atm_ct].seg_id);
				/*  
				   fprintf(STDOUT,"seq. : [%4d] %4d %4d %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s\n",
				   atm_ct,
				   my_protein.atom[atm_ct].serial_num,
				   my_protein.atom[atm_ct].atom_name,
				   my_protein.atom[atm_ct].res_name,
				   my_protein.atom[atm_ct].chain_id,
				   my_protein.atom[atm_ct].res_seq,
				   my_protein.atom[atm_ct].x,
				   my_protein.atom[atm_ct].y,
				   my_protein.atom[atm_ct].z,
				   my_protein.atom[atm_ct].occupancy,
				   my_protein.atom[atm_ct].seg_id);
				 */
			}
			fprintf(STDOUT, "seq. : \n");
		}
	}
	fprintf(STDOUT, "seq. : end of sequence\n");
	fprintf(STDOUT, "\n");
	fflush(STDOUT);
}

void print_tile(int i, struct pe_slt_tile * s_tiles)
{	
	fprintf(STDOUT, "slt_tiles[%d]: tile= x^%d y^%d z^%d w^%d %s,",
		i, s_tiles[i].top[0], s_tiles[i].top[1],
		s_tiles[i].top[2], s_tiles[i].top[3],
		my_trn_table[s_tiles[i].type].str);
	if (s_tiles[i].code == C_UP)
		fprintf(STDOUT, " code= U\n");
	else if (s_tiles[i].code == C_DOWN)
		fprintf(STDOUT, " code= D\n");
	else
		fprintf(STDOUT, " code= X\n");
	fprintf(STDOUT,
		"slt_tiles[%d]:     center=   (%8.3f, %8.3f, %8.3f)\n",
		i, s_tiles[i].cx, s_tiles[i].cy, s_tiles[i].cz);
	fprintf(STDOUT,
		"slt_tiles[%d]:     top=      (%8.3f, %8.3f, %8.3f)\n",
		i, s_tiles[i].tx, s_tiles[i].ty, s_tiles[i].tz);
	fprintf(STDOUT,
		"slt_tiles[%d]:     bottom=   (%8.3f, %8.3f, %8.3f)\n",
		i, s_tiles[i].bx, s_tiles[i].by, s_tiles[i].bz);
	fprintf(STDOUT,
		"slt_tiles[%d]:     direction= (%8.3f, %8.3f, %8.3f)\n",
		i, s_tiles[i].dx, s_tiles[i].dy, s_tiles[i].dz);
	fprintf(STDOUT,
		"slt_tiles[%d]:     error= %f against %d-th atom ", i,
		s_tiles[i].error, s_tiles[i].error_atom);
	if (my_protein.atom[s_tiles[i].error_atom].atom_name == PE_CC)
		fprintf(STDOUT, "(C)");
	else if (my_protein.atom[s_tiles[i].error_atom].atom_name == PE_CA)
		fprintf(STDOUT, "(CA)");
	else if (my_protein.atom[s_tiles[i].error_atom].atom_name == PE_NN)
		fprintf(STDOUT, "(N)");
	else
		fprintf(STDOUT, "(type unknown) ");
	fprintf(STDOUT, "    [flag= %d]\n", s_tiles[i].atom_flag);
	return;
}

void print_my_slt_tiles(struct pe_slt_tile * s_tiles)
{
	int i, seq_ct;

	if (my_protein.tile_num > my_protein.tile_max_num) {
		fprintf(STDOUT,
			"slt_tiles[  ]: ERROR    my_protein.tile_num= %d > my_protein.tile_max_num= %d\n",
			my_protein.tile_num, my_protein.tile_max_num);
		my_protein.tile_num = my_protein.tile_max_num;
	}

	seq_ct = 1;
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"slt_tiles[  ]: Slant tile table BEGIN (tile num= %d)\n",
		my_protein.tile_num);
	fprintf(STDOUT,
		"slt_tiles[  ]: --------------  sequence %d  --------------------------------------\n",
		seq_ct);
	for (i = 1; i <= my_protein.tile_num; i++) {
		print_tile(i, s_tiles);
		if ((s_tiles[i].ter > 0) && (i < my_protein.tile_num)) {	/* end of a sequence */
			seq_ct++;
			fprintf(STDOUT, "\n");
			fprintf(STDOUT,
				"slt_tiles[  ]: --------------  sequence %d  --------------------------------------\n",
				seq_ct);
		}
	}
	fprintf(STDOUT,
		"slt_tiles[  ]: --------------------------------------------------------------------\n");
	fprintf(STDOUT, "slt_tiles[  ]: Slant tile table END \n");
	fprintf(STDOUT, "\n");
	fflush(STDOUT);

	return;
}

void print_my_coordinates(void)
{
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[Coordinates]:  Origin = ( %8.3f, %8.3f, %8.3f ) \n",
		my_coor.origin.x, my_coor.origin.y, my_coor.origin.z);
	fprintf(STDOUT, "[Coordinates]:  x-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		my_coor.axes[0].x, my_coor.axes[0].y, my_coor.axes[0].z);
	fprintf(STDOUT, "[Coordinates]:  y-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		my_coor.axes[1].x, my_coor.axes[1].y, my_coor.axes[1].z);
	fprintf(STDOUT, "[Coordinates]:  z-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		my_coor.axes[2].x, my_coor.axes[2].y, my_coor.axes[2].z);
	fprintf(STDOUT, "[Coordinates]:  w-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		my_coor.axes[3].x, my_coor.axes[3].y, my_coor.axes[3].z);
	fflush(STDOUT);
	return;
}
