/*
 *  pe_trn_table.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/01/05.
 *  Copyright (C) 2005, 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern struct pe_protein my_protein;
extern struct pe_coor_system my_coor;
extern struct pe_settings my_settings;

struct pe_trn_table my_trn_table[24];
struct pe_trn_table my_trn_table_save[24];

int ston[256];
struct pe_qtos_table qtos[24];

void print_trn_table(void)
{
	int i;

	fprintf(STDOUT, "Transition table: Dump list starts\n");
	for (i = 0; i < 24; i++) {
		fprintf(STDOUT,
			"my_trn_table[%2d]:-------------------------------------------------------------------\n",
			i);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (current tile) type= 1%s, center= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].str, my_trn_table[i].cx,
			my_trn_table[i].cy, my_trn_table[i].cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (current tile)            bottom= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].bx, my_trn_table[i].by,
			my_trn_table[i].bz);

		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward tile) center= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_cx,
			my_trn_table[i].downward_cy,
			my_trn_table[i].downward_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward&&Up) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].downward_U_type].str,
			my_trn_table[i].downward_U_top[0],
			my_trn_table[i].downward_U_top[1],
			my_trn_table[i].downward_U_top[2],
			my_trn_table[i].downward_U_top[3],
			my_trn_table[i].downward_U_tx,
			my_trn_table[i].downward_U_ty,
			my_trn_table[i].downward_U_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward&&Up)             bottom vertex= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_U_bx,
			my_trn_table[i].downward_U_by,
			my_trn_table[i].downward_U_bz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_U_next_cx,
			my_trn_table[i].downward_U_next_cy,
			my_trn_table[i].downward_U_next_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward&&Down) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].downward_D_type].str,
			my_trn_table[i].downward_D_top[0],
			my_trn_table[i].downward_D_top[1],
			my_trn_table[i].downward_D_top[2],
			my_trn_table[i].downward_D_top[3],
			my_trn_table[i].downward_D_tx, my_trn_table[i].downward_D_ty,
			my_trn_table[i].downward_D_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward&&Down)             bottom vertex= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_D_bx,
			my_trn_table[i].downward_D_by,
			my_trn_table[i].downward_D_bz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_D_next_cx,
			my_trn_table[i].downward_D_next_cy,
			my_trn_table[i].downward_D_next_cz);

		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward tile) center= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_cx, my_trn_table[i].upward_cy,
			my_trn_table[i].upward_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward&&Up) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].upward_U_type].str,
			my_trn_table[i].upward_U_top[0],
			my_trn_table[i].upward_U_top[1],
			my_trn_table[i].upward_U_top[2],
			my_trn_table[i].upward_U_top[3],
			my_trn_table[i].upward_U_tx,
			my_trn_table[i].upward_U_ty,
			my_trn_table[i].upward_U_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward&&Up)             bottom vertex= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_U_bx,
			my_trn_table[i].upward_U_by,
			my_trn_table[i].upward_U_bz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_U_next_cx,
			my_trn_table[i].upward_U_next_cy,
			my_trn_table[i].upward_U_next_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward&&Down) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].upward_D_type].str,
			my_trn_table[i].upward_D_top[0],
			my_trn_table[i].upward_D_top[1],
			my_trn_table[i].upward_D_top[2],
			my_trn_table[i].upward_D_top[3],
			my_trn_table[i].upward_D_tx,
			my_trn_table[i].upward_D_ty,
			my_trn_table[i].upward_D_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward&&Down)             bottom vertex= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_D_bx,
			my_trn_table[i].upward_D_by,
			my_trn_table[i].upward_D_bz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_D_next_cx,
			my_trn_table[i].upward_D_next_cy,
			my_trn_table[i].upward_D_next_cz);
		fflush(STDOUT);
	}			/* for (i=0; i<24; i++) { */
	fprintf(STDOUT, "Transition table: Dump list ends.\n");
	return;
}

void copy_pe_trn_table(struct pe_trn_table * ptgt_trn_table,
		       struct pe_trn_table * psrc_trn_table)
{
	int tile_type, i;
	
	for (tile_type = 0; tile_type < 24; tile_type++) {
		strcpy(ptgt_trn_table[tile_type].str,
		       psrc_trn_table[tile_type].str);
		ptgt_trn_table[tile_type].cx = psrc_trn_table[tile_type].cx;
		ptgt_trn_table[tile_type].cy = psrc_trn_table[tile_type].cy;
		ptgt_trn_table[tile_type].cz = psrc_trn_table[tile_type].cz;
		ptgt_trn_table[tile_type].bx = psrc_trn_table[tile_type].bx;
		ptgt_trn_table[tile_type].by = psrc_trn_table[tile_type].by;
		ptgt_trn_table[tile_type].bz = psrc_trn_table[tile_type].bz;
		
		ptgt_trn_table[tile_type].downward_cx
			= psrc_trn_table[tile_type].downward_cx;
		ptgt_trn_table[tile_type].downward_cy
			= psrc_trn_table[tile_type].downward_cy;
		ptgt_trn_table[tile_type].downward_cz
			= psrc_trn_table[tile_type].downward_cz;
		
		ptgt_trn_table[tile_type].downward_U_next_cx
			= psrc_trn_table[tile_type].downward_U_next_cx;
		ptgt_trn_table[tile_type].downward_U_next_cy
			= psrc_trn_table[tile_type].downward_U_next_cy;
		ptgt_trn_table[tile_type].downward_U_next_cz
			= psrc_trn_table[tile_type].downward_U_next_cz;
		ptgt_trn_table[tile_type].downward_U_tx
			= psrc_trn_table[tile_type].downward_U_tx;
		ptgt_trn_table[tile_type].downward_U_ty
			= psrc_trn_table[tile_type].downward_U_ty;
		ptgt_trn_table[tile_type].downward_U_tz
			= psrc_trn_table[tile_type].downward_U_tz;
		ptgt_trn_table[tile_type].downward_U_bx
			= psrc_trn_table[tile_type].downward_U_bx;
		ptgt_trn_table[tile_type].downward_U_by
			= psrc_trn_table[tile_type].downward_U_by;
		ptgt_trn_table[tile_type].downward_U_bz
			= psrc_trn_table[tile_type].downward_U_bz;
		for (i = 0; i < 4; i++)
			ptgt_trn_table[tile_type].downward_U_top[i]
				= psrc_trn_table[tile_type].downward_U_top[i];
		ptgt_trn_table[tile_type].downward_U_type
			= psrc_trn_table[tile_type].downward_U_type;
		
		ptgt_trn_table[tile_type].downward_D_next_cx
			= psrc_trn_table[tile_type].downward_D_next_cx;
		ptgt_trn_table[tile_type].downward_D_next_cy
			= psrc_trn_table[tile_type].downward_D_next_cy;
		ptgt_trn_table[tile_type].downward_D_next_cz
			= psrc_trn_table[tile_type].downward_D_next_cz;
		ptgt_trn_table[tile_type].downward_D_tx
			= psrc_trn_table[tile_type].downward_D_tx;
		ptgt_trn_table[tile_type].downward_D_ty
			= psrc_trn_table[tile_type].downward_D_ty;
		ptgt_trn_table[tile_type].downward_D_tz
			= psrc_trn_table[tile_type].downward_D_tz;
		ptgt_trn_table[tile_type].downward_D_bx
			= psrc_trn_table[tile_type].downward_D_bx;
		ptgt_trn_table[tile_type].downward_D_by
			= psrc_trn_table[tile_type].downward_D_by;
		ptgt_trn_table[tile_type].downward_D_bz
			= psrc_trn_table[tile_type].downward_D_bz;
		for (i = 0; i < 4; i++)
			ptgt_trn_table[tile_type].downward_D_top[i]
				= psrc_trn_table[tile_type].downward_D_top[i];
		ptgt_trn_table[tile_type].downward_D_type
			= psrc_trn_table[tile_type].downward_D_type;
		
		ptgt_trn_table[tile_type].upward_cx
			= psrc_trn_table[tile_type].upward_cx;
		ptgt_trn_table[tile_type].upward_cy
			= psrc_trn_table[tile_type].upward_cy;
		ptgt_trn_table[tile_type].upward_cz
			= psrc_trn_table[tile_type].upward_cz;
		
		ptgt_trn_table[tile_type].upward_U_next_cx
			= psrc_trn_table[tile_type].upward_U_next_cx;
		ptgt_trn_table[tile_type].upward_U_next_cy
			= psrc_trn_table[tile_type].upward_U_next_cy;
		ptgt_trn_table[tile_type].upward_U_next_cz
			= psrc_trn_table[tile_type].upward_U_next_cz;
		ptgt_trn_table[tile_type].upward_U_tx
			= psrc_trn_table[tile_type].upward_U_tx;
		ptgt_trn_table[tile_type].upward_U_ty
			= psrc_trn_table[tile_type].upward_U_ty;
		ptgt_trn_table[tile_type].upward_U_tz
			= psrc_trn_table[tile_type].upward_U_tz;
		ptgt_trn_table[tile_type].upward_U_bx
			= psrc_trn_table[tile_type].upward_U_bx;
		ptgt_trn_table[tile_type].upward_U_by
			= psrc_trn_table[tile_type].upward_U_by;
		ptgt_trn_table[tile_type].upward_U_bz
			= psrc_trn_table[tile_type].upward_U_bz;
		for (i = 0; i < 4; i++)
			ptgt_trn_table[tile_type].upward_U_top[i]
				= psrc_trn_table[tile_type].upward_U_top[i];
		ptgt_trn_table[tile_type].upward_U_type
			= psrc_trn_table[tile_type].upward_U_type;
		
		ptgt_trn_table[tile_type].upward_D_next_cx
			= psrc_trn_table[tile_type].upward_D_next_cx;
		ptgt_trn_table[tile_type].upward_D_next_cy
			= psrc_trn_table[tile_type].upward_D_next_cy;
		ptgt_trn_table[tile_type].upward_D_next_cz
			= psrc_trn_table[tile_type].upward_D_next_cz;
		ptgt_trn_table[tile_type].upward_D_tx
			= psrc_trn_table[tile_type].upward_D_tx;
		ptgt_trn_table[tile_type].upward_D_ty
			= psrc_trn_table[tile_type].upward_D_ty;
		ptgt_trn_table[tile_type].upward_D_tz
			= psrc_trn_table[tile_type].upward_D_tz;
		ptgt_trn_table[tile_type].upward_D_bx
			= psrc_trn_table[tile_type].upward_D_bx;
		ptgt_trn_table[tile_type].upward_D_by
			= psrc_trn_table[tile_type].upward_D_by;
		ptgt_trn_table[tile_type].upward_D_bz
			= psrc_trn_table[tile_type].upward_D_bz;
		for (i = 0; i < 4; i++)
			ptgt_trn_table[tile_type].upward_D_top[i]
				= psrc_trn_table[tile_type].upward_D_top[i];
		ptgt_trn_table[tile_type].upward_D_type
			= psrc_trn_table[tile_type].upward_D_type;
	}
	return;
}

/* Set up the table for conversion from strings to integeres. */
void setup_string_to_integer_table(void)
{
	int i;
	
	for (i = 0; i < 256; i++)
		ston[i] = -1;
	
	ston[INDX(XX, YY, ZZ, WW)] = XYZW;
	ston[INDX(XX, ZZ, YY, WW)] = XZYW;
	ston[INDX(YY, XX, ZZ, WW)] = YXZW;
	ston[INDX(YY, ZZ, XX, WW)] = YZXW;
	ston[INDX(ZZ, XX, YY, WW)] = ZXYW;
	ston[INDX(ZZ, YY, XX, WW)] = ZYXW;
	
	ston[INDX(WW, XX, YY, ZZ)] = WXYZ;
	ston[INDX(YY, WW, XX, ZZ)] = YWXZ;
	ston[INDX(WW, YY, XX, ZZ)] = WYXZ;
	ston[INDX(XX, WW, YY, ZZ)] = XWYZ;
	ston[INDX(XX, YY, WW, ZZ)] = XYWZ;
	ston[INDX(YY, XX, WW, ZZ)] = YXWZ;
	
	ston[INDX(ZZ, WW, XX, YY)] = ZWXY;
	ston[INDX(WW, XX, ZZ, YY)] = WXZY;
	ston[INDX(XX, ZZ, WW, YY)] = XZWY;
	ston[INDX(ZZ, XX, WW, YY)] = ZXWY;
	ston[INDX(WW, ZZ, XX, YY)] = WZXY;
	ston[INDX(XX, WW, ZZ, YY)] = XWZY;
	
	ston[INDX(YY, ZZ, WW, XX)] = YZWX;
	ston[INDX(ZZ, YY, WW, XX)] = ZYWX;
	ston[INDX(ZZ, WW, YY, XX)] = ZWYX;
	ston[INDX(WW, YY, ZZ, XX)] = WYZX;
	ston[INDX(YY, WW, ZZ, XX)] = YWZX;
	ston[INDX(WW, ZZ, YY, XX)] = WZYX;
	return;
}

void setup_coordinates_to_tiletype_table(void)
{
	qtos[XYZW].p = XX;
	qtos[XYZW].q = YY;
	qtos[XYZW].r = ZZ;
	qtos[XYZW].s = WW;
	strcpy(qtos[XYZW].str, "[xyz]");
	qtos[XZYW].p = XX;
	qtos[XZYW].q = ZZ;
	qtos[XZYW].r = YY;
	qtos[XZYW].s = WW;
	strcpy(qtos[XZYW].str, "[xzy]");
	qtos[YXZW].p = YY;
	qtos[YXZW].q = XX;
	qtos[YXZW].r = ZZ;
	qtos[YXZW].s = WW;
	strcpy(qtos[YXZW].str, "[yxz]");
	qtos[YZXW].p = YY;
	qtos[YZXW].q = ZZ;
	qtos[YZXW].r = XX;
	qtos[YZXW].s = WW;
	strcpy(qtos[YZXW].str, "[yzx]");
	qtos[ZXYW].p = ZZ;
	qtos[ZXYW].q = XX;
	qtos[ZXYW].r = YY;
	qtos[ZXYW].s = WW;
	strcpy(qtos[ZXYW].str, "[zxy]");
	qtos[ZYXW].p = ZZ;
	qtos[ZYXW].q = YY;
	qtos[ZYXW].r = XX;
	qtos[ZYXW].s = WW;
	strcpy(qtos[ZYXW].str, "[zyx]");
	
	qtos[WXYZ].p = WW;
	qtos[WXYZ].q = XX;
	qtos[WXYZ].r = YY;
	qtos[WXYZ].s = ZZ;
	strcpy(qtos[WXYZ].str, "[wxy]");
	qtos[YWXZ].p = YY;
	qtos[YWXZ].q = WW;
	qtos[YWXZ].r = XX;
	qtos[YWXZ].s = ZZ;
	strcpy(qtos[YWXZ].str, "[ywx]");
	qtos[WYXZ].p = WW;
	qtos[WYXZ].q = YY;
	qtos[WYXZ].r = XX;
	qtos[WYXZ].s = ZZ;
	strcpy(qtos[WYXZ].str, "[wyx]");
	qtos[XWYZ].p = XX;
	qtos[XWYZ].q = WW;
	qtos[XWYZ].r = YY;
	qtos[XWYZ].s = ZZ;
	strcpy(qtos[XWYZ].str, "[xwy]");
	qtos[XYWZ].p = XX;
	qtos[XYWZ].q = YY;
	qtos[XYWZ].r = WW;
	qtos[XYWZ].s = ZZ;
	strcpy(qtos[XYWZ].str, "[xyw]");
	qtos[YXWZ].p = YY;
	qtos[YXWZ].q = XX;
	qtos[YXWZ].r = WW;
	qtos[YXWZ].s = ZZ;
	strcpy(qtos[YXWZ].str, "[yxw]");
	
	qtos[ZWXY].p = ZZ;
	qtos[ZWXY].q = WW;
	qtos[ZWXY].r = XX;
	qtos[ZWXY].s = YY;
	strcpy(qtos[ZWXY].str, "[zwx]");
	qtos[WXZY].p = WW;
	qtos[WXZY].q = XX;
	qtos[WXZY].r = ZZ;
	qtos[WXZY].s = YY;
	strcpy(qtos[WXZY].str, "[wxz]");
	qtos[XZWY].p = XX;
	qtos[XZWY].q = ZZ;
	qtos[XZWY].r = WW;
	qtos[XZWY].s = YY;
	strcpy(qtos[XZWY].str, "[xzw]");
	qtos[ZXWY].p = ZZ;
	qtos[ZXWY].q = XX;
	qtos[ZXWY].r = WW;
	qtos[ZXWY].s = YY;
	strcpy(qtos[ZXWY].str, "[zxw]");
	qtos[WZXY].p = WW;
	qtos[WZXY].q = ZZ;
	qtos[WZXY].r = XX;
	qtos[WZXY].s = YY;
	strcpy(qtos[WZXY].str, "[wzx]");
	qtos[XWZY].p = XX;
	qtos[XWZY].q = WW;
	qtos[XWZY].r = ZZ;
	qtos[XWZY].s = YY;
	strcpy(qtos[XWZY].str, "[xwz]");
	
	qtos[YZWX].p = YY;
	qtos[YZWX].q = ZZ;
	qtos[YZWX].r = WW;
	qtos[YZWX].s = XX;
	strcpy(qtos[YZWX].str, "[yzw]");
	qtos[ZYWX].p = ZZ;
	qtos[ZYWX].q = YY;
	qtos[ZYWX].r = WW;
	qtos[ZYWX].s = XX;
	strcpy(qtos[ZYWX].str, "[zyw]");
	qtos[ZWYX].p = ZZ;
	qtos[ZWYX].q = WW;
	qtos[ZWYX].r = YY;
	qtos[ZWYX].s = XX;
	strcpy(qtos[ZWYX].str, "[zwy]");
	qtos[WYZX].p = WW;
	qtos[WYZX].q = YY;
	qtos[WYZX].r = ZZ;
	qtos[WYZX].s = XX;
	strcpy(qtos[WYZX].str, "[wyz]");
	qtos[YWZX].p = YY;
	qtos[YWZX].q = WW;
	qtos[YWZX].r = ZZ;
	qtos[YWZX].s = XX;
	strcpy(qtos[YWZX].str, "[ywz]");
	qtos[WZYX].p = WW;
	qtos[WZYX].q = ZZ;
	qtos[WZYX].r = YY;
	qtos[WZYX].s = XX;
	strcpy(qtos[WZYX].str, "[wzy]");
	
	if (0) {
		int i;
		
		for (i = 0; i < 24; i++) {
			if (ston[INDX(qtos[i].p, qtos[i].q, qtos[i].r,
				      qtos[i].s)] != i) {
				fprintf(STDERR,
"setup_coordinates_to_tiletype_table() : ERROR  ston-qtos inconsistence\n");
				exit(1);
			}
		}
	}
	return;
}

void set_trn_table_centre_field(int i /* tile_type */)
{
	float VX_x, VX_y, VX_z;
	float VY_x, VY_y, VY_z;
	float VZ_x, VZ_y, VZ_z;
	float VW_x, VW_y, VW_z;
	
	VX_x = my_coor.axes[qtos[i].p].x;
	VX_y = my_coor.axes[qtos[i].p].y;
	VX_z = my_coor.axes[qtos[i].p].z;
	VY_x = my_coor.axes[qtos[i].q].x;
	VY_y = my_coor.axes[qtos[i].q].y;
	VY_z = my_coor.axes[qtos[i].q].z;
	VZ_x = my_coor.axes[qtos[i].r].x;
	VZ_y = my_coor.axes[qtos[i].r].y;
	VZ_z = my_coor.axes[qtos[i].r].z;
	VW_x = my_coor.axes[qtos[i].s].x;
	VW_y = my_coor.axes[qtos[i].s].y;
	VW_z = my_coor.axes[qtos[i].s].z;
	
	my_trn_table[i].cx = (3 * VX_x + 2 * VY_x + VZ_x) / 4;
	my_trn_table[i].cy = (3 * VX_y + 2 * VY_y + VZ_y) / 4;
	my_trn_table[i].cz = (3 * VX_z + 2 * VY_z + VZ_z) / 4;
	my_trn_table[i].bx = VX_x + VY_x + VZ_x;	/* == -VW */
	my_trn_table[i].by = VX_y + VY_y + VZ_y;
	my_trn_table[i].bz = VX_z + VY_z + VZ_z;
	return;
}

void set_trn_table_downward_field(int i /* tile_type */)
{
	float VX_x, VX_y, VX_z;
	float VY_x, VY_y, VY_z;
	float VZ_x, VZ_y, VZ_z;
	float VW_x, VW_y, VW_z;
	
	VX_x = my_coor.axes[qtos[i].p].x;
	VX_y = my_coor.axes[qtos[i].p].y;
	VX_z = my_coor.axes[qtos[i].p].z;
	VY_x = my_coor.axes[qtos[i].q].x;
	VY_y = my_coor.axes[qtos[i].q].y;
	VY_z = my_coor.axes[qtos[i].q].z;
	VZ_x = my_coor.axes[qtos[i].r].x;
	VZ_y = my_coor.axes[qtos[i].r].y;
	VZ_z = my_coor.axes[qtos[i].r].z;
	VW_x = my_coor.axes[qtos[i].s].x;
	VW_y = my_coor.axes[qtos[i].s].y;
	VW_z = my_coor.axes[qtos[i].s].z;
	
	/* |(x/w)[wyz]| = |x[yzx]| */
	my_trn_table[i].downward_cx = (5 * VX_x + 3 * VY_x + 2 * VZ_x) / 4;
	my_trn_table[i].downward_cy = (5 * VX_y + 3 * VY_y + 2 * VZ_y) / 4;
	my_trn_table[i].downward_cz = (5 * VX_z + 3 * VY_z + 2 * VZ_z) / 4;
	/* |(x/w)[wyx]| */
	my_trn_table[i].downward_U_next_cx = (5 * VX_x + 2 * VY_x - VW_x) / 4;
	my_trn_table[i].downward_U_next_cy = (5 * VX_y + 2 * VY_y - VW_y) / 4;
	my_trn_table[i].downward_U_next_cz = (5 * VX_z + 2 * VY_z - VW_z) / 4;
	/* x/w[wyz] */
	my_trn_table[i].downward_U_tx = VX_x - VW_x;
	my_trn_table[i].downward_U_ty = VX_y - VW_y;
	my_trn_table[i].downward_U_tz = VX_z - VW_z;
	/* x/w + wyz = xyz */
	my_trn_table[i].downward_U_bx = VX_x + VY_x + VZ_x;
	my_trn_table[i].downward_U_by = VX_y + VY_y + VZ_y;
	my_trn_table[i].downward_U_bz = VX_z + VY_z + VZ_z;
	/* x/w */
	my_trn_table[i].downward_U_top[qtos[i].p] = 1;
	my_trn_table[i].downward_U_top[qtos[i].q] = 0;
	my_trn_table[i].downward_U_top[qtos[i].r] = 0;
	my_trn_table[i].downward_U_top[qtos[i].s] = -1;
	/* [wyz] */
	my_trn_table[i].downward_U_type
		= ston[INDX(qtos[i].s, qtos[i].q, qtos[i].r, qtos[i].p)];
	
	/* |xy[zxy]| */
	my_trn_table[i].downward_D_next_cx
		= (6 * VX_x + 5 * VY_x + 3 * VZ_x) / 4;
	my_trn_table[i].downward_D_next_cy
		= (6 * VX_y + 5 * VY_y + 3 * VZ_y) / 4;
	my_trn_table[i].downward_D_next_cz
		= (6 * VX_z + 5 * VY_z + 3 * VZ_z) / 4;
	/* x[yzx] */
	my_trn_table[i].downward_D_tx = VX_x;
	my_trn_table[i].downward_D_ty = VX_y;
	my_trn_table[i].downward_D_tz = VX_z;
	/* x + yzx = xxyz */
	my_trn_table[i].downward_D_bx = 2 * VX_x + VY_x + VZ_x;
	my_trn_table[i].downward_D_by = 2 * VX_y + VY_y + VZ_y;
	my_trn_table[i].downward_D_bz = 2 * VX_z + VY_z + VZ_z;
	/* x */
	my_trn_table[i].downward_D_top[qtos[i].p] = 1;
	my_trn_table[i].downward_D_top[qtos[i].q] = 0;
	my_trn_table[i].downward_D_top[qtos[i].r] = 0;
	my_trn_table[i].downward_D_top[qtos[i].s] = 0;
	/* [yzx] */
	my_trn_table[i].downward_D_type = ston[INDX(qtos[i].q, qtos[i].r, qtos[i].p, qtos[i].s)];
	return;
}

void set_trn_table_upward_field(int i /* tile_type */)
{
	float VX_x, VX_y, VX_z;
	float VY_x, VY_y, VY_z;
	float VZ_x, VZ_y, VZ_z;
	float VW_x, VW_y, VW_z;
	
	VX_x = my_coor.axes[qtos[i].p].x;
	VX_y = my_coor.axes[qtos[i].p].y;
	VX_z = my_coor.axes[qtos[i].p].z;
	VY_x = my_coor.axes[qtos[i].q].x;
	VY_y = my_coor.axes[qtos[i].q].y;
	VY_z = my_coor.axes[qtos[i].q].z;
	VZ_x = my_coor.axes[qtos[i].r].x;
	VZ_y = my_coor.axes[qtos[i].r].y;
	VZ_z = my_coor.axes[qtos[i].r].z;
	VW_x = my_coor.axes[qtos[i].s].x;
	VW_y = my_coor.axes[qtos[i].s].y;
	VW_z = my_coor.axes[qtos[i].s].z;
	
	/* |1/z[zxy]|=|1[xyw]| */
	my_trn_table[i].upward_cx = (3 * VX_x + 2 * VY_x + VW_x) / 4;
	my_trn_table[i].upward_cy = (3 * VX_y + 2 * VY_y + VW_y) / 4;
	my_trn_table[i].upward_cz = (3 * VX_z + 2 * VY_z + VW_z) / 4;
	/* |1/z[zxw]| */
	my_trn_table[i].upward_U_next_cx = (2 * VX_x - VZ_x + VW_x) / 4;
	my_trn_table[i].upward_U_next_cy = (2 * VX_y - VZ_y + VW_y) / 4;
	my_trn_table[i].upward_U_next_cz = (2 * VX_z - VZ_z + VW_z) / 4;
	/* 1/z[zxy] */
	my_trn_table[i].upward_U_tx = - VZ_x;
	my_trn_table[i].upward_U_ty = - VZ_y;
	my_trn_table[i].upward_U_tz = - VZ_z;
	/* 1/z + zxy = xy */
	my_trn_table[i].upward_U_bx = VX_x + VY_x;
	my_trn_table[i].upward_U_by = VX_y + VY_y;
	my_trn_table[i].upward_U_bz = VX_z + VY_z;
	/* 1/z */
	my_trn_table[i].upward_U_top[qtos[i].p] = 0;
	my_trn_table[i].upward_U_top[qtos[i].q] = 0;
	my_trn_table[i].upward_U_top[qtos[i].r] = -1;
	my_trn_table[i].upward_U_top[qtos[i].s] = 0;
	/* [zxy] */
	my_trn_table[i].upward_U_type
		= ston[INDX(qtos[i].r, qtos[i].p, qtos[i].q, qtos[i].s)];
	/* |x[ywx]| */
	my_trn_table[i].upward_D_next_cx
		= (5 * VX_x + 3 * VY_x + 2 * VW_x) / 4;
	my_trn_table[i].upward_D_next_cy
		= (5 * VX_y + 3 * VY_y + 2 * VW_y) / 4;
	my_trn_table[i].upward_D_next_cz
		= (5 * VX_z + 3 * VY_z + 2 * VW_z) / 4;
	/* 1[xyw] */
	my_trn_table[i].upward_D_tx = 0.0;
	my_trn_table[i].upward_D_ty = 0.0;
	my_trn_table[i].upward_D_tz = 0.0;
	/* 1 + xyw = xyw */
	my_trn_table[i].upward_D_bx = VX_x + VY_x + VW_x;
	my_trn_table[i].upward_D_by = VX_y + VY_y + VW_y;
	my_trn_table[i].upward_D_bz = VX_z + VY_z + VW_z;
	/* 1 */
	my_trn_table[i].upward_D_top[qtos[i].p] = 0;
	my_trn_table[i].upward_D_top[qtos[i].q] = 0;
	my_trn_table[i].upward_D_top[qtos[i].r] = 0;
	my_trn_table[i].upward_D_top[qtos[i].s] = 0;
	/* [xyw] */
	my_trn_table[i].upward_D_type
		= ston[INDX(qtos[i].p, qtos[i].q, qtos[i].s, qtos[i].r)];
	return;
}

void check_trn_table(float ave_dis)
{
	int flag;
	int i; /* tile_type */
	int nxt_type;
	float wrk_float;
	
	flag = 0;
	for (i = 0; i < 24; i++) {
		/*downward && U */
		nxt_type = my_trn_table[i].downward_U_type;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].downward_U_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].downward_U_next_cx
			- my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].downward_U_next_cy
			- my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].downward_U_next_cz
			- my_trn_table[i].downward_U_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type]. downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].downward_U_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type].downward_U_next_cx
			- wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_U_next_cy
			- wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].downward_U_tz;
		wrk_float
		= my_trn_table[nxt_type].downward_U_next_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type].downward_U_type != i)
			flag = flag + 10000;
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR  - my_trn_table[%d], downward&&U (flag=%d) -\n",
				i, flag);
			exit(1);
		}
		
		/*downward && D */
		nxt_type = my_trn_table[i].downward_D_type;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].downward_D_next_cx
			- my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].downward_D_next_cy
			- my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].downward_D_next_cz -
			my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_U_next_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_U_next_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_U_next_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type]. upward_U_type != i)
			flag = flag + 10000;
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR   my_trn_table[%d], downward&&D (flag=%d) -\n",
				i, flag);
			exit(1);
		}
		
		/*upward && U */
		nxt_type = my_trn_table[i].upward_U_type;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].upward_U_next_cx
			- my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].upward_U_next_cy
			- my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].upward_U_next_cz
			- my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].downward_D_next_cx
			- wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_D_next_cy
			- wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].downward_D_next_cz
			- wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type].downward_D_type != i)
			flag = flag + 10000;
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR  - my_trn_table[%d], upward&&U (flag=%d) -\n",
				i, flag);
			exit(1);
		}
		
		/*upward && D */
		nxt_type = my_trn_table[i].upward_D_type;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].upward_D_next_cx
			- my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].upward_D_next_cy
			- my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].upward_D_next_cz -
			my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_D_next_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_D_next_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_D_next_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float, ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type].upward_D_type != i) {
			flag = flag + 10000;
		}
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR   my_trn_table[%d], upward&&D (flag=%d)\n",
				i, flag);
			exit(1);
		}		
	}
	return;
}

/* ston[ INDX(XX,YY,ZZ,WW) ] = XYZW */
/* qtos[XYZW].p =XX; qtos[XYZW].q =YY; qtos[XYZW].r =ZZ; qtos[XYZW].s =WW; */
/* qtos[XYZW].str = "[xyz]" */
/* ston[INDX(qtos[i].p, qtos[i].q, qtos[i].r, qtos[i].s)] == i */
void setup_transition_table(float ave_dis)
{
	int tile_type;

	setup_string_to_integer_table();	/* ston[] */
	setup_coordinates_to_tiletype_table();	/* qtos[] */

	/* Set up the transition table */
	for (tile_type = 0; tile_type < 24; tile_type++) {
		strcpy(my_trn_table[tile_type].str, qtos[tile_type].str);
		set_trn_table_centre_field(tile_type);
		set_trn_table_downward_field(tile_type);
		set_trn_table_upward_field(tile_type);
	}
	if (my_settings.rotate_tile == TRUE)
		copy_pe_trn_table(my_trn_table_save, my_trn_table);

	if (0) check_trn_table(ave_dis);

	return;
}
