/*
 *  pe_print_protein_code.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/9/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-03 add print_protein_ntile_code()
 *  2006-11-03 add print_ntcode(), ntile_to_char(), resname_to_char()
 *  2006-11-03 add ntile_to_atom_name()
 *  2006-11-03 support PE_CR, PE_CY, PE_CW
 *  2006-11-20 version 0.2 release
 *  2007-01-16 permit the surfix of ".pdb"
 *  2007-02-15 version 0.3 release
 *  2007-04-09 return PE_CG in the case of 'G' in ntile_to_atom_name()
 *  2007-04-09 support CO, CB, CG, CU, and CV in print_protein_code_pdb_data()
 *  2007-04-15 version 0.4 release
 *  2007-04-16 return PE_CO in the case of 'R' in ntile_to_atom_name()
 *  2007-04-16 version 0.4a release
 *  2007-07-27 return PE_CY in the case of 'R' in ntile_to_atom_name()
 *             return PE_CO in the case of 'B' or 'Q' in ntile_to_atom_name()
 *  2007-07-27 version 0.4b release
 *  2007-08-22 option "R" is supported in resname_to_char()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_pdb_data.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;
extern struct pe_trn_table my_trn_table[];
extern struct pe_coor_system my_coor;

PE_ATOM_NAME ntile_to_atom_name(struct pe_protein * pt_protein, int atm_ind)
{
	int half_length;
	int top_of_5tile;
	char work[5];
	int i;
	
	if (pt_protein->atom[atm_ind].ntile_length < 5)
		return PE_CA;
	half_length = (pt_protein->atom[atm_ind].ntile_length - 1) / 2;
	top_of_5tile = pt_protein->atom[atm_ind].ntile_head + half_length - 2;
	for (i=0; i < 5; i++) {
		if (pt_protein->ntile_code[top_of_5tile + i] == C_UP) {
			work[i] = 'U';
		} else if (pt_protein->ntile_code[top_of_5tile + i] == C_DOWN) {
			work[i] = 'D';
		} else {
			work[i] = 'X';
		}
	}
	if (!strncmp("DDDDD", work, 5)) {
		return PE_CW;	/* sheet */
	} else if (!strncmp("DUDUD", work, 5)) {
		return PE_CR;	/* helix */
	} else if (!strncmp("UUDUU", work, 5)) {
//		return PE_CO;	/* turn */	/* NM 2007-04-16 */
		return PE_CY;	/* turn */	/* NM 2007-07-27 */
	} else if (!strncmp("DUDUU", work, 5)) {
//		return PE_CY;	/* turn */
		return PE_CO;	/* C-cap */	/* NM 2007-07-27 */
	} else if (!strncmp("UUDUD", work, 5)) {
//		return PE_CY;	/* turn */
		return PE_CO;	/* N-cap */	/* NM 2007-07-27 */
	} else if (!strncmp("UDDDD", work, 5)) {
		return PE_CG;	/* code G */	/* NM 2007-04-09 */
	} else {
		return PE_CA;
	}
}

/* HEADER and COMPND of the PDB file */
void print_protein_code_header(char * filename, FILE * fp_udcode, char * outfile)
{
	FILE * fp_pdb;
	char PDBLine[LINE_WIDTH + 1];
	
	if ((fp_pdb = fopen(filename, "r")) == NULL) {
		fprintf(STDERR, "print_protein_code_header(): Can't open %s\n",
			filename);
		exit(1);
	}
	fprintf(fp_udcode,
		"----------------------------------------------------------------------------------\n");
	while (fgets(PDBLine, LINE_WIDTH + 1, fp_pdb) != NULL) {
		if (*PDBLine == 'C') {
			if (!strncmp("COMP", PDBLine, 4))
				fprintf(fp_udcode, "|%s|\n", PDBLine);
		} else if (*PDBLine == 'H') {
			if (!strncmp("HEAD", PDBLine, 4))
				fprintf(fp_udcode, "|%s|\n", PDBLine);
		} else if (*PDBLine == '\n') {
			continue;
		} else {
			break;
		}
	}
	fprintf(fp_udcode,
		"----------------------------------------------------------------------------------\n");
	if (fclose(fp_pdb)) {
		fprintf(STDERR, "print_protein_code_header(): Can't close %s\n",
			filename);
		abort();
	}
	if (0)
		fprintf(STDOUT,
			"print_protein_code_header() : DEBUG  PDBLine = %s\n",
			PDBLine);
	fprintf(fp_udcode, "  Num. of tiles per residue = %4.1f",
		my_settings.tiles_per_res);
	if (my_settings.ca_only_flag == TRUE
	    && my_settings.rotate_tile != TRUE) {
		if (my_settings.along_backbone == TRUE)
			fprintf(fp_udcode, "    (N by N along the backbone)");
		
		else
			fprintf(fp_udcode, "    (N by N)");
	} else {
		fprintf(fp_udcode, "    (1 by 1)");
	}
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode,
		"  Initial atom = %d-th, Distance between tiles = %8.3f\n",
		my_protein.best_ind, my_protein.ave_dis);
	if (my_settings.ca_only_flag == TRUE)
		fprintf(fp_udcode,
			"  Average distance between CA atoms = %8.3f\n",
			my_protein.ave_bond_length);
	else
		fprintf(fp_udcode, "  Average distance between atoms = %8.3f\n",
			my_protein.ave_bond_length);
	fprintf(fp_udcode,
		"  Average error between a tile and a CA atom = %8.3f\n",
		my_protein.ave_error);
	/*
	if (my_settings.variable_path_length == TRUE) {
		fprintf(fp_udcode,
			"  Parameters: npath_table size = %d (displacement = %d)\n",
			NP_TBL_SIZE, NP_TBL_DISP);
	}
	fprintf(fp_udcode,
		"  Parameters: dist_factor = %8.3f, error table size = %d\n",
		my_settings.dist_factor, my_settings.error_tab_size);
	 */
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode, "\n");
	return;
}

/* PDB data */
void print_protein_code_pdb_data(FILE * fp_udcode)
{
	int mod_ct, seq_ct, atm_ct;
	
	fprintf(fp_udcode,
		"<PDB_data atoms=\"%d\" residues=\"%d\" sequences=\"%d\" molecules=\"%d\" models=\"%d\">\n",
		my_protein.atm_num, my_protein.res_num, my_protein.seq_num,
		my_protein.mol_num, my_protein.mod_num);
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		fprintf(fp_udcode, "\n");
		fprintf(fp_udcode, "<model No.=\"%d\">\n", mod_ct);
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(fp_udcode,
				"<Atom_sequence No.=\"%d\" from=\"%d-th atom\" to=\"%d-th atom\">\n",
				seq_ct, my_protein.sequence[seq_ct].atom_head,
				my_protein.sequence[seq_ct].atom_tail);
			fprintf(fp_udcode,
				"       [  No. ]  ser.  name res chn num  (       x,        y,        z) occu. id\n");
			for (atm_ct = my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				fprintf(fp_udcode, "<atom> [%6d] %6d ", atm_ct,
					my_protein.atom[atm_ct].serial_num);
				if (my_protein.atom[atm_ct].atom_name == PE_CC)
					fprintf(fp_udcode, " C  ");
				else if (my_protein.atom[atm_ct].atom_name
					 == PE_CA)
					if (my_protein.atom[atm_ct].ntile_head < 0) {	/* 2006-11-03 */
						fprintf(fp_udcode, " CA ");
					} else {
						if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CR) {
							fprintf(fp_udcode, " CR ");
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CY) {
							fprintf(fp_udcode, " CY ");
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CW) {
							fprintf(fp_udcode, " CW ");
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CO) {
							fprintf(fp_udcode, " CO ");	/* NM 2007-04-09 */
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CB) {
							fprintf(fp_udcode, " CB ");	/* NM 2007-04-09 */
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CG) {
							fprintf(fp_udcode, " CG ");	/* NM 2007-04-09 */
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CU) {
							fprintf(fp_udcode, " CU ");	/* NM 2007-04-09 */
						} else if (ntile_to_atom_name(&my_protein, atm_ct) == PE_CV) {
							fprintf(fp_udcode, " CV ");	/* NM 2007-04-09 */
						} else {
							fprintf(fp_udcode, " CA ");
						}
						
					}
				else if (my_protein.atom[atm_ct].atom_name
					 == PE_NN)
					fprintf(fp_udcode, " N  ");
				else if (my_protein.atom[atm_ct].atom_name
					 == PE_RNA)	/* DNA or RNA */
					 fprintf(fp_udcode, "    ");
					 else
					 fprintf(fp_udcode, " UN ");
					 if (my_protein.atom[atm_ct].chain_id == ' ') {
						 /* To use sscanf() in ProteinViwer */
						 my_protein.atom[atm_ct].chain_id = '-';
					 }
					 fprintf(fp_udcode,
						 " %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s </atom>\n",
						 my_protein.atom[atm_ct].res_name,
						 my_protein.atom[atm_ct].chain_id,
						 my_protein.atom[atm_ct].res_seq,
						 my_protein.atom[atm_ct].x,
						 my_protein.atom[atm_ct].y,
						 my_protein.atom[atm_ct].z,
						 my_protein.atom[atm_ct].occupancy,
						 my_protein.atom[atm_ct].seg_id);
			}
			fprintf(fp_udcode, "</Atom_sequence>\n");
			if (seq_ct < my_protein.model[mod_ct].seq_tail)
				fprintf(fp_udcode, "\n");
		}
		fprintf(fp_udcode, "</model>\n");
	}
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode, "</PDB_data>\n");
	fprintf(fp_udcode, "\n");
	return;
}

/* Coordinates */
void print_protein_code_coordinates(FILE * fp_udcode)
{
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode, "<Coordinates>\n");
	fprintf(fp_udcode, "<origin> ( %8.3f, %8.3f, %8.3f ) </origin>\n",
		my_coor.origin.x, my_coor.origin.y, my_coor.origin.z);
	fprintf(fp_udcode, "<x-axis> ( %8.3f, %8.3f, %8.3f ) </x-axis>\n",
		my_coor.axes[0].x, my_coor.axes[0].y, my_coor.axes[0].z);
	fprintf(fp_udcode, "<y-axis> ( %8.3f, %8.3f, %8.3f ) </y-axis>\n",
		my_coor.axes[1].x, my_coor.axes[1].y, my_coor.axes[1].z);
	fprintf(fp_udcode, "<z-axis> ( %8.3f, %8.3f, %8.3f ) </z-axis>\n",
		my_coor.axes[2].x, my_coor.axes[2].y, my_coor.axes[2].z);
	fprintf(fp_udcode, "<w-axis> ( %8.3f, %8.3f, %8.3f ) </w-axis>\n",
		my_coor.axes[3].x, my_coor.axes[3].y, my_coor.axes[3].z);
	fprintf(fp_udcode, "</Coordinates>\n");
	fprintf(fp_udcode, "\n");
	return;
}

/* Tetrahedron tiles */
void print_protein_code_tiles(struct pe_slt_tile * s_tiles, FILE * fp_udcode)
{
	int mod_ct, seq_ct, i;
	int ind;
	
	if (my_protein.tile_num <= 0) return;
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode, "<Slant_tiles tiles=\"%d\">\n", my_protein.tile_num);
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		fprintf(fp_udcode, "\n");
		fprintf(fp_udcode, "<model No.=\"%d\">\n", mod_ct);
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(fp_udcode,
				"<Tile_sequence No.=\"%d\" from=\"%d-th tile\" to=\"%d-th tile\">\n",
				seq_ct, my_protein.sequence[seq_ct].tile_head,
				my_protein.sequence[seq_ct].tile_tail);
			for (i = my_protein.sequence[seq_ct].tile_head;
			     i <= my_protein.sequence[seq_ct].tile_tail; i++) {
				fprintf(fp_udcode,
					"<tile> [%d] x^%d y^%d z^%d w^%d %s,",
					i, s_tiles[i].top[0],
					s_tiles[i].top[1], s_tiles[i].top[2],
					s_tiles[i].top[3],
					my_trn_table[s_tiles[i].type].str);
				if (s_tiles[i].code == C_UP)
					fprintf(fp_udcode, " code= U,");
				
				else if (s_tiles[i].code == C_DOWN)
					fprintf(fp_udcode, " code= D,");
				else
					fprintf(fp_udcode, " code= X,");
				if (s_tiles[i].atom_flag == 0) {
					fprintf(fp_udcode, " atom= 0 (--)");
				} else {
					ind = s_tiles[i].error_atom;
					fprintf(fp_udcode, " atom= %d", ind);
					if (my_protein.atom[ind].atom_name
					    == PE_CC)
						fprintf(fp_udcode, " (C )");
					
					else if (my_protein.atom[ind].atom_name
						 == PE_CA)
						fprintf(fp_udcode, " (CA)");
					
					else if (my_protein.atom[ind].atom_name
						 == PE_NN)
						fprintf(fp_udcode, " (N )");
					
					else
						fprintf(fp_udcode, " (UN)");
				}
				fprintf(fp_udcode, " \n");

				fprintf(fp_udcode,
				"       center= (%8.3f, %8.3f, %8.3f), ",
					s_tiles[i].cx, s_tiles[i].cy,
					s_tiles[i].cz);
				fprintf(fp_udcode,
					"error= %f against atom [%d]",
					s_tiles[i].error,
					s_tiles[i].error_atom);
				if (my_settings.rotate_tile == TRUE) {
					fprintf(fp_udcode, "\n");
					fprintf(fp_udcode,
						"       direction= (%8.3f, %8.3f, %8.3f)",
						s_tiles[i].dx, s_tiles[i].dy,
						s_tiles[i].dz);
					fprintf(fp_udcode, " </tile>\n");
				} else {
					fprintf(fp_udcode, " </tile>\n");
				}
				if ((s_tiles[i].ter > 0)
				    && (i < my_protein.tile_num)) {
					/* end of a sequence */
					seq_ct++;
					fprintf(fp_udcode,
						"</Tile_sequence>\n");
					fprintf(fp_udcode, "\n");
					fprintf(fp_udcode,
"<Tile_sequence No.=\"%d\" from=\"%d-th tile\" to=\"%d-th tile\">\n",
						seq_ct,
						my_protein.sequence[seq_ct].
						tile_head,
						my_protein.sequence[seq_ct].
						tile_tail);
				}
			}
			fprintf(fp_udcode, "</Tile_sequence>\n");
		}
		fprintf(fp_udcode, "</model>\n");
		break;	/* 2006-11-03 */
	}
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode, "</Slant_tiles>\n");
	fprintf(fp_udcode, "\n");
	return;
}

void print_protein_ntile_code(struct pe_protein * pt_protein, FILE * fp_udcode)
{
	int mod_ct, seq_ct, atm_ct, t_ct;
	
	if (pt_protein->ntile_top < 0) return;
	fprintf(fp_udcode, "\n");
	fprintf(fp_udcode,
		"<N-tile_code atoms=\"%d\" residues=\"%d\" sequences=\"%d\" molecules=\"%d\" models=\"%d\">\n",
		pt_protein->atm_num, pt_protein->res_num, pt_protein->seq_num,
		pt_protein->mol_num, pt_protein->mod_num);
	for (mod_ct = 1; mod_ct <= pt_protein->mod_num; mod_ct++) {
		fprintf(fp_udcode, "\n");
		fprintf(fp_udcode, "<model No.=\"%d\">\n", mod_ct);
		for (seq_ct = pt_protein->model[mod_ct].seq_head;
		     seq_ct <= pt_protein->model[mod_ct].seq_tail; seq_ct++) {
			fprintf(fp_udcode,
				"<N-tile_sequence No.=\"%d\" from=\"%d-th atom\" to=\"%d-th atom\">\n",
				seq_ct, pt_protein->sequence[seq_ct].atom_head,
				pt_protein->sequence[seq_ct].atom_tail);
			fprintf(fp_udcode,
				"        [  No. ]  ser.  name res chn num  N-tile_code\n");
			for (atm_ct = pt_protein->sequence[seq_ct].atom_head;
			     atm_ct <= pt_protein->sequence[seq_ct].atom_tail;
			     atm_ct++) {
				fprintf(fp_udcode, "<ntile> [%6d] %6d ", atm_ct,
					pt_protein->atom[atm_ct].serial_num);
				if (pt_protein->atom[atm_ct].atom_name == PE_CC)
					fprintf(fp_udcode, " C  ");
				else if (pt_protein->atom[atm_ct].atom_name
					 == PE_CA)
					fprintf(fp_udcode, " CA ");
				else if (pt_protein->atom[atm_ct].atom_name
					 == PE_NN)
					fprintf(fp_udcode, " N  ");
				else if (pt_protein->atom[atm_ct].atom_name
					 == PE_RNA)	/* DNA or RNA */
					fprintf(fp_udcode, "    ");
				else
					fprintf(fp_udcode, " UN ");
				if (pt_protein->atom[atm_ct].chain_id == ' ') {
					/* To use sscanf() */
					pt_protein->atom[atm_ct].chain_id = '-';
				}
				fprintf(fp_udcode, " %.3s  %c  %4d ",
					pt_protein->atom[atm_ct].res_name,
					pt_protein->atom[atm_ct].chain_id,
					pt_protein->atom[atm_ct].res_seq);
				t_ct = pt_protein->atom[atm_ct].ntile_head;
				if (t_ct >= 0) {
					for (; t_ct < pt_protein->atom[atm_ct].ntile_head
						+ pt_protein->atom[atm_ct].ntile_length;
					     t_ct++) {
						if (pt_protein->ntile_code[t_ct] == C_UP)
							fprintf(fp_udcode, "U");
						else if (pt_protein->ntile_code[t_ct] == C_DOWN)
							fprintf(fp_udcode, "D");
						else
							fprintf(fp_udcode, "X");
					}
				} else {
					for (; t_ct < pt_protein->atom[atm_ct].ntile_head
						+ my_settings.code_length;
						t_ct++) {
							 fprintf(fp_udcode, "-");
					}
				}
				fprintf(fp_udcode, " </ntile>\n");
			}
			fprintf(fp_udcode, "</N-tile_sequence>\n");
			if (seq_ct < pt_protein->model[mod_ct].seq_tail)
				fprintf(fp_udcode, "\n");
		}
		fprintf(fp_udcode, "</model>\n");
		break;
	}
	return;
}

#define EXT_LENGTH (3+1)	/* ***.ent => ***.udcode + '\0' */

void print_protein_code(char * filename, struct pe_slt_tile * s_tiles)
{
	FILE * fp_udcode;
	char * outfile;		/* Out file */
	char * s;		/* Udcode file */
	int str_size;

	/* Set outfile to Udcode file */
	str_size = (int) strlen(filename);
	outfile = (char *) malloc((str_size + EXT_LENGTH) * sizeof(char));
	if (outfile == NULL) {
		fprintf(STDERR, 
"print_protein_code() : ERROR outfile[] allocation failed\n");
		exit(1);
	}
	strcpy(outfile, filename);
	s = strrchr(outfile, '.');	/* s = ".ent" */
	if ((s == NULL) 
	    || ((strcmp(s, ".ent") != 0) && (strcmp(s, ".pdb") != 0))) {	/* 2007-01-16 */
		fprintf(STDERR,
			"print_protein_code(): %s is not \".ent\" file\n",
			filename);
		exit(1);
	}
	strcpy(s+1, "udcode\0");
	fprintf(STDOUT, "OUTPUT FILE: %s\n", outfile);
	if ((fp_udcode = fopen(outfile, "w")) == NULL) {
		fprintf(STDERR, "print_protein_code(): can't open %s\n", s);
		exit(1);
	}
	
	print_protein_code_header(filename, fp_udcode, outfile);
	print_protein_code_pdb_data(fp_udcode);
	if (my_protein.ntile_top < 0) {
		print_protein_code_coordinates(fp_udcode);
		print_protein_code_tiles(s_tiles, fp_udcode);
	} else {
		print_protein_ntile_code(&my_protein, fp_udcode);
	}
		    
	if (fclose(fp_udcode)) {
		fprintf(STDERR, "print_protein_code(): Can't close %s\n",
			outfile);
		abort();
	}
	if (outfile != NULL) free(outfile);
	return;
}

char resname_to_char(char *r_name)
{
	char work_name[4] ="---\0";
	
	switch (*r_name) {
		case ('A'):if (!strncmp("ALA", r_name, 3)) {
				return 'A';
			} else if (!strncmp("ASP", r_name, 3)) {
				return 'D';
			} else if (!strncmp("ARG", r_name, 3)) {
				return 'R';
			} else if (!strncmp("ASN", r_name, 3)) {
				return 'N';
			} else if (!strncmp("ASX", r_name, 3)) {
				return 'B';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
		}
			
		case ('C'):
			if (!strncmp("CYS", r_name, 3)) {
				return 'C';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('G'):
			if (!strncmp("GLY", r_name, 3)) {
				return 'G';
			} else if (!strncmp("GLU", r_name, 3)) {
				return 'E';
			} else if (!strncmp("GLN", r_name, 3)) {
				return 'Q';
			} else if (!strncmp("GLX", r_name, 3)) {
				return 'Z';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('H'):
			if (!strncmp("HIS", r_name, 3)) {
				return 'H';
			} else if (!strncmp("HYP", r_name, 3)) {
				return '-';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('I'):
			if (!strncmp("ILE", r_name, 3)) {
				return 'I';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('L'):
			if (!strncmp("LEU", r_name, 3)) {
				return 'L';
			} else if (!strncmp("LYS", r_name, 3)) {
				return 'K';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('M'):
			if (!strncmp("MET", r_name, 3)) {
				return 'M';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('P'):
			if (!strncmp("PHE", r_name, 3)) {
				return 'F';
			} else if (!strncmp("PRO", r_name, 3)) {
				return 'P';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('S'):
			if (!strncmp("SER", r_name, 3)) {
				return 'S';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('T'):
			if (!strncmp("THR", r_name, 3)) {
				return 'T';
			} else if (!strncmp("TYR", r_name, 3)) {
				return 'Y';
			} else if (!strncmp("TRP", r_name, 3)) {
				return 'W';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('U'):
			if (!strncmp("UNK", r_name, 3)) {
				return '=';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		case ('V'):
			if (!strncmp("VAL", r_name, 3)) {
				return 'V';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
		
		case (' '):	// 2007-08-22
			if (!strncmp("G", r_name + 2, 1)) {	// '  G'
				return 'G';
			} else if (!strncmp("A", r_name + 2, 1)) {
				return 'A';
			} else if (!strncmp("C", r_name + 2, 1)) {
				return 'C';
			} else if (!strncmp("T", r_name + 2, 1)) {
				return 'T';
			} else if (!strncmp("U", r_name + 2, 1)) {
				return 'U';
			} else {
				strncpy(work_name, r_name, 3);
				fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
				return '?';
			}
			
		default:
			strncpy(work_name, r_name, 3);
			fprintf(STDERR, "resname_to_char() : CAUTION invalid residue name *%s*\n", work_name);
			return '?';
			
	}
}

/*
 * R denotes UUDUU, Q denotes UUDUD, ...,
 * B denotes DUDUU, A denotes DUDUD, ...,
 * 0 denotes DDDDD
 */
char ntile_to_char(int atm_ct)
{
	int code_top;
	int i;
	int val;
	
	code_top = my_protein.atom[atm_ct].ntile_head;
	if (code_top < 0) {
		return '-';
	}
	
	val = 0;
	for (i = 0; i < my_protein.atom[atm_ct].ntile_length; i++) {
		if (my_protein.ntile_code[code_top + i] == C_UP) {
			val = 2 * val + 1;
		} else if (my_protein.ntile_code[code_top + i] == C_DOWN) {
			val = 2 * val;
		} else {
			return '=';
		}
	}
	switch (val) {
		case 0: return '0'; case 1: return '1'; case 2: return '2';
		case 3: return '3'; case 4: return '4'; case 5: return '5';
		case 6: return '6'; case 7: return '7'; case 8: return '8';
		case 9: return '9'; case 10: return 'A'; case 11: return 'B';
		case 12: return 'C'; case 13: return 'D'; case 14: return 'E';
		case 15: return 'F'; case 16: return 'G'; case 17: return 'H';
		case 18: return 'I'; case 19: return 'J'; case 20: return 'K';
		case 21: return 'L'; case 22: return 'M'; case 23: return 'N';
		case 24: return 'O'; case 25: return 'P'; case 26: return 'Q';
		case 27: return 'R'; case 28: return 'S'; case 29: return 'T';
		case 30: return 'U'; case 31: return 'V'; case 32: return 'W';
		case 33: return 'X'; case 34: return 'Y'; case 35: return 'Z';
		default: return '?';
	}

}

void print_ntcode(FILE * fp_code, char * filename, int seq_ct)
{
	int atm_ct;
	char type = ' ';
	
	fprintf(fp_code, "[Res.] ");
	for (atm_ct = my_protein.sequence[seq_ct].atom_head;
	     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
	     atm_ct++) {
		if (THIS_IS_NOT_CA_ATOM(atm_ct))
			continue;
		type = resname_to_char(my_protein.atom[atm_ct].res_name);
		fprintf(fp_code, "%c", type);
	}
	fprintf(fp_code, " : %s c%d\n", filename, seq_ct);
	
	fprintf(fp_code, "[Code] ");
	for (atm_ct = my_protein.sequence[seq_ct].atom_head;
	     atm_ct <= my_protein.sequence[seq_ct].atom_tail;
	     atm_ct++) {
		if (THIS_IS_NOT_CA_ATOM(atm_ct))
			continue;
		type = ntile_to_char(atm_ct);
		fprintf(fp_code, "%c", type);
	}
	fprintf(fp_code, " : %s c%d\n", filename, seq_ct);
	return;
}

#define EXT1_LENGTH (1+1)	/* ***.ent => ***.code + '\0' */

void print_udcode(char * filename, struct pe_slt_tile * s_tiles)
{
	FILE * fp_code;
	char * outfile;		/* Out file */
	char * s;		/* Udcode file */
	int str_size;
	int mod_ct, seq_ct, t_ct;
	
	/* Set outfile to Udcode file */
	str_size = (int) strlen(filename);
	outfile = (char *) malloc((str_size + EXT1_LENGTH) * sizeof(char));
	if (outfile == NULL) {
		fprintf(STDERR, 
			"print_udcode() : ERROR outfile[] allocation failed\n");
		exit(1);
	}
	strcpy(outfile, filename);
	s = strrchr(outfile, '.');	/* s = ".ent" */
	strcpy(s+1, "code\0");
	fprintf(STDOUT, "CODE FILE: %s\n", outfile);
	if ((fp_code = fopen(outfile, "w")) == NULL) {
		fprintf(STDERR, "print_udcode(): can't open %s\n", s);
		exit(1);
	}
	
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			if (my_protein.ntile_top >= 0) {
				print_ntcode(fp_code, filename, seq_ct);
			} else if (my_protein.sequence[seq_ct].tile_tail == 0) {
				fprintf(fp_code, "No code : %s c%d\n", filename, seq_ct);
			} else {
				for (t_ct = my_protein.sequence[seq_ct].tile_head;
				     t_ct <= my_protein.sequence[seq_ct].tile_tail;
				     t_ct++) {
					if (s_tiles[t_ct].code == C_UP)
						fprintf(fp_code, "U");
					else if (s_tiles[t_ct].code == C_DOWN)
						fprintf(fp_code, "D");
					else
						fprintf(fp_code, "X");
				}
				fprintf(fp_code, " : %s c%d\n", filename, seq_ct);
			}
		}
		break;
	}
		    
	if (fclose(fp_code)) {
		fprintf(STDERR, "print_protein_code(): Can't close %s\n",
			outfile);
		abort();
	}
	if (outfile != NULL) free(outfile);
	return;
}
